#!/usr/bin/env perl

use 5.10.1;
use strict;
use warnings;

use MIME::Base64;
use Encode;
use Bugzilla::Util qw(html_quote);

# For local debugging
#use Data::Dumper;
#local $Data::Dumper::Indent    = 1;
#local $Data::Dumper::Sortkeys  = 1;

our %errors = (

  400 => {
    title => 'Bad Request',
    text  => 'The server was unable to process this request'
  },
  401 => {
    title => ' Authentication Required',
    text  =>
      q{This server could not verify that you are authorized to access that url. You either supplied the wrong credentials
                (e.g. bad password), or your browser doesn't understand how to supply the credentials required.}
  },
  403 => {
    title => 'Access Denied',
    text  => 'Access to the requested resource has been denied.'
  },
  404 => {
    title => 'Object Not Found',
    text  => 'The requested URL was not found on this server.'
  },
  500 => {
    title => 'Unrecoverable Error',
    text  =>
      'The server encountered an error and was unable to complete your request.'
  },
);

# For JSON errors the error will have already been printed.
my $contype
  = defined $ENV{'REDIRECT_CONTENT_TYPE'}
  ? $ENV{'REDIRECT_CONTENT_TYPE'}
  : 'text/html';

my $html = ($contype !~ m/json/ && defined $errors{$ENV{'REDIRECT_STATUS'}});

my $email
  = $ENV{'REDIRECT_SERVER_ADMIN'} ? $ENV{'REDIRECT_SERVER_ADMIN'}
  : $ENV{'SERVER_ADMIN'}          ? $ENV{'SERVER_ADMIN'}
  :                                 'unset';

my $zip = defined $ENV{'REDIRECT_HTTP_ACCEPT_ENCODING'}
  && $ENV{'REDIRECT_HTTP_ACCEPT_ENCODING'} =~ /zip/;

# BUGBUG Disabled until we work out how to capture printed data and reenable
# zip in HTTPD conf for API end points
$zip = 0;

# Only reprint JSON messages if gzip is set.
# Under HTTPD gzip will cause the already printed error messages to be eaten ...
_print_error($html, $ENV{'REDIRECT_STATUS'}, $ENV{'REDIRECT_ERROR_NOTES'},
  $email, $contype, $zip)
  if ($ENV{'REDIRECT_STATUS'} >= 400 && ($html || $zip));

exit;

sub _print_error {
  my ($html, $code, $msg, $email, $content_type, $zip) = @_;

  my $title = $errors{$code}->{title};
  my $text  = $errors{$code}->{text};

  unless ($msg) {
    if (exists $ENV{REDIRECT_BZ_ERROR_NOTES}) {
      $msg = decode('utf-8', decode_base64($ENV{REDIRECT_BZ_ERROR_NOTES}));
    }
    else {
      $msg = "No data";
    }
  }

  say "content-type: " . $content_type;
  if ($zip) {
    say "Content-Encoding: identity";
  }

  my $template = $msg;
  if ($html) {
    $msg      = html_quote($msg);
    $template = <<EOT;
<!DOCTYPE HTML>
<html>

<head>
    <meta charset="UTF-8">
    <title>Red Hat Bugzilla - $code $title</title>
    <link type="text/css" href="/skins/standard/global.css" rel="stylesheet">
    <link type="text/css" href="/extensions/RedHat/web/css/redhat.css" rel="stylesheet">
</head>

<body>
    <div id="header">
        <div id="titles">
            <span class="logo-crop">
                <img src="/extensions/RedHat/web/css/l_redhat-lg.png" alt="logo">
            </span>
            <span id="title">Red Hat Bugzilla - $title</span>
        </div>
    </div>
    <div id="bugzilla-body">
        <div class='message'>
            <h1>$title</h1>
            <p>$text</p>
        </div>
        <pre>
$msg
        </pre>
        <div>
            For administrative assistance please contact <a href="mailto:$email">$email</a>
        </div>
    </div>
</body>

</html>
EOT
  }

  say "Status: $code";
  say "";    # This ends the header.
  say $template;

  return;
}
