# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# This Source Code Form is "Incompatible With Secondary Licenses", as
# defined by the Mozilla Public License, v. 2.0.

package Bugzilla::Extension::ActivityReport::Reports;
use strict;
use warnings;
use 5.10.1;

use Bugzilla::Constants;
use Bugzilla::Error;
use Bugzilla::Field;
use Bugzilla::User;
use Bugzilla::Util qw(trim detaint_natural trick_taint correct_urlbase);

use Date::Parse;
use DateTime;
use JSON qw(-convert_blessed_universally);
use List::MoreUtils qw(uniq);

use base qw(Exporter);

our @EXPORT_OK = qw(user_activity_report
  user_admin_activity_report
  admin_activity_report
  group_admins_report
  email_queue_report
  group_membership_report);

sub user_activity_report {
  my ($vars) = @_;
  my $dbh    = Bugzilla->dbh;
  my $input  = Bugzilla->input_params;

  my @who    = ();
  my $from   = trim($input->{'from'} || '');
  my $to     = trim($input->{'to'} || '');
  my $action = $input->{'action'} || '';

  # fix non-breaking hyphens
  $from =~ s/\N{U+2011}/-/g;
  $to   =~ s/\N{U+2011}/-/g;

  if ($from eq '') {
    my $dt = DateTime->now()->subtract('weeks' => 8);
    $from = $dt->ymd('-');
  }
  if ($to eq '') {
    my $dt = DateTime->now();
    $to = $dt->ymd('-');
  }

  if ($action eq 'run') {
    if ($input->{'who'} eq '') {
      ThrowUserError('user_activity_missing_username');
    }
    Bugzilla::User::match_field({'who' => {'type' => 'multi'}});

    my $from_dt = _string_to_datetime($from);
    $from = $from_dt->ymd();

    my $to_dt = _string_to_datetime($to);
    $to = $to_dt->ymd();

    # add one day to include all activity that happened on the 'to' date
    $to_dt->add(days => 1);

    my ($activity_joins,    $activity_where)    = ('', '');
    my ($attachments_joins, $attachments_where) = ('', '');
    if (Bugzilla->params->{"insidergroup"}
      && !Bugzilla->user->in_group(Bugzilla->params->{'insidergroup'}))
    {
      $activity_joins = "LEFT JOIN attachments
                       ON attachments.attach_id = bugs_activity.attach_id";
      $activity_where    = "AND COALESCE(attachments.isprivate, 0) = 0";
      $attachments_where = $activity_where;
    }

    my @who_bits;
    foreach my $who (ref $input->{'who'} ? @{$input->{'who'}} : $input->{'who'}) {
      push @who,      $who;
      push @who_bits, '?';
    }
    my $who_bits = join(',', @who_bits);

    if (!@who) {
      my $template = Bugzilla->template;
      my $cgi      = Bugzilla->cgi;
      my $new_vars = {};
      $new_vars->{'script'}        = $cgi->url(-relative => 1);
      $new_vars->{'fields'}        = {};
      $new_vars->{'matches'}       = [];
      $new_vars->{'matchsuccess'}  = 0;
      $new_vars->{'matchmultiple'} = 1;
      print $cgi->header();
      $template->process("global/confirm-user-match.html.tmpl", $new_vars)
        || ThrowTemplateError($template->error());
      exit;
    }

    $from_dt = $from_dt->ymd() . ' 00:00:00';
    $to_dt   = $to_dt->ymd() . ' 23:59:59';
    my @params;
    for (1 .. 4) {
      push @params, @who;
      push @params, ($from_dt, $to_dt);
    }

    my $order
      = ($input->{'sort'} && $input->{'sort'} eq 'bug')
      ? 'bug_id, bug_when'
      : 'bug_when';

    my $comment_filter = '';
    if (!Bugzilla->user->is_insider) {
      $comment_filter = 'AND longdescs.isprivate = 0';
    }

    my $query = "
        SELECT 
                   fielddefs.name,
                   bugs_activity.bug_id,
                   bugs_activity.attach_id,
                   "
      . $dbh->sql_date_format('bugs_activity.bug_when', '%Y.%m.%d %H:%i:%s')
      . " AS ts,
                   bugs_activity.removed,
                   bugs_activity.added,
                   profiles.login_name,
                   bugs_activity.comment_id,
                   bugs_activity.bug_when,
                   bugs_activity.object_id
              FROM bugs_activity
                   $activity_joins
         LEFT JOIN fielddefs
                ON bugs_activity.fieldid = fielddefs.id
        INNER JOIN profiles
                ON profiles.userid = bugs_activity.who
             WHERE profiles.login_name IN ($who_bits)
                   AND bugs_activity.bug_when >= ? AND bugs_activity.bug_when <= ?
                   $activity_where

        UNION ALL

        SELECT 
                   'bug_id' AS name,
                   bugs.bug_id,
                   NULL AS attach_id,
                   "
      . $dbh->sql_date_format('bugs.creation_ts', '%Y.%m.%d %H:%i:%s') . " AS ts,
                   '(new bug)' AS removed,
                   bugs.short_desc AS added,
                   profiles.login_name,
                   NULL AS comment_id,
                   bugs.creation_ts AS bug_when,
                   NULL AS object_id
              FROM bugs
        INNER JOIN profiles
                ON profiles.userid = bugs.reporter
             WHERE profiles.login_name IN ($who_bits)
                   AND bugs.creation_ts >= ? AND bugs.creation_ts <= ?

        UNION ALL

        SELECT 
                   'longdesc' AS name,
                   longdescs.bug_id,
                   NULL AS attach_id,
                   "
      . $dbh->sql_date_format('longdescs.bug_when', '%Y.%m.%d %H:%i:%s') . " AS ts,
                   '' AS removed,
                   '' AS added,
                   profiles.login_name,
                   longdescs.comment_id AS comment_id,
                   longdescs.bug_when,
                   NULL AS object_id
              FROM longdescs
        INNER JOIN profiles
                ON profiles.userid = longdescs.who
             WHERE profiles.login_name IN ($who_bits)
                   AND longdescs.bug_when >= ? AND longdescs.bug_when <= ?
                   $comment_filter

        UNION ALL

        SELECT 
                   'attachments.description' AS name,
                   attachments.bug_id,
                   attachments.attach_id,
                   "
      . $dbh->sql_date_format('attachments.creation_ts', '%Y.%m.%d %H:%i:%s')
      . " AS ts,
                   '(new attachment)' AS removed,
                   attachments.description AS added,
                   profiles.login_name,
                   NULL AS comment_id,
                   attachments.creation_ts AS bug_when,
                   NULL AS object_id
              FROM attachments
        INNER JOIN profiles
                ON profiles.userid = attachments.submitter_id
             WHERE profiles.login_name IN ($who_bits)
                   AND attachments.creation_ts >= ? AND attachments.creation_ts <= ?
                   $attachments_where

          ORDER BY $order ";

    my $list = $dbh->selectall_arrayref($query, undef, @params);

    if ($input->{debug}) {
      while (my $param = shift @params) {
        $query =~ s/\?/$dbh->quote($param)/e;
      }
      $vars->{debug_sql} = $query;
    }

    my @operations;
    my $operation       = {};
    my $changes         = [];
    my $incomplete_data = 0;
    my %bug_ids;

    foreach my $entry (@$list) {
      my (
        $fieldname, $bugid, $attachid,   $when, $removed,
        $added,     $who,   $comment_id, undef, $object_id
      ) = @$entry;
      my %change;
      my $activity_visible = 1;

      next unless Bugzilla->user->can_see_bug($bugid);

      # check if the user should see this field's activity
      if ( $fieldname eq 'remaining_time'
        || $fieldname eq 'estimated_time'
        || $fieldname eq 'work_time'
        || $fieldname eq 'deadline')
      {
        $activity_visible = Bugzilla->user->is_timetracker;
      }
      elsif ($fieldname eq 'longdescs.isprivate'
        && !Bugzilla->user->is_insider
        && $added)
      {
        $activity_visible = 0;
      }
      else {
        $activity_visible = 1;
      }

## REDHAT EXTENSION BEGIN 842419
      if ($activity_visible) {
        Bugzilla::Hook::process(
          'bug_filter_change',
          {
            user      => Bugzilla->user,
            field     => $fieldname,
            added     => $added,
            removed   => $removed,
            object_id => $object_id,
            visible   => \$activity_visible,
          }
        );
      }
## REDHAT EXTENSION END 842419

      if ($activity_visible) {

        # Check for the results of an old Bugzilla data corruption bug
        if ( ($added eq '?' && $removed eq '?')
          || ($added =~ /^\? / || $removed =~ /^\? /))
        {
          $incomplete_data = 1;
        }

        # Start a new changeset if required (depends on the sort order)
        my $is_new_changeset;
        if ($order eq 'bug_when') {
          $is_new_changeset
            = $operation->{'who'}
            && ($who ne $operation->{'who'}
            || $when ne $operation->{'when'}
            || $bugid != $operation->{'bug'});
        }
        else {
          $is_new_changeset = $operation->{'bug'} && $bugid != $operation->{'bug'};
        }
        if ($is_new_changeset) {
          $operation->{'changes'} = $changes;
          push(@operations, $operation);
          $operation = {};
          $changes   = [];
        }

        $bug_ids{$bugid} = 1;

        $operation->{'bug'}  = $bugid;
        $operation->{'who'}  = $who;
        $operation->{'when'} = $when;

        $change{'fieldname'} = $fieldname;
        $change{'attachid'}  = $attachid;
        $change{'removed'}   = $removed;
        $change{'added'}     = $added;
        $change{'when'}      = $when;

        if ($comment_id) {
          $change{'comment'} = Bugzilla::Comment->new($comment_id);
          next if $change{'comment'}->count == 0;
        }

        if ($attachid) {
          $change{'attach'} = Bugzilla::Attachment->new($attachid);
        }

        push(@$changes, \%change);
      }
    }

    if ($operation->{'who'}) {
      $operation->{'changes'} = $changes;
      push(@operations, $operation);
    }

    $vars->{'incomplete_data'} = $incomplete_data;
    $vars->{'operations'}      = \@operations;

    my @bug_ids = sort { $a <=> $b } keys %bug_ids;
    $vars->{'bug_ids'} = \@bug_ids;
  }

  $vars->{'action'}    = $action;
  $vars->{'who'}       = join(',', @who);
  $vars->{'who_count'} = scalar @who;
  $vars->{'from'}      = $from;
  $vars->{'to'}        = $to;
  $vars->{'sort'}      = $input->{'sort'};

  return;
}

sub _string_to_datetime {
  my $input = shift;
  my $time  = _parse_date($input)
    or ThrowUserError('report_invalid_date', {date => $input});
  return _time_to_datetime($time);
}

sub _time_to_datetime {
  my $time = shift;
  return DateTime->from_epoch(epoch => $time)->set_time_zone('local')
    ->truncate(to => 'day');
}

sub _parse_date {
  my ($str) = @_;
  if ($str =~ /^(-|\+)?(\d+)([hHdDwWmMyY])$/) {

    # relative date
    my ($sign, $amount, $unit, $date) = ($1, $2, lc $3, time);
    my ($sec, $min, $hour, $mday, $month, $year, $wday) = localtime($date);
    $amount = -$amount if $sign && $sign eq '+';
    if ($unit eq 'w') {

      # convert weeks to days
      $amount = 7 * $amount + $wday;
      $unit   = 'd';
    }
    if ($unit eq 'd') {
      $date -= $sec + 60 * $min + 3600 * $hour + 24 * 3600 * $amount;
      return $date;
    }
    elsif ($unit eq 'y') {
      return str2time(sprintf("%4d-01-01 00:00:00", $year + 1900 - $amount));
    }
    elsif ($unit eq 'm') {
      $month -= $amount;
      while ($month < 0) { $year--; $month += 12; }
      return str2time(sprintf("%4d-%02d-01 00:00:00", $year + 1900, $month + 1));
    }
    elsif ($unit eq 'h') {

      # Special case 0h for 'beginning of this hour'
      if ($amount == 0) {
        $date -= $sec + 60 * $min;
      }
      else {
        $date -= 3600 * $amount;
      }
      return $date;
    }
    return;
  }
  return str2time($str);
}


sub group_admins_report {
  my ($vars) = @_;
  my $dbh    = Bugzilla->dbh;
  my $user   = Bugzilla->user;

  $user->in_group('editusers')
    || ThrowUserError('auth_failure',
    {group => 'editusers', action => 'run', object => 'group_admins'});

  my $query = "
        SELECT groups.name, "
    . $dbh->sql_group_concat('profiles.login_name', "','", 1) . "
          FROM groups
               LEFT JOIN user_group_map
                    ON user_group_map.group_id = groups.id
                    AND user_group_map.isbless = 1
                    AND user_group_map.grant_type = 0
               LEFT JOIN profiles
                    ON user_group_map.user_id = profiles.userid
         WHERE groups.isbuggroup = 1
      GROUP BY groups.name";

  my @groups;
  foreach my $group (@{$dbh->selectall_arrayref($query)}) {
    my @admins;
    if ($group->[1]) {
      foreach my $admin (split(/,/, $group->[1])) {
        push(@admins, Bugzilla::User->new({name => $admin}));
      }
    }
    push(@groups, {name => $group->[0], admins => \@admins});
  }

  $vars->{'groups'} = \@groups;

  return;
}

sub group_membership_report {
  my ($page, $vars) = @_;
  my $dbh  = Bugzilla->dbh;
  my $user = Bugzilla->user;
  my $cgi  = Bugzilla->cgi;

  $user->in_group('editusers')
    || ThrowUserError('auth_failure',
    {group => 'editusers', action => 'run', object => 'group_admins'});

  my $who = $cgi->param('who');
  if (!defined($who) || $who eq '') {
    if ($page eq 'group_membership.txt') {
      print $cgi->redirect("page.cgi?id=group_membership.html&output=txt");
      exit;
    }
    $vars->{'output'} = $cgi->param('output');
    return;
  }

  Bugzilla::User::match_field({'who' => {'type' => 'multi'}});
  $who = Bugzilla->input_params->{'who'};
  $who = ref($who) ? $who : [$who];

  my @users;
  foreach my $login (@$who) {
    my $u = Bugzilla::User->new(login_to_id($login, 1));

    # this is lifted from $user->groups()
    # we need to show which groups are direct and which are inherited

    my $groups_to_check = $dbh->selectcol_arrayref(
      q{SELECT DISTINCT group_id
                FROM user_group_map
               WHERE user_id = ? AND isbless = 0}, undef, $u->id
    );

    my $rows = $dbh->selectall_arrayref(
      "SELECT DISTINCT grantor_id, member_id
               FROM group_group_map
              WHERE grant_type = " . GROUP_MEMBERSHIP
    );

    my %group_membership;
    foreach my $row (@$rows) {
      my ($grantor_id, $member_id) = @$row;
      push(@{$group_membership{$member_id}}, $grantor_id);
    }

    my %checked_groups;
    my %direct_groups;
    my %indirect_groups;
    my %groups;

    foreach my $member_id (@$groups_to_check) {
      $direct_groups{$member_id} = 1;
    }

    while (scalar(@$groups_to_check) > 0) {
      my $member_id = shift @$groups_to_check;
      if (!$checked_groups{$member_id}) {
        $checked_groups{$member_id} = 1;
        my $members      = $group_membership{$member_id};
        my @new_to_check = grep(!$checked_groups{$_}, @$members);
        push(@$groups_to_check, @new_to_check);
        foreach my $id (@new_to_check) {
          $indirect_groups{$id} = $member_id;
        }
        $groups{$member_id} = 1;
      }
    }

    my @groups;
    my $ra_groups = Bugzilla::Group->new_from_list([keys %groups]);
    foreach my $group (@$ra_groups) {
      my $via;
      if ($direct_groups{$group->id}) {
        $via = '';
      }
      else {
        foreach my $g (@$ra_groups) {
          if ($g->id == $indirect_groups{$group->id}) {
            $via = $g->name;
            last;
          }
        }
      }
      push @groups, {name => $group->name, desc => $group->description, via => $via,};
    }

    push @users, {user => $u, groups => \@groups,};
  }

  $vars->{'who'}   = $who;
  $vars->{'users'} = \@users;

  return;
}

sub email_queue_report {
  my ($vars, $filter) = @_;
  my $dbh  = Bugzilla->dbh;
  my $user = Bugzilla->user;

  ## RED HAT EXTENSION START 1579624
  $user->is_insider
    || ThrowUserError('auth_failure',
    {group => 'admin', action => 'run', object => 'email_queue'});

  my $query = <<'EOQ';
        SELECT
            j.jobid,
            j.insert_time,
            j.run_after,
            j.grabbed_until,
            COUNT(e.jobid) AS error_count,
            MAX(e.error_time) AS error_time,
            (
                SELECT message
                FROM ts_error tse
                WHERE tse.jobid = j.jobid
                ORDER BY tse.error_time DESC
                LIMIT 1
            ) AS error_message,
            fm.funcname
        FROM ts_job j
            LEFT JOIN ts_error e ON e.jobid = j.jobid
            LEFT JOIN ts_funcmap fm ON j.funcid = fm.funcid
        GROUP BY j.jobid, fm.funcname
        ORDER BY j.run_after
        LIMIT 500
EOQ

  $vars->{'num_jobs'} = $dbh->selectall_arrayref('select count(*) from ts_job');
  ## RED HAT EXTENSION END 1579624

  ## RED HAT EXTENSION START 1720418
  $vars->{'job_summary'}
    = $dbh->selectall_arrayref(
    'select count(*), ts_funcmap.funcname from ts_job left join ts_funcmap on ts_job.funcid = ts_funcmap.funcid group by ts_funcmap.funcname order by funcname'
    );
  ## RED HAT EXTENSION END 1720418

  $vars->{'jobs'} = $dbh->selectall_arrayref($query, {Slice => {}});
  $vars->{'now'}  = (time);

  return;
}


sub user_admin_activity_report {
  my ($vars) = @_;

  my $dbh   = Bugzilla->dbh;
  my $user  = Bugzilla->user;
  my $input = Bugzilla->input_params;

  $user->in_group('admin')
    || ThrowUserError('auth_failure',
    {group => 'admin', action => 'run', object => 'group_admins'});


  my $from = trim($input->{'from'} || '');
  my $to   = trim($input->{'to'}   || '');
  my $who  = $input->{'who'};
  my $action = $input->{'action'} || '';
  my @list;

  if ($action eq 'run') {

    if ($input->{'who'} eq '') {
      ThrowUserError('user_activity_missing_username');
    }

    my $wuser = Bugzilla::User->check($who);

    # fix non-breaking hyphens
    $from =~ s/\N{U+2011}/-/g;
    $to   =~ s/\N{U+2011}/-/g;

    if ($from eq '') {
      my $dt = DateTime->now()->subtract('weeks' => 8);
      $from = $dt->ymd('-');
    }
    if ($to eq '') {
      my $dt = DateTime->now();
      $to = $dt->ymd('-');
    }

    my $from_dt = _string_to_datetime($from);
    $from    = $from_dt->ymd();
    $from_dt = $from_dt->ymd() . ' 00:00:00';

    my $to_dt = _string_to_datetime($to);
    $to = $to_dt->ymd();

    # add one day to include all activity that happened on the 'to' date
    $to_dt->add(days => 1);
    $to_dt = $to_dt->ymd() . ' 23:59:59';

    my $class_strings = <<EOT;
'Bugzilla::Extension::ExternalBugs::Bug',
'Bugzilla::Extension::ListOfBugs::BugList',
'Bugzilla::Extension::ExternalBugs::Type',
'Bugzilla::Extension::DependentProducts::Record',
'Bugzilla::Extension::RuleEngine::RuleDetail',
'Bugzilla::Extension::AgileTools::Pool'
EOT

    my $query = <<EOS;
SELECT *
FROM audit_log
WHERE user_id = ?
AND at_time BETWEEN ? AND ?
AND class NOT IN ($class_strings)
ORDER BY at_time
EOS

    my $list
      = $dbh->selectall_arrayref($query, undef, ($wuser->id, $from_dt, $to_dt));

    foreach my $row (@{$list}) {
      my $id    = $row->[2];
      my $class = $row->[1];
      $class =~ m/::([^:]*)$/;
      my $name = "$1 - $id";
      my $link = '';
      my $obj;
      if ($class =~ /(Bugzilla::Field)::Choice::(.*)/) {
        $class = $1;
        $name  = $2;
        $obj   = $class->new({name => $name});
      }
      elsif ($class =~ /(Bugzilla::Extension::ExternalBugs::Type)::(.*)/) {
        $class = $1;
        $name  = "ExternalBugs::Type::$2 - $id";
        $obj   = $class->new($id);
      }
      elsif ($class =~ /Bugzilla::Extension::Releases/) {
        $obj = $class->new($id);
      }
      else {
        $obj = $class->new({id => $id});
        $obj = $class->new($id) unless ($obj);
      }
      if ($obj) {
        $link = $obj->admin_link;
      }
      else {
        $link = "$name - DELETED";
      }

      push(@list, [$wuser->login, $link, $row->[3], $row->[4], $row->[5], $row->[6]]);
    }
  }

  $vars->{'action'} = $action;
  $vars->{'who'}    = $who;
  $vars->{'from'}   = $from;
  $vars->{'to'}     = $to;
  $vars->{list}     = \@list;

  return;
}


sub admin_activity_report {
  my ($vars) = @_;

  my $dbh   = Bugzilla->dbh;
  my $user  = Bugzilla->user;
  my $input = Bugzilla->input_params;

  $user->in_group('admin')
    || ThrowUserError('auth_failure',
    {group => 'admin', action => 'run', object => 'group_admins'});


  my $from = trim($input->{'from'} || '');
  my $to   = trim($input->{'to'}   || '');
  my @obj_ids
    = $input->{'object_ids'} ? split(/,\s*/, $input->{'object_ids'}) : ();
  my $action      = $input->{'action'} || '';
  my $cur_classes = $input->{'cur_classes'} // [];
  my $who         = $input->{'who'};

  if (ref $cur_classes ne 'ARRAY') {
    $cur_classes = [$cur_classes];
  }

  my @list;
  my %users;

  if ($action eq 'run') {

    my $wuser;

    $wuser = Bugzilla::User->check($who) if ($who);

    # fix non-breaking hyphens
    $from =~ s/\N{U+2011}/-/g;
    $to   =~ s/\N{U+2011}/-/g;

    if ($from eq '') {
      my $dt = DateTime->now()->subtract('weeks' => 8);
      $from = $dt->ymd('-');
    }
    if ($to eq '') {
      my $dt = DateTime->now();
      $to = $dt->ymd('-');
    }

    my $from_dt = _string_to_datetime($from);
    $from    = $from_dt->ymd();
    $from_dt = $from_dt->ymd() . ' 00:00:00';

    my $to_dt = _string_to_datetime($to);
    $to = $to_dt->ymd();

    # add one day to include all activity that happened on the 'to' date
    $to_dt->add(days => 1);
    $to_dt = $to_dt->ymd() . ' 23:59:59';

    my $class_strings = '';

    if (@$cur_classes) {
      my $qms = "?, " x @$cur_classes;
      chop($qms);
      chop($qms);
      $class_strings = "AND class IN ($qms)";
    }

    my $obj_id = '';
    if (@obj_ids) {
      my $qms = "?, " x @obj_ids;
      chop($qms);
      chop($qms);
      $obj_id = "AND object_id IN($qms)";
    }

    my $user_sql = '';
    if ($wuser) {
      $user_sql = "AND user_id = ?";
    }

    my $query = <<EOS;
SELECT *
FROM audit_log
WHERE
at_time BETWEEN ? AND ?
$user_sql
$obj_id
$class_strings
ORDER BY at_time
EOS

    my @args = ($from_dt, $to_dt);

    if ($wuser) {
      trick_taint($who);
      push(@args, $wuser->id);
    }

    foreach my $id (@obj_ids) {
      trick_taint($id);
      push(@args, $id);
    }

    foreach my $class (@$cur_classes) {
      trick_taint($class);
      push(@args, $class);
    }

    my $list = $dbh->selectall_arrayref($query, undef, @args);

    foreach my $row (@{$list}) {
      my $id    = $row->[2];
      my $class = $row->[1];
      $class =~ m/::([^:]*)$/;
      my $name = "$1 - $id";
      my $link = '';
      my $obj;
      if ($class =~ /(Bugzilla::Field)::Choice::(.*)/) {
        $class = $1;
        $name  = $2;
        $obj   = $class->new({name => $name});
      }
      elsif ($class =~ /(Bugzilla::Extension::ExternalBugs::Type)::(.*)/) {
        $class = $1;
        $name  = "ExternalBugs::Type::$2 - $id";
        $obj   = $class->new($id);
      }
      elsif ($class =~ /Bugzilla::Extension::Releases/) {
        $obj = $class->new($id);
      }
      else {
        $obj = $class->new({id => $id});
        $obj = $class->new($id) unless ($obj);
      }
      if ($obj) {
        $link = $obj->admin_link;
      }
      else {
        $link = "$name - DELETED";
      }

      my $who_link = 'unknown';
      if ($row->[0]) {
        unless (defined($users{$row->[0]})) {
          $users{$row->[0]} = Bugzilla::User->check({id => $row->[0]})->admin_link(1);
        }
        $who_link = $users{$row->[0]};
      }

      push(@list, [$who_link, $link, $row->[3], $row->[4], $row->[5], $row->[6]]);
    }
  }

  my $reportable_classes = $dbh->selectcol_arrayref(
    "SELECT DISTINCT class FROM audit_log ORDER BY class");

  $vars->{'action'}             = $action;
  $vars->{'object_ids'}         = join(', ', @obj_ids);
  $vars->{'from'}               = $from;
  $vars->{'to'}                 = $to;
  $vars->{'who'}                = $who;
  $vars->{'list'}               = \@list;
  $vars->{'cur_classes'}        = $cur_classes;
  $vars->{'reportable_classes'} = $reportable_classes;

  return;
}

1;
