# -*- Mode: perl; indent-tabs-mode: nil -*-
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (C) 2012 Jolla Ltd.
# Contact: Pami Ketolainen <pami.ketolainen@jollamobile.com>

package Bugzilla::Extension::AgileTools;
use strict;
use warnings;
use 5.10.1;
use base qw(Bugzilla::Extension);

use Bugzilla;
use Bugzilla::Bug qw(LogActivityEntry);
use Bugzilla::Config qw(SetParam write_params);
use Bugzilla::Constants;
use Bugzilla::Error;
use Bugzilla::Field;
use Bugzilla::Util qw(detaint_natural);
use Bugzilla::WebService::Util qw(filter_wants);

use Bugzilla::Extension::AgileTools::Burn;
use Bugzilla::Extension::AgileTools::Constants;
use Bugzilla::Extension::AgileTools::Pages;
use Bugzilla::Extension::AgileTools::Pages::Team;
use Bugzilla::Extension::AgileTools::Pages::Scrum;
use Bugzilla::Extension::AgileTools::Pool;
use Bugzilla::Extension::AgileTools::Role;
use Bugzilla::Extension::AgileTools::Schema;
use Bugzilla::Extension::AgileTools::Team;
use Bugzilla::Extension::AgileTools::Util;

use JSON;
use List::MoreUtils qw(none);
use Scalar::Util qw(looks_like_number);

our $VERSION = '0.03';

my %template_handlers;

use constant PAGE_HANDLERS => (
  [qr/^agiletools\/([^\/]+)\./,         'Pages'],
  [qr/^agiletools\/scrum\/([^\/.]+)\./, 'Pages::Scrum'],
  [qr/^agiletools\/team\/([^\/.]+)\./,  'Pages::Team'],
);

# Hooks in to other modules.
BEGIN {
  ## RED HAT EXTENSION START 1613297
  *Bugzilla::Component::default_pool     = \&_default_pool;
  *Bugzilla::Component::set_default_pool = \&_set_default_pool;
  *Bugzilla::Component::default_pool_obj = \&_default_pool_obj;
  *Bugzilla::Extension::SubComponents::RHSubComponent::default_pool
    = \&_default_pool;
  *Bugzilla::Extension::SubComponents::RHSubComponent::set_default_pool
    = \&_set_default_pool;
  *Bugzilla::Extension::SubComponents::RHSubComponent::default_pool_obj
    = \&_default_pool_obj;
  ## RED HAT EXTENSION END 1613297
  ## RED HAT EXTENSION START 1613300
  *Bugzilla::Bug::default_pool_obj = \&_bug_default_pool_obj;
  ## RED HAT EXTENSION END 1613300
  ## RED HAT EXTENSION START 1622371
  *Bugzilla::Bug::permitted_teams       = \&_permitted_teams;
  *Bugzilla::Product::agile_team        = \&_agile_team;
  *Bugzilla::Product::set_agile_team    = \&_set_agile_team;
  *Bugzilla::Product::agile_team_obj    = \&_agile_team_obj;
  *Bugzilla::Product::permitted_teams   = \&_permitted_teams;
  *Bugzilla::Component::agile_team      = \&_agile_team;
  *Bugzilla::Component::agile_team_obj  = \&_agile_team_obj;
  *Bugzilla::Component::set_agile_team  = \&_set_agile_team;
  *Bugzilla::Component::permitted_teams = \&_permitted_teams;
  *Bugzilla::Extension::SubComponents::RHSubComponent::agile_team = \&_agile_team;
  *Bugzilla::Extension::SubComponents::RHSubComponent::set_agile_team
    = \&_set_agile_team;
  *Bugzilla::Extension::SubComponents::RHSubComponent::agile_team_obj
    = \&_agile_team_obj;
  *Bugzilla::Extension::SubComponents::RHSubComponent::permitted_teams
    = \&_permitted_teams;
  ## RED HAT EXTENSION END 1622371
  *Bugzilla::active_pools_to_vars = \&_active_pools_to_vars;
  *Bugzilla::User::active_pools   = \&_user_active_pools;
  *Bugzilla::Product::my_team     = \&_my_team;
  *Bugzilla::Component::my_team   = \&_my_team;
  *Bugzilla::Extension::SubComponents::RHSubComponent::my_team = \&_my_team;
  *Bugzilla::Bug::my_team                                      = \&_my_team;

  ## REDHAT EXTENSION BEGIN 1882298
  *Bugzilla::Group::my_team = \&_my_team;
  ## REDHAT EXTENSION END 1882298
}

###################
# Template handlers
###################

# Helper to add a handler for the given template.
sub _add_template_handler {
  my ($name, $sub) = @_;
  push @{$template_handlers{$name} ||= []}, $sub;

  return;
}

_add_template_handler(
  'list/list-burn.html.tmpl',
  sub {
    my ($vars)  = @_;
    my $cgi     = Bugzilla->cgi;
    my @bug_ids = map { $_->{bug_id} } @{$vars->{bugs}};

    my $start = $cgi->param("burn_start") || undef;
    ThrowUserError("invalid_parameter",
      {name => "burn_start", err => "Date format should be YYYY-MM-DD"})
      if (defined $start && !($start =~ /^\d\d\d\d-\d\d-\d\d$/));

    my $end = $cgi->param("burn_end") || undef;
    ThrowUserError("invalid_parameter",
      {name => "burn_end", err => "Date format should be YYYY-MM-DD"})
      if (defined $end && !($end =~ /^\d\d\d\d-\d\d-\d\d$/));

    my $data = get_burndata(\@bug_ids, $start, $end);
    $vars->{burn_type}  = $cgi->param("burn_type") || 'items';
    $vars->{burn_start} = $start;
    $vars->{burn_end}   = $end;
    $vars->{burn_json}  = encode_json($data);

    return;
  }
);

sub _user_active_pools {
  my $user = shift;

  if (!defined $user->{active_pools}) {
    if (user_in_agiletools_group()) {

      my $all_pools = Bugzilla::Extension::AgileTools::Pool->match({is_active => 1});
      my @vis_pools;
      foreach my $pool (@{$all_pools}) {
        my $handle;
        if ($pool->is_sprint) {
          ## REDHAT EXTENSION 1677779 - don't show sprints for SST teams
          next if ($pool->team->name =~ m/^sst_/);
          $handle = Bugzilla::Extension::AgileTools::Sprint->new($pool->id);
        }
        elsif ($pool->is_backlog) {
          $handle = Bugzilla::Extension::AgileTools::Backlog->new($pool->id);
        }

        # If a pool isn't attached to a team then anyone can see it.
        if (!$handle || !$handle->team || $handle->team->user_can_view()) {
          push(@vis_pools, $pool);
        }
      }
      $user->{active_pools} = \@vis_pools;
    }
  }

  return $user->{active_pools};
}

sub _active_pools_to_vars {
  my $vars = shift;
  Bugzilla->login(LOGIN_OPTIONAL);
  my $user = Bugzilla->user;
  $vars->{active_pools} = $user->active_pools;

  return;
}

## REDHAT EXTENSION START 1613309
#_add_template_handler("bug/edit.html.tmpl",           \&_active_pools_to_vars);
#_add_template_handler("list/edit-multiple.html.tmpl", \&_active_pools_to_vars);
## REDHAT EXTENSION END 1613309

_add_template_handler(
  "bug/field-help.none.tmpl",
  sub {
    return unless Bugzilla->params->{agile_use_points};

    # This is done here instead of hook/global/field-descs-end.none.tmpl, so
    # that field names are reflected also in the help strings
    my $vars = shift;
    $vars->{vars}->{field_descs}->{actual_time}    = "Actual Points";
    $vars->{vars}->{field_descs}->{work_time}      = "Points Worked";
    $vars->{vars}->{field_descs}->{remaining_time} = "Points Left";
  }
);


#######################################
# Page and template processing handlers
#######################################

sub page_before_template {
  my ($self, $params) = @_;
  my $page_id = $params->{page_id};
  my $vars    = $params->{'vars'};

  if ($page_id =~ /^agiletools\//) {

    Bugzilla->login(LOGIN_REQUIRED);
    user_in_agiletools_group(1);

    foreach (PAGE_HANDLERS()) {
      my ($rex, $mod) = @$_;
      if ($page_id =~ /$rex/) {
        my $handler = $1;
        $mod = "Bugzilla::Extension::AgileTools::$mod";
        my $sub = $mod->can($handler);
        if ($sub) {
          $sub->($params->{vars});
          last;
        }
      }
    }
  }
## REDHAT EXTENSION START 1613309
  elsif ($page_id =~ m/^agile_team_membership.(html|csv|json)$/) {
    my $type = $1;
    Bugzilla->login(LOGIN_REQUIRED);
    if ($type eq 'csv' || $type eq 'json') {
      my $cgi = Bugzilla->cgi;
      $cgi->set_dated_content_disp('attachment', 'team_membership', $type);
    }
    _agile_team_membership($vars);
  }
  elsif ($page_id =~ m/^agile_component_mapping.(html|csv|json)$/) {
    my $type = $1;
    Bugzilla->login(LOGIN_REQUIRED);
    _agile_component_mapping($vars);
    if (($type eq 'csv' || $type eq 'json') && $vars->{'product'}) {

    $vars->{json_data}
      = JSON->new->utf8->allow_blessed->convert_blessed->encode($vars->{rows})
      if ($type eq 'json');

      my $cgi = Bugzilla->cgi;
      $cgi->set_dated_content_disp('attachment',
        'component_mapping-' . $vars->{'product'}->name, $type);
    }
  }
## REDHAT EXTENSION END 1613309

  return;
}

sub template_before_process {
  my ($self, $params) = @_;
  my $subs = $template_handlers{$params->{file}};
  for my $sub (@{$subs || []}) {
    $sub->($params->{vars});
  }

  return;
}

## REDHAT EXTENSION START 1613309
sub _agile_team_membership {
  my $vars = shift;
  Bugzilla::Extension::AgileTools::Pages::Team::list($vars);

  my @rows;
  foreach my $team (@{$vars->{active_teams}}) {
    my $rls = _team_roles_tidy($team);
    push(@rows, $rls);
  }
  $vars->{rows} = \@rows;

  my $role_names = Bugzilla->dbh->selectcol_arrayref(
    'select distinct name from agile_role order by name');
  $vars->{role_names} = $role_names;

  $vars->{json_data}
    = JSON->new->utf8->allow_blessed->convert_blessed->encode(\@rows);
  return;
}

sub _team_roles_tidy {
  my $team       = shift;
  my $team_roles = $team->roles;

  my %rls = (name => $team->{name}, id => $team->id);
  foreach my $member (@{$team->members}) {
    my $has_roles = 0;
    foreach my $role (@{$team_roles->{$member->id}}) {
      $has_roles++;
      push(@{$rls{$role->{name}}}, $member->{login_name});
    }
    unless ($has_roles) {
      push(@{$rls{'Other'}}, $member->{login_name});
    }
  }
  return (\%rls);
}

sub _agile_component_mapping {
  my $vars = shift;
  _active_pools_to_vars($vars);

  my $dbh   = Bugzilla->dbh;
  my $input = Bugzilla->input_params;
  my $user  = Bugzilla->user;

  # If the user cannot enter bugs in any product, stop here.
  scalar @{$user->get_selectable_products} || ThrowUserError('no_products');

  # Create data structures representing each classification
  my @classifications = ();
  foreach my $c (@{$user->get_selectable_classifications}) {

    # Create hash to hold attributes for each classification.
    my %classification = (
      'name'     => $c->name,
      'products' => [@{$user->get_selectable_products($c->id)}]
    );

    # Assign hash back to classification array.
    push @classifications, \%classification;
  }
  $vars->{'classifications'} = \@classifications;

  my $product_name = $input->{'product'} // '';

  return if !$product_name;

  # We need to check and make sure that the user has permission
  # to see this product.
  my $product = Bugzilla::Product->new({name => $product_name, cache => 1});
  if (!$product || !$user->can_see_product($product->name)) {
    return;
  }

  $product_name = $product->name;

  $vars->{'product'} = $product;

  my $roles
    = $dbh->selectall_arrayref("SELECT id, name FROM agile_role ORDER BY name");

  my $roles_sql  = '';
  my $fields_sql = '';
  my $count      = 1;

  # Add users with no roles
  $roles_sql .= <<EOT;
LEFT JOIN user_group_map ugm ON ugm.group_id = cteam.group_id
 AND ugm.user_id NOT IN (
    SELECT DISTINCT user_id FROM agile_user_role aur WHERE aur.team_id = cteam.id
  )
LEFT JOIN profiles u_other ON ugm.user_id = u_other.userid
EOT

  ## BUGBUG sql_group_concat doesn't work with PG for this
  $fields_sql .= <<EOF;
    ARRAY_TO_STRING(ANYARRAY_UNIQ(ARRAY_AGG(u_other.login_name::text)), ', '::text) as "Other",
EOF

  # Add users with roles
  foreach my $row (@$roles) {
    my ($id, $name) = @$row;

    ## BUGBUG sql_group_concat doesn't work with PG for this
    $fields_sql .= <<EOF;
    ARRAY_TO_STRING(ANYARRAY_UNIQ(ARRAY_AGG(u$count.login_name::text)), ', '::text) as "$name",
EOF

    $roles_sql .= <<EOT;
LEFT JOIN agile_user_role urole$count ON urole$count.team_id = cteam.id AND urole$count.role_id = $id
LEFT JOIN profiles u$count ON urole$count.user_id = u$count.userid
EOT

    $count++;
  }

  chomp($roles_sql);
  chomp($fields_sql);
  chop($fields_sql);

  my $query = <<EOS;
SELECT c.name comp_name, '' sc_name1, '' sc_name2, '' sc_name3,
    cpool.name pool_name, cteam.name team, cteam.id team_num,
    $fields_sql
FROM components c
LEFT JOIN agile_pool cpool ON c.default_pool = cpool.id
LEFT JOIN agile_backlog cbacklog ON cpool.id = cbacklog.pool_id
LEFT JOIN agile_team cteam ON cbacklog.team_id = cteam.id
$roles_sql
WHERE product_id IN (
    SELECT id FROM products WHERE name = '$product_name'
)
GROUP BY comp_name, pool_name, team, team_num, sc_name1, sc_name2, sc_name3
UNION ALL
SELECT c.name comp_name, sc1.name sc_name1, '' sc_name2, '' sc_name3,
    cpool.name pool_name, cteam.name team, cteam.id team_num,
    $fields_sql
FROM components c
RIGHT JOIN rh_sub_components sc1 ON sc1.component_id = c.id
LEFT JOIN agile_pool cpool ON sc1.default_pool = cpool.id
LEFT JOIN agile_backlog cbacklog ON cpool.id = cbacklog.pool_id
LEFT JOIN agile_team cteam ON cbacklog.team_id = cteam.id
$roles_sql
WHERE product_id IN (
    SELECT id FROM products WHERE name = '$product_name'
)
GROUP BY comp_name, pool_name, team, team_num, sc_name1, sc_name2, sc_name3
UNION ALL
SELECT c.name comp_name, sc1.name sc_name1, sc2.name sc_name2, '' sc_name3,
    cpool.name pool_name, cteam.name team, cteam.id team_num,
    $fields_sql
FROM components c
RIGHT JOIN rh_sub_components sc1 ON sc1.component_id = c.id
RIGHT JOIN rh_sub_components sc2 ON sc2.parent_id = sc1.id
LEFT JOIN agile_pool cpool ON sc2.default_pool = cpool.id
LEFT JOIN agile_backlog cbacklog ON cpool.id = cbacklog.pool_id
LEFT JOIN agile_team cteam ON cbacklog.team_id = cteam.id
$roles_sql
WHERE product_id IN (
    SELECT id FROM products WHERE name = '$product_name'
)
GROUP BY comp_name, pool_name, team, team_num, sc_name1, sc_name2, sc_name3
UNION ALL
SELECT c.name comp_name, sc1.name sc_name1, sc2.name sc_name2, sc3.name sc_name3,
    cpool.name pool_name, cteam.name team, cteam.id team_num,
    $fields_sql
FROM components c
RIGHT JOIN rh_sub_components sc1 ON sc1.component_id = c.id
RIGHT JOIN rh_sub_components sc2 ON sc2.parent_id = sc1.id
RIGHT JOIN rh_sub_components sc3 ON sc3.parent_id = sc2.id
LEFT JOIN agile_pool cpool ON sc3.default_pool = cpool.id
LEFT JOIN agile_backlog cbacklog ON cpool.id = cbacklog.pool_id
LEFT JOIN agile_team cteam ON cbacklog.team_id = cteam.id
$roles_sql
WHERE product_id IN (
    SELECT id FROM products WHERE name = '$product_name'
)
GROUP BY comp_name, pool_name, team, team_num, sc_name1, sc_name2, sc_name3
ORDER BY comp_name, sc_name1, sc_name2, sc_name3
EOS

  $vars->{rows} = $dbh->selectall_arrayref($query, {Slice => {}});

  my $role_names = $dbh->selectcol_arrayref(
    'select distinct name from agile_role order by name');

  $vars->{role_names} = $role_names;

  return;
}

sub _get_child_rows {
  my $sc    = shift;
  my $rows  = shift;
  my $names = shift;

  foreach my $subcomp (@{$sc->children}) {
    my %sc_row;
    $sc_row{name} = $names->[0];
    my $depth = $subcomp->level;
    for (my $i = 1; $i <= $depth; $i++) {
      $sc_row{"sc_name$i"} = $names->[$i];
    }
    my $key = $sc_row{"sc_name" . ($depth + 1)} = $subcomp->name;

    if ($subcomp->default_pool) {
      $sc_row{pool_name} = $subcomp->default_pool_obj->name;
      $sc_row{team}      = _team_roles_tidy($subcomp->default_pool_obj->team);
    }
    push(@$rows, \%sc_row);
    if ($subcomp->children) {
      _get_child_rows($subcomp, $rows, [@$names, $subcomp->name]);
    }
  }

  return;
}
## REDHAT EXTENSION END 1613309

#############################
# BayotBase page header links
#############################

sub bb_common_links {
  my ($self, $args) = @_;
  if (0) {    # we don't use the menus
    return unless user_in_agiletools_group();
    $args->{links}->{agile_teams} = [{
      text     => "All teams",
      href     => "page.cgi?id=agiletools/team/list.html",
      priority => 11
    }];
    $args->{links}->{agile_summary} = [{
      text     => "My teams",
      href     => "page.cgi?id=agiletools/team/list.html&user_teams=1",
      priority => 10
    }];
  }

  return;
}

sub bb_group_params {
  my ($self, $args) = @_;
  push(@{$args->{group_params}}, 'agile_user_group', 'agile_nonhuman_group');

  return;
}

############################
# Additional buglist columns
############################

sub buglist_columns {
  my ($self, $args) = @_;
  my $columns = $args->{columns};
  $columns->{"agile_pool.name"}
    = {name => "COALESCE(agile_pool.name,'')", title => "Pool"};
  $columns->{"bug_agile_pool.pool_order"}
    = {name => "COALESCE(bug_agile_pool.pool_order, -1)", title => "Pool order"};
  $columns->{"bug_agile_pool.pool_id"}
    = {name => "COALESCE(bug_agile_pool.pool_id, -1)", title => "Pool ID"};
  $columns->{"agile_team.name"}
    = {name => "COALESCE(agile_team.name,'')", title => "Pool team"};
  return;
}

#################################################
# Table joins required for the additional columns
#################################################

sub buglist_column_joins {
  my ($self, $args) = @_;
  my $joins = $args->{column_joins};
  $joins->{"agile_pool.name"} = {
    table   => "bug_agile_pool",
    as      => "bug_agile_pool",
    then_to => {
      as    => "agile_pool",
      table => "agile_pool",
      from  => "bug_agile_pool.pool_id",
      to    => "id",
    },
  };

  $joins->{"bug_agile_pool.pool_order"}
    = {table => "bug_agile_pool", as => "bug_agile_pool",};

  $joins->{"bug_agile_pool.pool_id"}
    = {table => "bug_agile_pool", as => "bug_agile_pool",};

  $joins->{"agile_team.name"} = {
    table   => "bug_agile_pool",
    as      => "bug_agile_pool",
    then_to => {
      as        => "agile_backlog",
      table     => "agile_backlog",
      from      => "bug_agile_pool.pool_id",
      to        => "pool_id",
      extra_sql => 'LEFT JOIN agile_team on agile_backlog.team_id = agile_team.id',
    },
  };

  return;
}

##########################################
# Additional operations when updating bugs
##########################################

# Helper to determine which bugs are being updated by user
# TODO Move this to BayotBase
sub _bugs_being_updated {
  my $cache = Bugzilla->request_cache;
  if (!defined $cache->{bugs_being_updated}) {
    my $params = Bugzilla->input_params;
    my @ids;

    # Change several
    @ids = map { $_ =~ /^id_([0-9]*)/ } grep(/^id_/, keys %$params);
    unless (@ids) {
      if (defined $params->{ids}) {

        # Webservice
        @ids = @{$params->{ids}} if defined;
      }
      elsif (defined $params->{id}) {

        # Single bug
        @ids = ($params->{id});
      }
    }
    $cache->{bugs_being_updated} = \@ids;
  }
  return @{$cache->{bugs_being_updated}};
}

sub bug_end_of_update {
  my ($self, $args) = @_;

  my ($bug, $old_bug, $changes, $timestamp)
    = @$args{qw(bug old_bug changes timestamp)};
  my $user = Bugzilla->user;
  my $cgi  = Bugzilla->cgi;

  if (my $status_change = $changes->{'bug_status'}) {
    my $old_status = new Bugzilla::Status({name => $status_change->[0]});
    my $new_status = new Bugzilla::Status({name => $status_change->[1]});
    if (!$new_status->is_open && $old_status->is_open) {

      # Bug is being closed
      my $non_human = Bugzilla->params->{agile_nonhuman_group};
      if (
        (
          !Bugzilla->params->{agile_check_time_only_sprint}
          || ($bug->pool && $bug->pool->is_sprint)
        )
        && (!$non_human || !$user->in_group($non_human))
        )
      {
        # Check that actual time is set if it is required for the
        # severity and resolution
        my $check_severity = grep { $bug->bug_severity eq $_ }
          @{Bugzilla->params->{"agile_check_time_severity"}};
        my $check_resolution = grep { $bug->resolution eq $_ }
          @{Bugzilla->params->{"agile_check_time_resolution"}};
        if ($check_severity && $check_resolution) {
          ThrowUserError("agile_actual_time_required") if ($bug->actual_time == 0);
        }
      }

      # Remove closed bug from any backlog
      # RED HAT EXTENSION 1696864 - DIRTY HACK FIX in 1677779
      if ($bug->pool && $bug->pool->is_backlog && $bug->pool->name !~ /^sst_/) {
        $bug->pool->remove_bug($bug);
        $changes->{'bug_agile_pool.pool_id'} = [$old_bug->pool_id, $bug->pool_id];

        # Activity log has been writen at this point so we need to
        # add this entry
        LogActivityEntry($bug->id, 'bug_agile_pool.pool_id', $old_bug->pool_id,
          $bug->pool_id, $user->id, $timestamp);
      }
    }
  }

  return;
}

sub bug_check_can_change_field {
  my ($self, $params) = @_;
  my ($bug, $field, $new_value, $priv_results)
    = @$params{qw(bug field new_value priv_results)};
  if ($field eq 'estimated_time') {

    # Check if user is allowed to edit the original estimates of items in
    # sprint
    if (!user_can_change_estimated_time($bug)) {
      push(@$priv_results, PRIVILEGES_REQUIRED_EMPOWERED);
    }
  }
  if ($field eq 'pool_id') {

    # User needs to be logged in and in agile_user_group to change bug pool
    my $user = Bugzilla->login(LOGIN_OPTIONAL);
    if (!user_in_agiletools_group()) {
      push(@$priv_results, PRIVILEGES_REQUIRED_EMPOWERED);
    }
  }

  return;
}

sub object_before_set {
  my ($self, $args) = @_;
  my ($obj, $field, $value) = @$args{qw(object field value)};
  if ($field eq 'estimated_time' && $obj->isa('Bugzilla::Bug')) {
    if ($obj->estimated_time != $value && !user_can_change_estimated_time($obj)) {
      ThrowUserError('agile_estimated_time_locked');
    }
  }

  return;
}


sub object_end_of_set_all {
  my ($self, $args) = @_;
  my $bug = $args->{object};
  return unless Bugzilla->usage_mode == USAGE_MODE_BROWSER;
  return unless $bug->isa("Bugzilla::Bug");

  # If we are editing bug via browser, we need to manually set pool_id,
  # because it is not included in set_all in process_bug.cgi
  my $cgi        = Bugzilla->cgi;
  my $dontchange = $cgi->param('dontchange') || '';
  my $pool_id    = $cgi->param('pool_id');
  return if (!defined $pool_id || $pool_id eq $dontchange);
  $bug->set_pool_id($pool_id);

  return;
}

sub object_end_of_update {
  my ($self, $args) = @_;
  my ($obj, $old_obj, $changes) = @$args{qw(object old_object changes)};

  if ($obj->isa("Bugzilla::Bug")) {

    # Update remaining time if estimated time is changed
    if (defined $changes->{estimated_time} && !defined $changes->{remaining_time}) {
      my ($old, $new) = @{$changes->{estimated_time}};
      if ($obj->{remaining_time} != $new) {
        Bugzilla->dbh->do(
          "UPDATE bugs " . "SET remaining_time = ? " . "WHERE bug_id = ?",
          undef, $new, $obj->id);
        $changes->{remaining_time} = [$old_obj->{remaining_time}, $new];
      }
    }

    my $reset_pool
      = defined Bugzilla->input_params->{'set_default_pool'}
      ? Bugzilla->input_params->{'set_default_pool'}
      : 0;
    if ($reset_pool && $obj->default_pool_obj) {
      $obj->set_pool_id($obj->default_pool_obj->id);
    }

    # Update pool_id and pool_order if they have been changed
    if ($obj->pool_id != $old_obj->pool_id) {
      if ($obj->pool_id) {
        $obj->pool->add_bug($obj, $obj->{pool_order_set} ? $obj->pool_order : undef);
      }
      elsif ($old_obj->pool_id) {
        $old_obj->pool->remove_bug($obj);
      }

      my $old_name = $old_obj->pool_id ? $old_obj->pool->name : "";
      my $new_name = $obj->pool_id     ? $obj->pool->name     : "";

      $changes->{'bug_agile_pool.pool_id'} = [$old_name, $new_name];
    }
    elsif ($obj->pool_order != $old_obj->pool_order) {
      $obj->pool->add_bug($obj, $obj->pool_order);
    }
  }

  # Update params if group names change
  if ($obj->isa("Bugzilla::Group") && defined $changes->{name}) {
    if ( Bugzilla->params->{agile_user_group}
      && Bugzilla->params->{agile_user_group} eq $old_obj->name)
    {
      SetParam('agile_user_group', $obj->name);
      write_params();
    }
    if ( Bugzilla->params->{agile_nonhuman_group}
      && Bugzilla->params->{agile_nonhuman_group} eq $old_obj->name)
    {
      SetParam('agile_nonhuman_group', $obj->name);
      write_params();
    }
  }

  return;
}

################################################
# Search operators for the additional bug fields
################################################

sub search_operator_field_override {
  my ($self, $args) = @_;
  my $operators = $args->{'operators'};
  my $search    = $args->{'search'};

  $operators->{'agile_pool.name'}->{_default} = sub {
    _add_agile_pool_join($search, @_);
    _search_extra_filters($search, @_);    ## REDHAT EXTENSION 1323078
  };
  $operators->{'bug_agile_pool.pool_order'}->{_default} = sub {
    _add_bug_agile_pool_join($search, @_);
    _search_extra_filters($search, @_);    ## REDHAT EXTENSION 1323078
  };
  $operators->{'bug_agile_pool.pool_id'}->{_default} = sub {
    _add_bug_agile_pool_join($search, @_);
    _search_extra_filters($search, @_);    ## REDHAT EXTENSION 1323078
  };

  $operators->{'agile_team.name'}->{_default} = sub {
    _add_team_agile_pool_join($search, @_);
    _search_extra_filters($search, @_);    ## REDHAT EXTENSION 1323078
  };

  return;
}

## REDHAT EXTENSION START 1323078
# This function prevents users for searching for values they aren;t allowed to see
# e.g. if you can see pool_id field, but can't see value '4', then this will prevent '4' from matching
sub _search_extra_filters {
  my $search = shift;
  my ($invocant, $args) = @_;
  my $user = $search->_user;

  $args->{term} .= " AND ( bug_agile_pool.pool_id IS NULL ";
  my @vis_pools = @{$user->visible_pool_ids()};
  if (@vis_pools) {
    $args->{term}
      .= " OR bug_agile_pool.pool_id IN (" . join(",", @vis_pools) . ")";
  }
  $args->{term} .= " )";

  return;
}
## REDHAT EXTENSION END 1323078

# Table join required for Pool name
sub _add_agile_pool_join {
  my $search = shift;
  my ($invocant, $args) = @_;
  my ($joins) = @$args{qw(joins)};
  if (none { $_->{table} eq 'bug_agile_pool' } @$joins) {
    my $join = {
      table   => "bug_agile_pool",
      as      => "bug_agile_pool",
      then_to => {
        table => "agile_pool",
        as    => "agile_pool",
        from  => "bug_agile_pool.pool_id",
        to    => "id",
      },
    };
    push(@$joins, $join);
  }
  $args->{full_field} = "COALESCE($args->{full_field}, '')";
  $search->_do_operator_function($args);

  return;
}

# Table join required for Pool id and order
sub _add_bug_agile_pool_join {
  my $search = shift;
  my ($invocant, $args) = @_;
  my ($joins) = @$args{qw(joins)};
  if (none { $_->{table} eq 'bug_agile_pool' } @$joins) {
    my $join = {table => "bug_agile_pool", as => "bug_agile_pool",};
    push(@$joins, $join);
  }
  $args->{full_field} = "COALESCE($args->{full_field}, -1)";
  $search->_do_operator_function($args);

  return;
}

sub _add_team_agile_pool_join {
  my $search = shift;

  my ($invocant, $args) = @_;
  my ($joins) = @$args{qw(joins)};
  if (none { $_->{table} eq 'agile_backlog' } @$joins) {
    my $join = {
      table   => "bug_agile_pool",
      as      => "bug_agile_pool",
      then_to => {
        as        => "agile_backlog",
        table     => "agile_backlog",
        from      => "bug_agile_pool.pool_id",
        to        => "pool_id",
        extra_sql => 'LEFT JOIN agile_team on agile_backlog.team_id = agile_team.id',
      },
    };
    push(@$joins, $join);
  }
  $args->{full_field} = "COALESCE($args->{full_field}, '')";
  $search->_do_operator_function($args);

  return;
}

#################
# Database schema
#################

sub db_schema_abstract_schema {
  my ($self, $args) = @_;
  agiletools_schema_init($args->{schema});

  return;
}

##########################################
# Database updates performed in checksetup
##########################################

sub install_update_db {
  my ($self, $args) = @_;
  agiletools_schema_update();
  agiletools_db_init();

  return;
}

###############
# Sanity checks
###############
sub _get_bad_pools {
  my $dbh = Bugzilla->dbh;

  my $pool_ids = $dbh->selectcol_arrayref('SELECT id FROM agile_pool');

  my $sth
    = $dbh->prepare_cached(
        'SELECT pool_order FROM bug_agile_pool WHERE pool_id = ? '
      . 'ORDER BY pool_order ASC');

  my %bad_pools;
  for my $pool_id (@$pool_ids) {
    $sth->execute($pool_id);
    my $expected = 1;
    my @gaps;
    while (my ($real) = $sth->fetchrow_array) {
      if ($real != $expected) {
        push(@gaps, {start => $expected, end => $real});
        $expected = $real + 1;
      }
      else {
        $expected += 1;
      }
    }
    if (@gaps) {
      $bad_pools{$pool_id} = \@gaps;
    }
  }
  return \%bad_pools;
}

sub _sanitycheck_repair_pg {

  my $dbh = Bugzilla->dbh;

  my $fix_dupe = $dbh->prepare_cached('
UPDATE bug_agile_pool bap1
SET pool_order = bap2.order
FROM (
  SELECT bap2.*, row_number() over (ORDER BY bap2.pool_order, bap2.bug_id) AS order
    FROM bug_agile_pool bap2
    WHERE bap2.pool_id = ?
) bap2
WHERE bap1.bug_id= bap2.bug_id
');

  my $bad_pools = _get_bad_pools();
  for my $pool_id (keys %$bad_pools) {
    $fix_dupe->execute($pool_id);
  }

  return;
}

sub _sanitycheck_repair_not_pg {
  my ($self, $args) = @_;

  my $status = $args->{'status'};
  my $dbh = Bugzilla->dbh;

  my $fix_gap
    = $dbh->prepare_cached('UPDATE bug_agile_pool '
      . 'SET pool_order = pool_order - ? '
      . 'WHERE pool_id = ? AND pool_order > ?');
  my $get_dupes = $dbh->prepare_cached(
    'SELECT bug_id FROM bug_agile_pool ' . 'WHERE pool_id = ? AND pool_order = ?');
  my $make_room
    = $dbh->prepare_cached('UPDATE bug_agile_pool '
      . 'SET pool_order = pool_order + ? '
      . 'WHERE pool_id = ? AND pool_order > ?');
  my $fix_dupe
    = $dbh->prepare_cached('UPDATE bug_agile_pool '
      . 'SET pool_order = pool_order + ? '
      . 'WHERE pool_id = ? AND bug_id = ?');

  my $bad_pools = _get_bad_pools();
  for my $pool_id (keys %$bad_pools) {
    for my $gap (reverse @{$bad_pools->{$pool_id}}) {
      my $change = $gap->{end} - $gap->{start};
      if ($change > 0) {
        $fix_gap->execute($change, $pool_id, $gap->{start});
      }
      elsif ($change == -1) {

        # Duplicate values
        $get_dupes->execute($pool_id, $gap->{end});
        my @dupes = map { $_->[0] } @{$get_dupes->fetchall_arrayref};
        shift @dupes;
        $make_room->execute(scalar @dupes, $pool_id, $gap->{end});
        $change = 1;
        for my $bug_id (@dupes) {
          $fix_dupe->execute($change, $pool_id, $bug_id);
        }
      }
      else {
        $status->(
          'agiletools_repair_pool_order_weird_alert',
          {pool => $pool_id, gap => $gap}, 'alert'
        );
      }
    }
  }

  return;
}

sub sanitycheck_repair {
  my ($self, $args) = @_;

  my $cgi    = Bugzilla->cgi;
  my $status = $args->{'status'};
  if ($cgi->param('agiletools_repair_pool_order')) {
    $status->('agiletools_repair_pool_order_start');

    if (Bugzilla->is_pg) {
      _sanitycheck_repair_pg();
    }
    else {
      _sanitycheck_repair_not_pg($self, $args);
    }

    $status->('agiletools_repair_pool_order_end');
  }

  return;
}

sub sanitycheck_check {
  my ($self, $args) = @_;
  my $status = $args->{'status'};

  $status->('agiletools_check_pool_order');
  my $bad_pools = _get_bad_pools();
  if (%$bad_pools) {
    $status->('agiletools_check_pool_order_alert', {pools => $bad_pools}, 'alert');
    $status->('agiletools_chek_pool_order_prompt');
  }

  return;
}

#####################
# Webservice bindings
#####################

sub webservice {
  my ($self, $args) = @_;
  $args->{dispatch}->{'Agile.Team'}
    = "Bugzilla::Extension::AgileTools::WebService::Team";
  $args->{dispatch}->{'Agile.Sprint'}
    = "Bugzilla::Extension::AgileTools::WebService::Sprint";
  $args->{dispatch}->{'Agile.Pool'}
    = "Bugzilla::Extension::AgileTools::WebService::Pool";
  $args->{dispatch}->{'Agile.Backlog'}
    = "Bugzilla::Extension::AgileTools::WebService::Backlog";

  #RHBZ 1717852 XMLRPC converts the extra '.' to a '/'
  $args->{dispatch}->{'Agile/Team'}
    = "Bugzilla::Extension::AgileTools::WebService::Team";
  $args->{dispatch}->{'Agile/Sprint'}
    = "Bugzilla::Extension::AgileTools::WebService::Sprint";
  $args->{dispatch}->{'Agile/Pool'}
    = "Bugzilla::Extension::AgileTools::WebService::Pool";
  $args->{dispatch}->{'Agile/Backlog'}
    = "Bugzilla::Extension::AgileTools::WebService::Backlog";

  return;
}

########################
# Admin interface panels
########################

sub config_add_panels {
  my ($self, $args) = @_;
  my $modules = $args->{panel_modules};
  $modules->{AgileTools} = "Bugzilla::Extension::AgileTools::Params";

  return;
}

########################
## REDHAT EXTENSION START 1323078
# This function blanks out values in a field the user can't see
# e.g. if a user can see a bug but not the pool on the bug, then this blanks the pool
# because even though the user can't see the pool they can still see the bug
########################

sub user_see_field_values {
  my ($self, $args) = @_;
  my $user = $args->{user};

  my @filtered_vals = ();

  if ($args->{field} eq 'bug_agile_pool.pool_id') {
    my $foo = $args->{values};
    @filtered_vals = @{$user->visible_pool_ids()};
    @$foo          = @filtered_vals;
  }

  return;
}

# Hook to identify which fields need to be filtered
sub bug_filter_fields {
  my ($self, $args) = @_;

  my $fields = $args->{fields};
  my @filtered_fields;
  foreach my $field (@$fields) {

# bug_agile_pool.pool_id
# bug_agile_pool.pool_order
# agile_pool.name
    if ($field->name =~ m/agile_pool/) {

      # Some fields (like partner, are restricted on what you can see)
      my $values = Bugzilla->user->see_field_values($field);

      # not defined = no restrictions, scalar = some values
      if (not defined $values or scalar(@$values)) {
        push @filtered_fields, $field;
      }
    }
    else {
      push(@filtered_fields, $field);
    }
  }
  @$fields = @filtered_fields;

  return;
}
## REDHAT EXTENSION END 1323078

## RED HAT EXTENSION START 1613297
sub _default_pool { return $_[0]->{'default_pool'}; }

sub _set_default_pool {
  $_[0]->set('default_pool', $_[1]);
  delete $_[0]->{default_pool_obj};
  return;
}
_add_template_handler("admin/components/edit.html.tmpl",
  \&_active_pools_to_vars);
_add_template_handler("pages/subcomponents/edit.html.tmpl",
  \&_active_pools_to_vars);

sub _default_pool_obj {
  my ($self, $args) = @_;
  if ( (not defined $self->{default_pool_obj})
    && defined $self->default_pool
    && $self->default_pool ne '')
  {
    $self->{default_pool_obj}
      = Bugzilla::Extension::AgileTools::Pool->new($self->default_pool);
  }

  return ($self->{default_pool_obj});
}

sub object_columns {
  my ($self,  $args)    = @_;
  my ($class, $columns) = @$args{qw(class columns)};

  if ( $class->isa('Bugzilla::Component')
    || $class->isa('Bugzilla::Extension::SubComponents::RHSubComponent'))
  {
    push(@$columns, 'default_pool');
  }

  ## RED HAT EXTENSION START 1622371
  if ( $class->isa('Bugzilla::Product')
    || $class->isa('Bugzilla::Component')
    || $class->isa('Bugzilla::Extension::SubComponents::RHSubComponent'))
  {
    push(@$columns, 'agile_team');
  }
  ## RED HAT EXTENSION END 1622371

  return;
}

sub object_update_columns {
  my ($self,   $args)    = @_;
  my ($object, $columns) = @$args{qw(object columns)};

  if ( $object->isa('Bugzilla::Component')
    || $object->isa('Bugzilla::Extension::SubComponents::RHSubComponent'))
  {
    push(@$columns, 'default_pool');
  }

  ## RED HAT EXTENSION START 1622371
  if ( $object->isa('Bugzilla::Product')
    || $object->isa('Bugzilla::Component')
    || $object->isa('Bugzilla::Extension::SubComponents::RHSubComponent'))
  {
    push(@$columns, 'agile_team');
  }
  ## RED HAT EXTENSION END 1622371

  return;
}

sub object_validators {
  my ($self,  $args)       = @_;
  my ($class, $validators) = @$args{qw(class validators)};

  if ( $class->isa('Bugzilla::Component')
    || $class->isa('Bugzilla::Extension::SubComponents::RHSubComponent'))
  {
    $validators->{default_pool} = \&_check_pool;
  }

  ## RED HAT EXTENSION START 1622371
  if ( $class->isa('Bugzilla::Product')
    || $class->isa('Bugzilla::Component')
    || $class->isa('Bugzilla::Extension::SubComponents::RHSubComponent'))
  {
    $validators->{agile_team} = \&_check_team_id;
  }
  ## RED HAT EXTENSION END 1622371

  return;
}

sub _check_pool {
  my ($invocant, $pool_id) = @_;

  return unless ($pool_id);

  my $pool;
  if (looks_like_number($pool_id)) {
    eval { $pool = Bugzilla::Extension::AgileTools::Pool->new($pool_id); };
  }
  else {
    eval { $pool = Bugzilla::Extension::AgileTools::Pool->new({name => $pool_id}); };
  }

  # Check that the pool exists
  $pool || ThrowUserError('agile_invalid_pool', {id => $pool_id});

  return $pool->id;
}

sub bug_end_of_create {
  my ($self, $args) = @_;

  my $bug = $args->{'bug'};

  ## RED HAT EXTENSION START 1613300
  if (!$bug->pool && $bug->default_pool_obj) {

    # Do as site user to ensure ACLs.
    my $curr_user = Bugzilla->user;
    Bugzilla->set_user(Bugzilla::User->check(Bugzilla->params->{system_user}));
    $bug->set_pool_id($bug->default_pool_obj->id);
    $bug->update();
    Bugzilla->set_user($curr_user);
  }
  ## RED HAT EXTENSION END 1613300

  return;
}

## RED HAT EXTENSION END 1613297

## RED HAT EXTENSION START 1613300
sub _bug_default_pool_obj {
  my ($self, $args) = @_;

  if (not defined $self->{default_pool_obj}) {
    my $def_pool = $self->component_obj->default_pool_obj;
    my $sub_comp = $self->rh_sub_component_objs->[0];

    if ($sub_comp && $sub_comp->default_pool) {
      $def_pool = $sub_comp->default_pool_obj;
    }

    $self->{default_pool_obj} = $def_pool;    # if($def_pool);
  }

  return ($self->{default_pool_obj});
}
## RED HAT EXTENSION END 1613300

## RED HAT EXTENSION START 1622371
sub _agile_team { return $_[0]->{'agile_team'}; }

sub _set_agile_team {
  $_[0]->set('agile_team', $_[1]);
  $_[0]->{agile_team_obj} = undef;
  return;
}

sub _agile_team_obj {
  my ($self, $args) = @_;
  if ((not defined $self->{agile_team_obj}) && defined $self->agile_team) {
    my $team = Bugzilla::Extension::AgileTools::Team->new($self->agile_team);

    $self->{agile_team_obj} = $team;
  }

  return ($self->{agile_team_obj});
}

sub _check_team_id {
  my ($invocant, $team_id) = @_;
  return unless ($team_id);

  my $team;
  if (looks_like_number($team_id)) {
    eval { $team = Bugzilla::Extension::AgileTools::Team->new($team_id); };
  }
  else {
    eval { $team = Bugzilla::Extension::AgileTools::Team->new({name => $team_id}); };
  }

  $team || ThrowUserError('agile_invalid_team', {id => $team_id});

  return ($team->id);
}

sub _permitted_teams {
  my ($invocant) = shift;

  my @teams;

  if ($invocant->isa('Bugzilla::Bug')) {
    foreach my $subcomp (@{$invocant->rh_sub_component_objs}) {
      _permitted_teams_obj($subcomp, \@teams);
    }

    if ($invocant->component_obj) {
      _permitted_teams_obj($invocant->component_obj, \@teams);
    }

    if ($invocant->product_obj) {
      _permitted_teams_obj($invocant->product_obj, \@teams);
    }
  }

  if ($invocant->isa('Bugzilla::Component')) {
    _permitted_teams_obj($invocant,          \@teams);
    _permitted_teams_obj($invocant->product, \@teams);
  }

  if ($invocant->isa('Bugzilla::Product')) {
    _permitted_teams_obj($invocant, \@teams);
  }

  if ($invocant->isa('Bugzilla::Extension::SubComponents::RHSubComponent')) {
    _permitted_teams_obj($invocant,                     \@teams);
    _permitted_teams_obj($invocant->component,          \@teams);
    _permitted_teams_obj($invocant->component->product, \@teams);
  }

  return (\@teams);
}

sub _permitted_teams_obj {
  my ($obj, $teams) = @_;

  if ($obj && $obj->agile_team_obj) {
    push(@$teams, $obj->agile_team_obj->id);
  }

  return;
}

sub _my_team {
  my $invocant = shift;

  unless ($invocant->{my_team}) {

    if ($invocant->isa('Bugzilla::Group')) {
      my $teams
        = Bugzilla::Extension::AgileTools::Team->match({group_id => $invocant->id});
      if (@$teams) {
        $invocant->{my_team} = $teams->[0];
      }
    }
    else {
      my $teams = $invocant->permitted_teams();

      if (@$teams) {
        $invocant->{my_team} = Bugzilla::Extension::AgileTools::Team->new(@$teams);
      }
    }
  }

  return $invocant->{my_team};
}

## RED HAT EXTENSION END 1622371

## RED HAT EXTENSION START 1677378
sub quicksearch_map {
  my ($self, $args) = @_;
  my $map = $args->{'map'};

  $map->{'pool'} = 'agile_pool.name';
  $map->{'team'} = 'agile_team.name';

  return;
}
## RED HAT EXTENSION END 1677378

sub bugs_to_hashes {
  my ($self, $args) = @_;

  my $params = $args->{params};
  my $user   = Bugzilla->user;

  my $orig_extra_fields = $params->{'allowed_extra_fields'};

  $params->{'allowed_extra_fields'} = [qw(
    pool
    bug_agile_pool.pool_order
    pool_order
    agile_pool.name
    agile_team.name
    bug_agile_pool.pool_id
    pool_id
  )];

  for (my $i = 0; $i < scalar @{$args->{bugs}}; $i++) {
    my $bug  = $args->{bugs}->[$i];
    my $item = $args->{hashes}->[$i];

    ## REDHAT EXTENSION START 1613309
    if ( (filter_wants $params, 'pool', 'extra')
      || (filter_wants $params, 'agile_pool.name', 'extra'))
    {
      if ($bug->pool) {
        $bug->pool->team;    # have to force team to load from DB
        $item->{pool} = $bug->pool;
      }
      else {
        $item->{pool} = {};
      }
    }
    ## REDHAT EXTENSION END 1613309

    ## REDHAT EXTENSION START 1347097
    if ( (filter_wants $params, 'bug_agile_pool.pool_order', 'extra')
      || (filter_wants $params, 'pool_order', 'extra'))
    {
      $item->{'pool_order'} = $bug->pool_order;
    }

    if ( (filter_wants $params, 'bug_agile_pool.pool_id', 'extra')
      || (filter_wants $params, 'pool_id', 'extra'))
    {
      $item->{'bug_agile_pool'}->{'pool_id'} = $bug->pool_id;
    }

    if (filter_wants $params, 'agile_team.name', 'extra') {
      if ($bug->pool) {
        $item->{'agile_team'} = $self->_team_to_hash($bug->pool->team);
      }
    }

  }

  $params->{'allowed_extra_fields'} = $orig_extra_fields;

  return;
}

## REDHAT EXTENSION START 1347097
sub _team_to_hash {
  my ($self, $team) = @_;

  my $item;

  $item = {
    id   => $team->id,
    name => $team->name,
    }
    if ($team);

  return $item;
}
## REDHAT EXTENSION END 1347097

__PACKAGE__->NAME;


__END__

=head1 NAME

Bugzilla::Extension::AgileTools

=head1 DESCRIPTION

This extension allows control of Teams, Sprints & Backlogs.

=head2 Teams

The "My Teams" menu will take you to a page that shows the teams you are a member of. The "All Teams" menu item will take you to a list of all teams you can view or edit.

The list shows the meta data for a team, if you can see the "planning" link on a team then you have edit privileges on that team; allowing you to manage the team and it's backlog & sprints.

To edit a team click on the teams name at the top of the team box, this will open the team page.

To edit team details click the "Show/Hide Editing" button. This will show edit widgets for fields that can be edited.

You will have edit access to a team if you have the product owner, scrum master, product manager, or team lead role in the team, or if you are in the edit group set on the team.

Each Agile team has it's own matching Bugzilla Group, changing membership of the team is changing membership of the group.

By default the teams are very open, as is the way for Open Source, if you want to restrict edit access you need to select groups to limit edit access to. Often the team itself is a suitable option. View access can be restricted in a similar way.

When setting up a group you will probably want to limit which bugs are available in the teams backlog, usually by product and/or component. To do this:

=over

=item 1.
click "show/hide editing" to get the edit widgets to appear

=item 2.
click the pencil icon next to "Unprioritized items" this will bring up a window with the advanced search page in it

=item 3.
Adjust the search to fit your requirements

=item 4.
I<IMPORTANT> run the search!

=item 5.
if you are unhappy, click edit search and go to 3

=item 6.
click close in bottom right corner, this will store your search

=item 6.
click save button

=back

B<Note> if you create a team via the RPC interface, then you can supply a product to be used in to the "Unprioritized items". Alternatively you can supply a complete advanced search string. See the create function in L<Bugzilla::Extension::AgileTools::WebService::Team>.

=head2 Sprints

To edit the sprint for a team, click the planning link.

The planning page has three columns, but only two will be displayed at any time, click the left/right arrows to move the columns left and right. Initially you will see the current sprint and backlog
columns, and the "Unprioritized items" column will not be shown. Click right to display the backlog and "Unprioritized items" columns.

If there are a large number of bugs it will take a while to load, there is a spinner in each column to let you know it's fetching data.

Once the bugs are loaded you will be able to drag and drop them from one column to another, or to reorder the columns. Higher up the list is supposed to represent more important, assuming you are being all agile.

=over

=item -
Drag some bugs in to the backlog.

=item -
The column changes are committed in real time via ajax, so there is not save button for these changes.

=item -
Click left to display the current sprint and backlog columns.

=item -
Drag some bugs in to the sprint.

=item -
You can click the edit button on the sprint to change the dates and capacity for the sprint. Remember to click commit if you make changes and want to save them.

=back

The bugs have 3 icons:

=over

=item 1.
The up/down arrow shows/hides extra bug details

=item 2.
The pencil allows editing basic fields

=item 3.
The plus allows creation of a new bug as a dependency

=back

If you have dependent bugs in the same column they will be shown nested inside the bug, or bugs, they depend on. You can drag a bug from inside the nested view in to a higher column, but it will drag any bugs inside it in to the same column.

e.g. If you have bug A depending on B, and B depending on C, and you drag B from unprioritized to the backlog, then C will come along as B cannot be completed without C.

=cut

