# -*- Mode: perl; indent-tabs-mode: nil -*-
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (C) 2012 Jolla Ltd.
# Contact: Pami Ketolainen <pami.ketolainen@jollamobile.com>

package Bugzilla::Extension::BugViewPlus;
use strict;
use warnings;
use 5.10.1;
use base qw(Bugzilla::Extension);

use Bugzilla::Config qw(SetParam write_params);
use Bugzilla::Error;
use Bugzilla::Field qw(get_legal_field_values);
use Bugzilla::Template;
use Bugzilla::Util qw(detaint_natural trick_taint);
use Bugzilla::User::Setting qw(add_setting);

use Bugzilla::Extension::BugViewPlus::Template;
use List::MoreUtils qw(any);
use Scalar::Util qw(looks_like_number);

use JSON;

our $VERSION = '0.01';

BEGIN {
  ## REDHAT EXTENSION START 1326981
  *Bugzilla::Component::set_template_id = \&_set_template_id;
  *Bugzilla::Component::template_id     = \&_template_id;
  *Bugzilla::Component::template_name   = \&_template_name;
  *Bugzilla::default_template_id        = \&_default_template_id;
  ## REDHAT EXTENSION END 1326981

  ## REDHAT EXTENSION START 2004321
  *Bugzilla::User::advanced_fields     = \&_user_advanced_fields;
  *Bugzilla::User::advanced_fields_raw = \&_user_advanced_fields_raw;
  *Bugzilla::no_display_fields         = \&_no_display_fields;
  ## REDHAT EXTENSION END 2004321
}

sub install_update_db {
  my ($self, $args) = @_;

  my $old_group
    = Bugzilla->dbh->selectrow_hashref(
    "SELECT name, isactive, isbuggroup FROM groups " . "WHERE name = ?",
    undef, 'bvp_edit_description');
  if (defined $old_group && !$old_group->{isbuggroup} && $old_group->{isactive}) {

    # Fix earlier installations group type
    Bugzilla->dbh->do("UPDATE groups SET isbuggroup = 1, isactive = 0 WHERE id = ?",
      undef, $old_group->id);
  }

  my $dbh = Bugzilla->dbh;
  if (!$dbh->bz_column_info('components', 'bvp_template_id')) {
    my $txt
      = "Description of problem:\n\n\nVersion-Release number of selected component (if applicable):\n\n\nHow reproducible:\n\n\nSteps to Reproduce:\n1.\n2.\n3.\n\nActual results:\n\n\nExpected results:\n\n\nAdditional info:\n";
    $dbh->do(
      q{INSERT INTO bvp_templates (name, is_active, description, content) VALUES (?,1,'',?)},
      undef, 'Default', $txt
    );

    my $default = $dbh->selectrow_array(
      "SELECT id FROM bvp_templates where name = 'Default'");

    $dbh->bz_add_column('components', 'bvp_template_id',
      {TYPE => 'INT4', DEFAULT => $default, NOTNULL => 1});
    $dbh->bz_add_fk("components", "bvp_template_id",
      {TABLE => 'bvp_templates', COLUMN => 'id'});
  }

  ## REDHAT EXTENSION START 2004321
  $dbh->bz_add_column('profiles', 'advanced_fields', {TYPE => 'LONGTEXT'});
  ## REDHAT EXTENSION END 2004321

  return;
}

sub install_before_final_checks {
  my $dbh = Bugzilla->dbh;

  my $count
    = $dbh->selectrow_array(
    q{SELECT count(*) FROM setting WHERE name = 'bvp_hide_advanced' AND default_value = 'rememberhide'}
    );

  if ($count == 0) {
    add_setting('bvp_hide_advanced', ['remember', 'rememberhide', 'on', 'off'],
      'rememberhide', undef, 1);
  }

  return;
}

sub config_add_panels {
  my ($self, $args) = @_;
  my $modules = $args->{panel_modules};
  $modules->{BugViewPlus} = "Bugzilla::Extension::BugViewPlus::Params";

  return;
}

sub bb_group_params {
  my ($self, $args) = @_;
  push(@{$args->{group_params}}, 'bvp_description_edit_group');

  return;
}

sub bug_end_of_update {
  my ($self, $args) = @_;
  my $cgi = Bugzilla->cgi;
  my $dbh = Bugzilla->dbh;
  my ($bug, $timestamp) = @$args{qw(bug timestamp)};

  # Get the bug id passed to processbug.cgi
  my $bug_id = $cgi->param('id');
  return unless $bug_id;
  return unless detaint_natural($bug_id);

  # Make sure we are udating the same bug and not some related bug
  if ($bug->bug_id == $bug_id) {

    # Edit description related stuff
    my $group         = Bugzilla->params->{'bvp_description_edit_group'};
    my $type_editable = grep { $_ eq $bug->bug_severity }
      @{Bugzilla->params->{bvp_description_editable_types}};
    if ($type_editable && Bugzilla->user->in_group($group)) {

      # Current description text
      my $old_desc = ${@{$bug->comments}[0]}{'thetext'};

      # Current descriptions comment id (in longdescs table)
      my $comment_id = ${@{$bug->comments}[0]}{'comment_id'};

      # The new description
      my $new_desc = $cgi->param('bvp_description');

      if ($new_desc) {
        trick_taint($new_desc);

        # Remove trailing whitespace
        $new_desc =~ s/\s*$//s;
        $new_desc =~ s/\r\n?/\n/g;    # Get rid of \r.

        if ($old_desc ne $new_desc) {

          # Add or update the timestamp in description
          my $stamp = "> Edited by " . Bugzilla->user->name . " on " . $timestamp;
          if (!($new_desc =~ s/> Edited by .* on .*$/$stamp/)) {
            $new_desc .= "\n\n" . $stamp;
          }

          # Push to database
          $dbh->do(
            'UPDATE longdescs SET thetext = ? ' . 'WHERE bug_id = ? AND comment_id = ?',
            undef, $new_desc, $bug->bug_id, $comment_id);
        }
      }
    }
  }

  return;
}

sub bug_format_comment {
  my ($self, $args) = @_;
  return unless Bugzilla->params->{bvp_linkify_severity};

  my $regexes = $args->{regexes};

  # Turn '<severity> #' into bug link in comments
  if (!defined $self->{bug_severities}) {
    $self->{bug_severities} = get_legal_field_values('bug_severity');
  }
  foreach my $value (@{$self->{bug_severities}}) {
    push(@$regexes,
      {match => qr/($value)\s*(\d+)/i, replace => \&_replace_bug_link});
  }

  return;
}

sub db_schema_abstract_schema {
  my ($self, $args) = @_;
  my $schema = $args->{schema};

  # Table for storing the bug templates
  $schema->{bvp_templates} = {
    FIELDS => [
      id          => {TYPE => 'MEDIUMSERIAL', NOTNULL => 1, PRIMARYKEY => 1},
      name        => {TYPE => 'TINYTEXT',     NOTNULL => 1},
      is_active   => {TYPE => 'BOOLEAN',      NOTNULL => 1, DEFAULT => 0},
      description => {TYPE => 'TINYTEXT'},
      content     => {TYPE => 'MEDIUMTEXT'},
    ],
    INDEXES => [],
  };

  return;
}

sub object_end_of_update {
  my ($self, $args) = @_;
  my ($new_obj, $old_obj, $changes) = @$args{qw(object old_object changes)};

  # Update user group param if group name changes
  if ($new_obj->isa("Bugzilla::Group") && defined $changes->{name}) {
    if ($old_obj->name eq Bugzilla->params->{bvp_description_edit_group}) {
      SetParam('bvp_description_edit_group', $new_obj->name);
      write_params();
    }
  }

  return;
}

sub page_before_template {
  my ($self, $args) = @_;
  return unless ($args->{page_id} eq 'bvp_template.html');
  ThrowUserError('auth_failure',
    {group => 'admin', action => 'access', object => 'administrative_pages'})
    unless Bugzilla->user->in_group('admin');
  my $vars   = $args->{vars};
  my $cgi    = Bugzilla->cgi;
  my $tid    = $cgi->param('tid');
  my $action = $cgi->param('action') || '';
  my $current;

  if (defined $tid) {
    $current = Bugzilla::Extension::BugViewPlus::Template->check({id => $tid});
  }
  if ($action) {
    my $values = {
      name        => scalar $cgi->param('name'),
      is_active   => scalar $cgi->param('is_active'),
      description => scalar $cgi->param('description'),
      content     => scalar $cgi->param('content'),
    };
    if ($action eq 'create') {
      $current = Bugzilla::Extension::BugViewPlus::Template->create($values);
      $vars->{message} = 'bvp_template_created';
    }
    elsif ($action eq 'save') {
      ThrowCodeError('param_required', {param => 'tid', function => 'save'})
        unless defined $current;
      $current->set_all($values);
      $current->update();
      $vars->{message} = 'bvp_template_saved';
    }
    elsif ($action eq 'remove') {
      ThrowCodeError('param_required', {param => 'tid', function => 'remove'})
        unless defined $current;
      $current->remove_from_db();
      $vars->{message} = 'bvp_template_removed';
      $vars->{name}    = $current->name;
      $current         = undef;
    }
    else {
      ThrowCodeError('param_invalid', {param => $action, function => 'action'});
    }
  }
  $vars->{current}   = $current;
  $vars->{templates} = [Bugzilla::Extension::BugViewPlus::Template->get_all()];

  return;
}

sub template_before_process {
  my ($self, $args) = @_;
  my $vars = $args->{vars};

  if ($args->{file} eq 'bug/comments.html.tmpl') {
    return if (defined($args->{vars}->{bug}->{'error'}));

    my $group         = Bugzilla->params->{'bvp_description_edit_group'};
    my $severity      = $args->{vars}->{bug}->bug_severity;
    my $type_editable = grep { $_ eq $severity }
      @{Bugzilla->params->{bvp_description_editable_types}};
    $args->{vars}->{can_edit_description}
      = $type_editable && Bugzilla->user->in_group($group) ? 1 : 0;
  }
  elsif ($args->{file} eq 'bug/create/create.html.tmpl'
    || $args->{file} eq 'pages/quickideas/enter.html.tmpl'
    || $args->{file} =~ m{bug/create/create-guided-.*\.html\.tmpl}
  ){
    $args->{vars}->{bvp_templates_json}
      = JSON->new->utf8->allow_blessed->convert_blessed->encode(
      Bugzilla::Extension::BugViewPlus::Template->match({is_active => 1}));
  }
  elsif (
    any { $args->{file} eq $_ }
    (qw(
      admin/components/edit.html.tmpl
      bug/create/create-defaultcontent.html.tmpl
      admin/components/create.html.tmpl
      admin/products/create.html.tmpl
    ))
    )
  {
    $vars->{templates} = [Bugzilla::Extension::BugViewPlus::Template->get_all()];
  }

  return;
}

sub _replace_bug_link {
  my $args = shift;
  my $type = $args->{matches}->[0];
  my $id   = $args->{matches}->[1];

  return Bugzilla::Template::get_bug_link($id, $type . " " . $id);
}

## REDHAT EXTENSION START 1326981 per-component default template
sub object_columns {
  my ($self,  $args)    = @_;
  my ($class, $columns) = @$args{qw(class columns)};

  if ($class->isa('Bugzilla::Component')) {
    push(@$columns, 'bvp_template_id');
  }

  return;
}

sub object_update_columns {
  my ($self,   $args)    = @_;
  my ($object, $columns) = @$args{qw(object columns)};

  if ($object->isa('Bugzilla::Component')) {
    push(@$columns, 'bvp_template_id');
  }
  elsif ($object->isa('Bugzilla::User')) {
    ## REDHAT EXTENSION START 2004321
    push(@$columns, 'advanced_fields');
    ## REDHAT EXTENSION END 2004321
  }

  return;
}

sub object_validators {
  my ($self,  $args)       = @_;
  my ($class, $validators) = @$args{qw(class validators)};

  if ($class->isa('Bugzilla::Component')) {
    $validators->{bvp_template_id} = \&_check_template_id;
  }
  elsif ($class->isa('Bugzilla::User')) {
    ## REDHAT EXTENSION START 2004321
    $validators->{'advanced_fields'} = sub {
      my ($self, $value) = @_;
      return $value;
    };
    ## REDHAT EXTENSION END 2004321
  }

  return;
}


sub _check_template_id {
  my ($invocant, $template_id) = @_;

  my $templ;

  if (looks_like_number($template_id)) {
    $templ = Bugzilla::Extension::BugViewPlus::Template->new($template_id);
  }
  else {
    $templ
      = Bugzilla::Extension::BugViewPlus::Template->new({name => $template_id});
  }

  # Check that the template exists
  $templ || ThrowUserError('component_invalid_template');

  return $templ->id;
}

sub _set_template_id { $_[0]->set('bvp_template_id', $_[1]); return; }
sub _template_id { return $_[0]->{'bvp_template_id'}; }

sub _template_name {
  my $templ
    = Bugzilla::Extension::BugViewPlus::Template->new($_[0]->{'bvp_template_id'});
  return $templ->name;
}

sub _default_template_id {
  my $templ
    = Bugzilla::Extension::BugViewPlus::Template->new({name => 'Default'});
  return $templ->id;
}

## REDHAT EXTENSION END 1326981

## REDHAT EXTENSION START 2004321
sub _user_advanced_fields {
  my $self = shift;

  if (!defined($self->{advanced_fields})) {
    my $advanced_fields
      = Bugzilla->dbh->selectrow_array(
      "SELECT advanced_fields FROM profiles WHERE userid = ?",
      undef, $self->id);

    if (!$advanced_fields || $advanced_fields eq '') {
      $advanced_fields = Bugzilla->params->{'bvp_advanced_fields'};
    }

    my @advanced_fields = split(/[, ]+/, $advanced_fields);
    $self->{advanced_fields} = \@advanced_fields;
  }

  return $self->{advanced_fields};
}

sub _user_advanced_fields_raw {
  my $self = shift;

  if (!defined($self->{advanced_fields_raw})) {
    my $advanced_fields
      = Bugzilla->dbh->selectrow_array(
      "SELECT advanced_fields FROM profiles WHERE userid = ?",
      undef, $self->id);

    if (!$advanced_fields || $advanced_fields eq '') {
      $advanced_fields = Bugzilla->params->{'bvp_advanced_fields'};
    }

    $self->{advanced_fields_raw} = $advanced_fields;
  }

  return $self->{advanced_fields_raw};
}

sub _no_display_fields {

  if (!defined(Bugzilla->request_cache->{no_display_fields})) {
    my @fields = split(/[, ]+/, Bugzilla->params->{'bvp_no_display_fields'});
    Bugzilla->request_cache->{no_display_fields} = \@fields;
  }

  return Bugzilla->request_cache->{no_display_fields};
}

sub user_preferences {
  my ($self, $args) = @_;
  my $tab     = $args->{'current_tab'};
  my $save    = $args->{'save_changes'};
  my $handled = $args->{'handled'};
  my $vars    = $args->{'vars'};
  my $params  = Bugzilla->input_params;

  return unless $tab eq 'advanced_fields';

  # Create a new user object so we don't mess with the main one, as we
  # don't know where it's been...
  my $user = new Bugzilla::User($vars->{'change_user'}->id);

  if ($save) {
    delete $self->{advanced_fields};

    my $advanced_fields = join(', ',
      map { $_ =~ m/^squelch_(.*)$/; $1 } grep {/^squelch_/} keys %$params);

    $user->set('advanced_fields', $advanced_fields);
    $user->update();
  }

  # Set the 'handled' scalar reference to true so that the caller
  # knows the panel name is valid and that an extension took care of it.
  $$handled = 1;

  return;
}
## REDHAT EXTENSION END 2004321

__PACKAGE__->NAME;
