# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

use strict;
use warnings;
use 5.10.1;

package Bugzilla::Extension::DependentProducts::Record;

use Scalar::Util qw(blessed);

use Bugzilla::Constants;
use Bugzilla::Field;
use Bugzilla::Util;
use Bugzilla::Error;

use base qw(Bugzilla::Object);

use Bugzilla::Extension::DependentProducts::Util;
use Bugzilla::Extension::DependentProducts::Contact;

###############################
####    Initialization     ####
###############################

use constant DB_TABLE   => 'dependent_products';
use constant DB_COLUMNS => qw(
  id
  product_id
  depends_on
  vis_group
);

use constant UPDATE_COLUMNS => qw(
  product_id
  depends_on
  vis_group
);

use constant VALIDATORS => {
  product_id => \&_check_product_id,
  depends_on => \&_check_depends_on,
  vis_group  => \&_check_vis_group,
};

###############################
####      Validators       ####
###############################

sub _check_vis_group {
  my ($invocant, $value) = @_;
  my $old_id = blessed($invocant) ? $invocant->vis_group : 0;
  $value = trim($value);
  my $id;

  if ($value) {
    my $object = Bugzilla::Group->check({id => $value});
    $id = $object->id;
  }
  return $id;
}

sub _check_product_id {
  my ($invocant, $value) = @_;
  my $old_id = blessed($invocant) ? $invocant->product_id : 0;
  $value = trim($value);
  my $id;

  if ($value) {
    my $object = Bugzilla::Product->check({id => $value});
    $id = $object->id;
  }
  return $id;
}

sub _check_depends_on {
  my ($invocant, $value) = @_;
  my $old_id = blessed($invocant) ? $invocant->vis_group : 0;
  $value = trim($value);
  my $id;

  if ($value) {
    my $object = Bugzilla::Product->check({id => $value});
    $id = $object->id;
  }
  return $id;
}

###############################
####       Methods         ####
###############################

sub set_product    { $_[0]->set('product_id', $_[1]); return; }
sub set_depends_on { $_[0]->set('depends_on', $_[1]); return; }
sub set_vis_group  { $_[0]->set('vis_group',  $_[1]); return; }

sub _set_contacts {
  my $self   = shift;
  my $emails = shift;
  my $role   = shift || DP_CONTACT;

  my @ids;
  my $dbh = Bugzilla->dbh;

  foreach my $email (split(/,\s*/, $emails)) {
    if (my $user = Bugzilla::User->new({name => $email})) {
      push(@ids, $user->id);
    }
  }

  my $query
    = "DELETE FROM dependent_products_contacts WHERE record_id = ? AND contact_role = ?";

  if (@ids) {
    my $commas = '?,' x @ids;
    chop($commas);
    $query .= " AND contact NOT IN ($commas)";
  }

  $dbh->do($query, undef, $self->id, $role, @ids);

  foreach my $user_id (@ids) {
    my $contact = Bugzilla::Extension::DependentProducts::Contact->new({
      condition => 'record_id = ? AND contact = ? AND contact_role = ?',
      values    => [$self->id, $user_id, $role]
    });

    $contact
      = Bugzilla::Extension::DependentProducts::Contact->create(
      {record_id => $self->id, contact => $user_id, contact_role => $role})
      unless ($contact);

    $contact->update();
  }

  return;
}

sub set_role_contacts {
  my $self   = shift;
  my $emails = shift;
  my $role   = shift;

  return $self->_set_contacts($emails, $role);
}

sub set_contacts {
  my $self   = shift;
  my $emails = shift;

  return $self->_set_contacts($emails);
}

sub set_qe_contacts {
  my $self   = shift;
  my $emails = shift;

  return $self->_set_contacts($emails, DP_QE_MGR);
}

sub set_dm_contacts {
  my $self   = shift;
  my $emails = shift;

  return $self->_set_contacts($emails, DP_DEV_MGR);
}

sub set_pm_contacts {
  my $self   = shift;
  my $emails = shift;

  return $self->_set_contacts($emails, DP_PROD_MGR);
}

sub _contacts {
  my $self = shift;
  my $role = shift || DP_CONTACT;

  my $name = 'contacts';
  $name = 'qe_contacts' if ($role == DP_QE_MGR);
  $name = 'dm_contacts' if ($role == DP_DEV_MGR);
  $name = 'pm_contacts' if ($role == DP_PROD_MGR);

  if (!defined $self->{$name}) {
    my $dbh = Bugzilla->dbh();
    my $ids = $dbh->selectcol_arrayref(
      q{
            SELECT id FROM dependent_products_contacts
            WHERE record_id = ? AND contact_role = ?}, undef, $self->id, $role
    );

    $self->{$name}
      = Bugzilla::Extension::DependentProducts::Contact->new_from_list($ids);
  }
  return $self->{$name};
}

sub _contacts_as_email {
  my $self = shift;
  my $role = shift || DP_CONTACT;

  my $name = 'contacts_as_email';
  $name = 'qe_contacts_as_email' if ($role == DP_QE_MGR);
  $name = 'dm_contacts_as_email' if ($role == DP_DEV_MGR);
  $name = 'pm_contacts_as_email' if ($role == DP_PROD_MGR);

  if (!defined $self->{$name}) {
    my @contacts = map { $_->contact_email } @{$self->_contacts($role)};
    $self->{$name} = \@contacts;
  }
  return $self->{$name};

}

###############################
####      Accessors        ####
###############################

sub contacts {
  my $self = shift;
  return $self->_contacts();
}

sub contacts_as_email {
  my $self = shift;
  return $self->_contacts_as_email();
}

sub role_contacts_as_email {
  my $self = shift;
  my $role = shift;

  return $self->_contacts_as_email($role);
}

sub qe_contacts {
  my $self = shift;
  return $self->_contacts(DP_QE_MGR);
}

sub dm_contacts {
  my $self = shift;
  return $self->_contacts(DP_DEV_MGR);
}

sub pm_contacts {
  my $self = shift;
  return $self->_contacts(DP_PROD_MGR);
}

sub qe_contacts_as_email {
  my $self = shift;
  return $self->_contacts_as_email(DP_QE_MGR);
}

sub dm_contacts_as_email {
  my $self = shift;
  return $self->_contacts_as_email(DP_DEV_MGR);
}

sub pm_contacts_as_email {
  my $self = shift;
  return $self->_contacts_as_email(DP_PROD_MGR);
}

sub product_id { return $_[0]->{product_id}; }
sub depends_on { return $_[0]->{depends_on}; }
sub vis_group  { return $_[0]->{vis_group}; }

sub all_contacts {
  my $self = shift;

  if (!defined $self->{all_contacts}) {
    my @contacts;

    foreach my $role (DP_CONTACT, DP_QE_MGR, DP_DEV_MGR, DP_PROD_MGR) {
      my $conts = $self->_contacts($role);
      @contacts = (@contacts, @{$conts}) if (@{$conts});
    }
    $self->{all_contacts} = \@contacts;
  }

  return ($self->{all_contacts});
}

sub user_can_view {
  my ($self, $the_user) = @_;

  $the_user //= Bugzilla->user;
  $the_user = _get_user($the_user);

  return (0) unless ($the_user->id);
  return (1) unless ($self->vis_group);

  return ($the_user->in_group_id($self->vis_group));
}

sub _get_user {
  my $the_user = shift;
  if (!blessed $the_user) {
    if ($the_user =~ /^\d+$/) {
      $the_user = Bugzilla::User->check({id => $the_user});
    }
    else {
      $the_user = Bugzilla::User->check($the_user);
    }
  }

  return $the_user;
}

1;

__END__

=head1 Description

Bugzilla::Extension::DependentProducts::Record - This modules maps the relationship of products to dependent products.

=head1 Fields

=over 4

=item product_id

The ID of the dependent product.

=item depends_on

The ID of the product dependend on.

=item vis_group

The ID of the group allowed to see this relationship.

=back

=head1 Methods

These methods allow you to set the specified field.

=over 4

=item set_contacts

=item set_role_contacts

=item set_qe_contacts

=item set_dm_contacts

=item set_pm_contacts

=back

=head1 Accessors


These methods allow you to get the specified field.

=over 4

=item contacts

=item contacts_as_email

=item role_contacts

=item role_contacts_as_email

=item qe_contacts

=item qe_contacts_as_email

=item dm_contacts

=item dm_contacts_as_email

=item pm_contacts

=item pm_contacts_as_email

=item product_id

=item depends_on

=item vis_group

=item all_contacts

=item user_can_view

Can the specified, or current, user view this field on bugs in this product?

=back


