# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::ExternalBugs::Type::SFDC_Sandbox;

use strict;
use warnings;
use 5.10.1;

use base qw(Bugzilla::Extension::ExternalBugs::Type);

use Bugzilla;
use Bugzilla::Constants;
use Bugzilla::Error;

use Bugzilla::WebService::Server::XMLRPC;
use Bugzilla::WebService::Bug;

use Bugzilla::Extension::ExternalBugs::Util qw(mail_error sfdc_get_directory);

use Data::Dumper;
use File::Spec::Functions qw(catfile);
use SOAP::Lite;
use WWW::Salesforce::Simple;

use constant TYPE => 'SFDC_Sandbox';

use constant WSDL_VERSION => '46.0';

our $HOSTNAME = Bugzilla->params->{'sfdc_sandbox_server'};
our $SERVER   = "https://test.salesforce.com/services/Soap/u/" . WSDL_VERSION;
our $CACHE    = sfdc_get_directory('Sandbox', $HOSTNAME);
our $WSDL     = 'file:' . catfile(bz_locations()->{datadir}, 'WSDL-' . WSDL_VERSION . '.xml');

$WWW::Salesforce::SF_URI = 'urn:partner.soap.sforce.com';

my $logger = Bugzilla->logger;

sub get_data {
  my ($self, $ext_bug) = @_;

  unless ($ext_bug->ext_bz_bug_id =~ /^\d+$/) {
    $logger->debug("SFDC Bug id " . $ext_bug->ext_bz_bug_id . " is not a number.");
    return;
  }

  unless ($self->ext_login && $self->ext_password) {
    $logger->debug(
      "SFDC SOAP: Warning: No username/password information for " . $self->url);
    return;
  }

  my $result = undef;
  my $sforce = WWW::Salesforce::Simple->new(
    serverurl => $SERVER,
    username  => $self->ext_login,
    password  => $self->ext_password,
  );
  if (not $sforce) {
    die "SFDC SOAP: WWW::Salesforce login failure";
  }
  $logger->debug("SFDC SOAP: Login successful");

  my $soql
    = "SELECT Severity__c, Status, Subject, Account.Name FROM Case WHERE CaseNumber = '"
    . $ext_bug->ext_bz_bug_id . "'";

  $result = $sforce->do_query($soql);

  if (!$result) {

    # Probably not a valid case number
    return;
  }

  my %data
    = (priority => $result->[0]{Severity__c}, status => $result->[0]{Status},);

  if ( defined($result->[0]{'Account'})
    && defined($result->[0]{'Account'}->{'Name'}))
  {
    $data{description}
      = '[' . $result->[0]{'Account'}->{'Name'} . '] ' . $result->[0]{Subject};
  }
  else {
    $data{description} = '[unknown] ' . $result->[0]{Subject};
  }

  return \%data;
}

sub can_sync {
  my ($self, $ext_bug) = @_;
  return 1;
}

sub send_changes {
  my ($self, $params) = @_;

  $logger->debug("SFDC SOAP: Entering send_changes");

  my $ext_bug = $params->{'ext_bug'};
  my $bug     = $params->{'bug'};
  my $start   = $params->{'start'};

  my $serializer = Bugzilla::XMLRPC::Serializer->new();

  # Send a full copy of the bug report using Bug.get
  my $bug_data = Bugzilla::WebService::Bug->get({
    ids          => [$bug->id],
    extra_fields => ['external_bugs', 'description', 'flags'],
  });
  my $bug_xml_data = $serializer->envelope('response', 'Bug.get', $bug_data);

  # Get the comments for this bug
  my $bug_comments = Bugzilla::WebService::Bug->comments(
    {ids => [$bug->id], new_since => $start});
  my $bug_xml_comments
    = $serializer->envelope('response', 'Bug.comments', $bug_comments);

  ## REDHAT EXTENSION 965415: Bugzilla 4.4 switches the above calls to
  # the shadow db. You need to switch make to the main one
  Bugzilla->switch_to_main_db();

  # We now need to send the events to the Job Queue. However, we must set
  # We need to record the ext_bz info since there is no guarantee that the
  # row will exist at the time we send the event.
  Bugzilla->job_queue->insert(
    'send_sfdc',
    {
      ext_bz_type_id   => $ext_bug->ext_bz_id,
      bug_id           => $ext_bug->bug_id,
      bug_xml_data     => $bug_xml_data,
      bug_xml_comments => $bug_xml_comments,
    }
  );

  return 1;
}

sub send_to_sfdc {
  my $self             = shift;
  my $bug_id           = shift;
  my $bug_xml_data     = shift;
  my $bug_xml_comments = shift;

  unless ($self->ext_login && $self->ext_password) {
    $logger->debug(
      "SFDC SOAP: Warning: No username/password information for " . $self->url);
    return 1;
  }

  eval {
    my $sforce = WWW::Salesforce::Simple->new(
      serverurl => $SERVER,
      username  => $self->ext_login,
      password  => $self->ext_password,
    );
    if (not $sforce) {
      die "SFDC SOAP: WWW::Salesforce login failure";
    }
    $logger->debug("SFDC SOAP: Login successful");
    my $client = $sforce->get_client(1);
    my @errors;
    my $result = $client->cache_dir($CACHE)->service($WSDL)
      ->updateBug($bug_xml_data, $sforce->get_session_header());
    push(@errors, "updateBug returned failure ($result)")
      if (!$result || $result eq 'false');
    if ($bug_xml_comments) {
      $result = $client->cache_dir($CACHE)->service($WSDL)
        ->upsertComments($bug_xml_comments, $sforce->get_session_header());
      push(@errors, "upsertComments returned failure ($result)")
        if (!$result || $result eq 'false');
    }

    if (@errors) {
      die("Call(s) failed: " . join(", ", @errors));
    }
  };


  if (my $call_error = $@) {
    die $call_error;
  }

  $logger->debug("SFDC SOAP: Success: changes sent for bug $bug_id");
  return 1;
}

sub sfdc_hostname {
  return 'c.cs11.visual.force.com';
}

1;
