#!/usr/bin/perl -T
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

################################################################################
# Script Initialization
################################################################################

use strict;
use warnings;
use 5.10.1;

use lib qw(/var/www/html/bugzilla/ .);

use constant MAX_SITEMAP_LINKS => 50000;

use Bugzilla;
use Bugzilla::Util qw(correct_urlbase);

# create some handles that we'll need
my $dbh = Bugzilla->dbh;

################################################################################
# Main Body Execution
################################################################################

# This script needs to the database for bugs that are publicly viewable
# and records the url to access the bug in a sitemap xml file. This file can
# be used by search engines to determine which bugs needs to be re-indexed
# on a daily basis.
#
# A sitemap xml file can only contain a maximum of 50,000 links so we are
# breaking the list up into separate sitemap files and then use an index
# sitemap pointing to each of the individual files.
#
# Each individual sitemap file will be numbered and compressed.
#
# Sample sitemap files are after __END__

main(@ARGV) if not caller();

sub main {
  my $filecount = 1;
  my @filelist  = ();

  my $sth = $dbh->prepare(
    "SELECT bugs.bug_id, " . $dbh->sql_date_format('bugs.delta_ts', '%Y-%m-%d') . "
                             FROM   bugs LEFT JOIN bug_group_map ON bugs.bug_id = bug_group_map.bug_id 
                             WHERE  bug_group_map.group_id IS NULL ORDER BY bugs.bug_id "
      . $dbh->sql_limit(MAX_SITEMAP_LINKS, '?')
  );

  while (1) {
    my @bugs;

    $sth->execute(MAX_SITEMAP_LINKS * ($filecount - 1));

    while (my ($bug_id, $delta_ts) = $sth->fetchrow_array()) {
      push(@bugs, {bug_id => $bug_id, delta_ts => $delta_ts});
    }

    last if !@bugs;

    push(@filelist, _generate_sitemap_file($filecount, @bugs));
    $filecount++;
  }

  # Generate index file
  _generate_sitemap_index(@filelist);

  return;
}


################################################################################
# Functions
################################################################################

sub _generate_sitemap_index {
  my @filelist = @_;
  my $OUT;

  my $timestamp = $dbh->selectrow_array(
    "SELECT " . $dbh->sql_date_format('NOW()', '%Y-%m-%d'));

  open($OUT, ">", "sitemap_index.xml") || die "Could not open sitemap index: $!";
  print $OUT <<END;
<?xml version="1.0" encoding="UTF-8"?>
<sitemapindex xmlns="http://www.sitemaps.org/schemas/sitemap/0.9">
END

  foreach my $filename (@filelist) {
    print $OUT "
  <sitemap>
    <loc>" . correct_urlbase() . "$filename</loc>
    <lastmod>$timestamp</lastmod>
  </sitemap>
";
  }

  print $OUT <<END;
</sitemapindex>
END

  close($OUT) || die "Could not close sitemap index: $!";

  return;
}

sub _generate_sitemap_file {
  my ($filecount, @bugs) = @_;
  my $OUT;

  open($OUT, ">", "sitemap$filecount.xml")
    || die "Could not open sitemap index: $!";

  print $OUT <<END;
<?xml version="1.0" encoding="UTF-8"?>
<urlset xmlns="http://www.sitemaps.org/schemas/sitemap/0.9">
END

  foreach my $bugref (@bugs) {
    print $OUT "
  <url>
    <loc>" . correct_urlbase() . "show_bug.cgi?id=$bugref->{bug_id}</loc>
    <lastmod>$bugref->{delta_ts}</lastmod>
  </url> 
";
  }

  print $OUT <<END;
</urlset>
END

  close($OUT) || die "Could not close sitemap index: $!";

  # Compress the file
  system("/bin/gzip", "-f", "sitemap$filecount.xml") == 0
    || die "Error executing gzip -f sitemap$filecount.xml";

  return "sitemap$filecount.xml.gz";
}

__END__

Sitemap Index:

<?xml version="1.0" encoding="UTF-8"?>
<sitemapindex xmlns="http://www.sitemaps.org/schemas/sitemap/0.9">
  <sitemap>
    <loc>https://bugzilla.redhat.com/sitemap1.xml.gz</loc>
    <lastmod>2004-10-01</lastmod>
  </sitemap>
  <sitemap>
    <loc>https://bugzilla.redhat.com/sitemap2.xml.gz</loc>
    <lastmod>2005-01-01</lastmod>
  </sitemap>
</sitemapindex>

Single Sitemap:

<?xml version="1.0" encoding="UTF-8"?>
<urlset xmlns="http://www.sitemaps.org/schemas/sitemap/0.9">
  <url>
    <loc>https://bugzilla.redhat.com/show_bug.cgi?id=XXXXXX</loc>
    <lastmod>2005-01-01</lastmod>
  </url>  
</urlset>

=head1 NAME

sitemap.pl - Generates a list of sitemap files that contain 
the urls to the public bugs in Bugzilla.

=head1 SYNOPSIS

cd /path/to/bugzilla && perl sitemap.pl

# cron example
MAILTO="bugzilla-owner@redhat.com"
# run 13 minutes past midnight every day
0 13 * * * cd <your-bugzilla-directory>; perl -wT ./sitemap.pl 

=head1 DESCRIPTION

Generates a list of sitemap files that contain the urls to the 
public bugs in Bugzilla. Each numbered file contains up to 
50,000 links and there is also a sitemap_index.xml that points 
to each of the individual compressed files.

=head1 ENVIRONMENT 

Needs to be in the root directory of the Bugzilla installation in order
to load the perl libs.

