#!/usr/bin/perl -w
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

# Bug 523845, 1114398
# This script will unshare a saved search if it has not been used
# for X amount of time.

use strict;
use warnings;

use lib qw(. /var/www/html/bugzilla/);

use Bugzilla;
use Bugzilla::Constants;
use Bugzilla::Search;
use Bugzilla::Search::Saved;
use Bugzilla::User;
use Bugzilla::Mailer;
use Bugzilla::Util;

my $conf = Bugzilla->params->{log4perl_config} || die;
Log::Log4perl::init_once(\$conf);


if (Bugzilla->params->{expired_search_time} eq 'Never') {
  exit 0;
}

my $time = Bugzilla->params->{expired_search_time};

my %table = (
  '1 Month'   => '30',
  '3 Months'  => '90',
  '6 Months'  => '180',
  '9 Months'  => '270',
  '12 Months' => '365',
);

# Make sure the currently defined time in the config is valid
unless (defined($table{$time})) {
  print STDERR "Invalid expiry time of $time";
  exit 0;
}

$time = $table{$time};

process_queries();

exit(0);

sub process_queries {

  my $dbh      = Bugzilla->dbh;
  my $sql_date = $dbh->sql_date_math('NOW()', '-', $time, 'DAY');

  my $sql = "SELECT userid, namedquery_id, last_used
               FROM namedquery_group_map nqgm
               JOIN namedqueries nq ON (nq.id = nqgm.namedquery_id)
               WHERE ($sql_date) > last_used
               ORDER BY userid";

  my $result = $dbh->selectall_arrayref($sql);

  # Build up a list of users and their queries.
  # If a user has multiple queries being expired then we process
  # them all at once and only send the user one mail.
  my %users;

  $dbh->bz_start_transaction();

  foreach my $row (@$result) {
    my $user = Bugzilla::User->new($row->[0]);
    push @{$users{$row->[0]}}, $row->[1];

    $dbh->do('DELETE FROM namedquery_group_map WHERE namedquery_id = ?',
      undef, $row->[1]);
  }

  $dbh->bz_commit_transaction();

  notify_owners(\%users);

  return;
}

sub notify_owners {
  my $owners = shift;

  my $timestamp = Bugzilla->dbh->selectrow_array('SELECT NOW()');


  my %args = (
    timeperiod => Bugzilla->params->{expired_search_time},
    from       => Bugzilla->params->{'mailfrom'},
    date       => format_time($timestamp, '%a, %d %b %Y %T %z', 'UTC'),
  );

  while (my ($userid, $querylist) = each(%$owners)) {
    my $user = Bugzilla::User->new($userid);

    next if $user->email_disabled;

    my @queries = map { Bugzilla::Search::Saved->new($_) } @$querylist;

    $args{to_user}  = $user;
    $args{searches} = \@queries;

    my $template = Bugzilla->template_inner('en');

    my $msg;
    $template->process('email/unshare-search.txt.tmpl', \%args, \$msg);

    MessageToMTA($msg);
  }

  return;
}

