# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::RedHat::Schema;

use strict;
use warnings;
use 5.10.1;

use Bugzilla;
use Bugzilla::Constants;
use Bugzilla::User::Setting;

use base qw(Exporter);
our @EXPORT = qw(
  redhat_db_schema_abstract_schema
  redhat_install_before_final_checks
  redhat_install_update_db
);

## REDHAT EXTENSION BEGIN 917935
use YAML::Syck qw(LoadFile);
## REDHAT EXTENSION END 917935

use JSON            qw(from_json);
use List::MoreUtils qw(none);

sub redhat_db_schema_abstract_schema {
  my ($self, $args) = @_;

  my $schema = $args->{schema};

  ## REDHAT EXTENSION START 420461
  $schema->{'longdescs_activity'} = {
    FIELDS => [
      comment_id => {
        TYPE       => 'INT4',
        NOTNULL    => 1,
        REFERENCES =>
          {TABLE => 'longdescs', COLUMN => 'comment_id', DELETE => 'CASCADE'}
      },
      who => {
        TYPE       => 'INT3',
        NOTNULL    => 1,
        REFERENCES => {TABLE => 'profiles', COLUMN => 'userid', DELETE => 'CASCADE'}
      },
      change_when => {TYPE => 'DATETIME', NOTNULL => 1},
      old_comment => {TYPE => 'LONGTEXT', NOTNULL => 1},
      security    => {TYPE => 'BOOLEAN',  NOTNULL => 1, DEFAULT => 0},
    ],
    INDEXES => [
      longdescs_activity_comment_id_idx             => ['comment_id'],
      longdescs_activity_change_when_idx            => ['change_when'],
      longdescs_activity_comment_id_change_when_idx => [qw(comment_id change_when)],
    ],
  };
  ## REDHAT EXTENSION END 420461

  $schema->{bugs_release} = {
    FIELDS => [
      bug_id => {
        TYPE       => 'INT3',
        NOTNULL    => 1,
        REFERENCES => {TABLE => 'bugs', COLUMN => 'bug_id', DELETE => 'CASCADE'}
      },
      value => {TYPE => 'varchar(64)', NOTNULL => 1, DEFAULT => "'---'"},
    ],
    INDEXES => [
      bugs_release_bug_id_idx       => ['bug_id'],
      bugs_release_target_value_idx => ['value'],
    ],
  };

  $schema->{releases} = {
    FIELDS => [
      id         => {TYPE => 'MEDIUMSERIAL', NOTNULL => 1, PRIMARYKEY => 1},
      product_id => {
        TYPE       => 'INT2',
        NOTNULL    => 1,
        REFERENCES => {TABLE => 'products', COLUMN => 'id', DELETE => 'CASCADE'}
      },
      value    => {TYPE => 'varchar(64)', NOTNULL => 1},
      sortkey  => {TYPE => 'INT2',        NOTNULL => 1, DEFAULT => 0},
      isactive => {TYPE => 'BOOLEAN',     NOTNULL => 1, DEFAULT => 'TRUE'},
    ],
    INDEXES => [
      releases_product_id_idx => {FIELDS => [qw(product_id value)], TYPE => 'UNIQUE'},
    ],
  };

  $schema->{doc_type_text} = {
    FIELDS => [
      id             => {TYPE => 'MEDIUMSERIAL', NOTNULL => 1, PRIMARYKEY => 1},
      cf_doc_type_id => {
        TYPE       => 'INT2',
        NOTNULL    => 1,
        REFERENCES => {TABLE => 'cf_doc_type', COLUMN => 'id', DELETE => 'CASCADE'}
      },
      value => {TYPE => 'LONGTEXT', NOTNULL => 1},
    ],
    INDEXES => [
      doc_type_text_cf_doc_type_id_idx =>
        {FIELDS => [qw(cf_doc_type_id)], TYPE => 'UNIQUE'},
    ],
  };

  ## REDHAT EXTENSION START 1824881
  $schema->{'comment_private_groups'} = {
    FIELDS => [
      comment_id => {
        TYPE       => 'INT4',
        NOTNULL    => 1,
        REFERENCES =>
          {TABLE => 'longdescs', COLUMN => 'comment_id', DELETE => 'CASCADE'}
      },
      group_id => {
        TYPE       => 'INT3',
        NOTNULL    => 1,
        REFERENCES => {TABLE => 'groups', COLUMN => 'id', DELETE => 'CASCADE'}
      },
    ],
    INDEXES => [
      comment_private_groups_join_idx =>
        {FIELDS => [qw(comment_id group_id)], TYPE => 'UNIQUE'}
    ],
  };
  ## REDHAT EXTENSION END 1824881

  ## REDHAT EXTENSION START 1676959
  $schema->{'report_groups'} = {
    FIELDS => [
      report_id => {
        TYPE       => 'INT4',
        NOTNULL    => 1,
        REFERENCES => {TABLE => 'reports', COLUMN => 'id', DELETE => 'CASCADE'}
      },
      group_id => {
        TYPE       => 'INT3',
        NOTNULL    => 1,
        REFERENCES => {TABLE => 'groups', COLUMN => 'id', DELETE => 'CASCADE'}
      },
    ],
    INDEXES => [
      report_groups_join_idx => {FIELDS => [qw(report_id group_id)], TYPE => 'UNIQUE'}
    ],
  };
  ## REDHAT EXTENSION END 1676959

  ## REDHAT EXTENSION START 1999932
  # Create params table, one day this will be used to replace params.json.
  $schema->{'params'} = {
    FIELDS => [
      id       => {TYPE => 'MEDIUMSERIAL', NOTNULL => 1, PRIMARYKEY => 1},
      name     => {TYPE => 'LONGTEXT',     NOTNULL => 1},
      value    => {TYPE => 'LONGTEXT',     NOTNULL => 1},
      isactive => {TYPE => 'BOOLEAN',      NOTNULL => 1, DEFAULT => 'TRUE'},
      inui     => {TYPE => 'BOOLEAN',      NOTNULL => 1, DEFAULT => 'TRUE'},
    ],
    INDEXES => [params_name_idx => {FIELDS => [qw(name)], TYPE => 'UNIQUE'}],
  };
  ## REDHAT EXTENSION END 1999932

  ## REDHAT EXTENSION START 2162148
  $schema->{product_private_comment_groups} = {
    FIELDS => [
      id         => {TYPE => 'MEDIUMSERIAL', NOTNULL => 1, PRIMARYKEY => 1},
      product_id => {
        TYPE       => 'INT2',
        NOTNULL    => 1,
        REFERENCES => {TABLE => 'products', COLUMN => 'id', DELETE => 'CASCADE'}
      },
      group_id => {
        TYPE       => 'INT3',
        NOTNULL    => 1,
        REFERENCES => {TABLE => 'groups', COLUMN => 'id', DELETE => 'CASCADE'}
      },
    ],
    INDEXES => [
      product_private_comment_groups_join_idx =>
        {FIELDS => [qw(product_id group_id)], TYPE => 'UNIQUE'},
    ],
  };
  ## REDHAT EXTENSION END 2162148

  return;
}

sub redhat_install_update_db {
  my ($self, $args) = @_;

  my $dbh = Bugzilla->dbh;

  ## REDHAT EXTENSION START 2236629
  $dbh->bz_add_column('products', 'banner',
      {TYPE => 'LONGTEXT', NOTNULL => 1, DEFAULT => "''"});
  ## REDHAT EXTENSION END 2236629

  _update_custom_fields(['cf_doc_type']);

  ## REDHAT EXTENSION 406171
  $dbh->bz_add_column('bug_status', 'description', {TYPE => 'MEDIUMTEXT'});

  ## REDHAT - Removing password expiry related columns from profiles
  #          table , bug #509100
  $dbh->bz_drop_column('profiles', 'password_expire_days');
  $dbh->bz_drop_column('profiles', 'password_last_changed');

  # 2008-01-31 dkl@redhat.com - Bug 174037
  $dbh->bz_add_column('flagtypes', 'view_group_id', {TYPE => 'INT3'});

  _update_status_and_resolution();

  ## REDHAT EXTENSION START 831328
  $dbh->bz_add_column('bugs_activity', 'object_id', {TYPE => 'INT3'});
  ## REDHAT EXTENSION END 831328

  _populate_releases_table();

  # Made the size of the value field in the releases table match
  # the size of the value column in the bugs_release table.
  $dbh->bz_alter_column('releases', 'value',
    {TYPE => 'varchar(64)', NOTNULL => 1});

  _add_products_defaultrelease();


  if (!exists $dbh->bz_column_info('releases', 'sortkey')->{DEFAULT}) {
    $dbh->bz_alter_column('releases', 'sortkey',
      {TYPE => 'INT2', NOTNULL => 1, DEFAULT => 0});
  }


  if ($dbh->bz_column_info('products', 'releaseurl')) {
    $dbh->bz_alter_column('products', 'releaseurl',
      {TYPE => 'TINYTEXT', NOTNULL => 1, DEFAULT => "''"});
  }

  $dbh->bz_add_column('releases', 'id',
    {TYPE => 'MEDIUMSERIAL', NOTNULL => 1, PRIMARYKEY => 1});

  $dbh->bz_drop_column('products', 'releaseurl');

  # 2011-02-16 Drop the column (bugs_release table replaces it)
  if ($dbh->bz_column_info('bugs', 'target_release')) {
    $dbh->bz_drop_column('bugs', 'target_release');
  }

  ## REDHAT EXTENSION BEGIN 706784
  if (!$dbh->bz_column_info('products', 'multiple_components')) {
    $dbh->bz_add_column('products', 'multiple_components',
      {TYPE => 'BOOLEAN', NOTNULL => 1, DEFAULT => 0},
    );
  }

  if (!$dbh->bz_column_info('products', 'multiple_target_releases')) {
    $dbh->bz_add_column('products', 'multiple_target_releases',
      {TYPE => 'BOOLEAN', NOTNULL => 1, DEFAULT => 0},
    );
  }

  if (!$dbh->bz_column_info('products', 'multiple_versions')) {
    $dbh->bz_add_column('products', 'multiple_versions',
      {TYPE => 'BOOLEAN', NOTNULL => 1, DEFAULT => 0},
    );
  }
  ## REDHAT EXTENSION END 706784

  ## REDHAT EXTENSION BEGIN 653915
  if (!$dbh->bz_column_info('setting', 'required_groups')) {
    $dbh->bz_add_column('setting', 'required_groups', {TYPE => 'VARCHAR(32)'},);
  }
  ## REDHAT EXTENSION END 653915

  ## REDHAT EXTENSION BEGIN 711031
  if (!$dbh->bz_column_info('flagtypes', 'category')) {
    $dbh->bz_add_column('flagtypes', 'category', {TYPE => 'VARCHAR(50)'});
  }
  ## REDHAT EXTENSION END 711031

  ## REDHAT EXTENSION BEGIN 751352
  if (!$dbh->bz_column_info('groups', 'category')) {
    $dbh->bz_add_column('groups', 'category', {TYPE => 'VARCHAR(50)'});
  }
  ## REDHAT EXTENSION END 751352

  # As recommended by Monty Widenius for GNOME's upgrade.
  if (Bugzilla->localconfig->{db_driver} eq 'mysql') {
    $dbh->do('SET SESSION myisam_sort_buffer_size = 3221225472');
  }

  ## REDHAT EXTENSION BEGIN
  if (!$dbh->bz_column_info('releases', 'isactive')) {
    $dbh->bz_add_column('releases', 'isactive',
      {TYPE => 'BOOLEAN', NOTNULL => 1, DEFAULT => 'TRUE'});
  }
  ## REDHAT EXTENSION END

  ## REDHAT EXTENSION START 829154
  $dbh->bz_add_column('products', 'always_private',
    {TYPE => 'BOOLEAN', NOTNULL => 1, DEFAULT => 'FALSE'});
  ## REDHAT EXTENSION END 829154

  # This column is no longer used, so should be dropped
  $dbh->bz_drop_column('fielddefs', 'public');

  ## REDHAT EXTENSION START 523845
  # This will monitor when a saved search was lased used, so it can be
  # removed after an expiry period.
  $dbh->bz_add_column('namedquery_group_map', 'last_used',
    {TYPE => 'DATETIME', 'DEFAULT' => 'NOW()', NOTNULL => 1});
  ## REDHAT EXTENSION END 523845

  ## REDHAT EXTENSION START 876015
  $dbh->bz_add_column('bugs', 'docs_contact', {TYPE => 'INT3'});
  $dbh->bz_add_fk("bugs", "docs_contact",
    {TABLE => 'profiles', COLUMN => 'userid', DELETE => 'CASCADE'});
  $dbh->bz_add_column('components', 'initialdocscontact', {TYPE => 'INT3'});
  $dbh->bz_add_fk('components', 'initialdocscontact',
    {TABLE => 'profiles', COLUMN => 'userid', DELETE => 'SET NULL'});
  ## REDHAT EXTENSION END 876015

  # BUGZILLA 4.4 UPGRADE
  $dbh->bz_alter_column('longdescs_activity', 'comment_id',
    {TYPE => 'INT4', NOTNULL => 1});

  if (!$dbh->bz_column_info('longdescs_activity', 'security')) {

    # Default new field off
    $dbh->bz_add_column('longdescs_activity', 'security',
      {TYPE => 'BOOLEAN', NOTNULL => 1, DEFAULT => 0},
    );

    # But set it on for all existing changes
    $dbh->do('UPDATE longdescs_activity SET security = 1');
  }

  ## REDHAT EXTENSION BEGIN 903031
  $dbh->bz_alter_column('bugs_aliases', 'alias', {TYPE => 'varchar(40)'});
  ## REDHAT EXTENSION END 903031

  ## REDHAT EXTENSION 993863: While these are not strictly related to this
  # bug, they were uncovered when implementing it
  $dbh->bz_alter_column('namedqueries', 'name',
    {TYPE => 'varchar(255)', NOTNULL => 1});

  $dbh->bz_alter_column('profiles_activity', 'oldvalue',
    {TYPE => 'varchar(2000)'});
  $dbh->bz_alter_column('profiles_activity', 'newvalue',
    {TYPE => 'varchar(2000)'});

  ## REDHAT EXTENSION START 1067210
  $dbh->bz_add_index('component_cc', 'component_cc_comp_id_idx',
    ['component_id']);
  ## REDHAT EXTENSION END 1067210

  ## REDHAT EXTENSION START 1326559
  # was rh_on_dev_products
  if (!$dbh->bz_column_info('products', 'allows_on_dev')) {
    $dbh->bz_add_column('products', 'allows_on_dev',
      {TYPE => 'BOOLEAN', NOTNULL => 1, DEFAULT => 0},
    );
  }

  # was rh_no_clear_acks
  if (!$dbh->bz_column_info('products', 'allows_no_clear_acks')) {
    $dbh->bz_add_column('products', 'allows_no_clear_acks',
      {TYPE => 'BOOLEAN', NOTNULL => 1, DEFAULT => 0},
    );
  }

  # was rh_always_cc
  if (!$dbh->bz_column_info('components', 'always_cc')) {
    $dbh->bz_add_column('components', 'always_cc',
      {TYPE => 'varchar(1024)', NOTNULL => 1, DEFAULT => "''"},
    );
  }
  ## REDHAT EXTENSION END 1326559

  ## REDHAT EXTENSION START 2019778
  if (!$dbh->bz_column_info('products', 'default_to_open')) {
    $dbh->bz_add_column('products', 'default_to_open',
      {TYPE => 'BOOLEAN', NOTNULL => 1, DEFAULT => 0},
    );
  }
  ## REDHAT EXTENSION END 2019778

  ## REDHAT EXTENSION BEGIN 1651832
  if (Bugzilla->params->{'group_categories'} =~ m/Admin/) {
    my $sql
      = q{select count (*) from group_group_map where member_id = 6 and grantor_id not in (select id from groups where category = 'Admin')};
    my $res   = $dbh->selectall_arrayref($sql);
    my $count = $res->[0]->[0];
    if ($count ne '0') {
      $sql
        = q{delete from group_group_map where member_id = 6 and grantor_id not in (select id from groups where category = 'Admin')};
      $dbh->do($sql);
    }
  }
  ## REDHAT EXTENSION END 1651832

  ## REDHAT EXTENSION BEGIN 1737448
  if (!$dbh->bz_column_info('logincookies', 'id')
    && lc(Bugzilla->localconfig->{db_driver}) ne 'sqlite')
  {

    # Primary key changes need to be committed before adding new pkey.
    $dbh->do(
      'ALTER TABLE logincookies DROP CONSTRAINT IF EXISTS logincookies_pkey');

    $dbh->_bz_real_schema->set_column('logincookies', 'cookie',
      {'TYPE' => 'varchar(16)', 'NOTNULL' => 1});
    $dbh->_bz_store_real_schema;

    $dbh->bz_alter_column('logincookies', 'cookie',
      {TYPE => 'varchar(22)', NOTNULL => 1});
    $dbh->bz_add_index('logincookies', 'logincookies_cookie_idx',
      {FIELDS => ['cookie'], TYPE => 'UNIQUE'});

    $dbh->bz_add_column('logincookies', 'id',
      {TYPE => 'INTSERIAL', NOTNULL => 1, PRIMARYKEY => 1});
    $dbh->bz_add_index('logincookies', 'logincookies_cookie_idx',
      {FIELDS => ['cookie'], TYPE => 'UNIQUE'});
  }
  ## REDHAT EXTENSION END 1737448

  ## REDHAT EXTENSION BEGIN 2075342
  if (!$dbh->bz_column_info('logincookies', 'expires')
    && lc(Bugzilla->localconfig->{db_driver}) ne 'sqlite')
  {
    $dbh->bz_start_transaction();
    $dbh->do(q{DELETE FROM logincookies});
    $dbh->bz_add_column('logincookies', 'expires', {TYPE => 'DATETIME'});
    $dbh->bz_commit_transaction();
  }

  if (Bugzilla->expire_cookies) {
    $dbh->do(q{
      DELETE FROM logincookies
      WHERE expires is NULL OR expires <= NOW()
    });
  }
  ## REDHAT EXTENSION END 2075342

  ## REDHAT EXTENSION BEGIN 1853587
  if (!$dbh->bz_index_info('bugs_aliases', 'bugs_aliases_lower_idx')
    && Bugzilla->is_pg)
  {

    # Delete all the duplicates
    $dbh->do(
      q{
  DELETE FROM bugs_aliases
  WHERE alias IN (
      SELECT
          alias
      FROM (
          SELECT
              alias,
              row_number() OVER w as rnum
          FROM bugs_aliases
          WINDOW w AS (
              PARTITION BY lower(alias)
              ORDER BY alias
          )

      ) t
  WHERE t.rnum > 1);
}
    );

    # Add index
    $dbh->bz_add_index('bugs_aliases', 'bugs_aliases_lower_idx',
      {FIELDS => ['lower(alias)'], TYPE => 'UNIQUE'});
  }
  ## REDHAT EXTENSION END 1853587

  ## REDHAT EXTENSION START 1426216
  ## Create an index that matches user names really fast
  ## Search like:
  ##   select userid, login_name, is_enabled from profiles
  ##   where login_name ilike '%test%' OR realname ilike '%test%'
  if (Bugzilla->is_pg) {
    $dbh->do(<<'SQLDATA');
DO
$$
BEGIN
   IF NOT EXISTS (
      SELECT *
      FROM   pg_class c
      JOIN   pg_namespace n ON n.oid = c.relnamespace
      WHERE  c.relname = 'profiles_login_name_realname_trgm_idx'
      AND    n.nspname = 'public'
   ) THEN
      CREATE INDEX profiles_login_name_realname_trgm_idx ON public.profiles USING gin (login_name gin_trgm_ops, realname gin_trgm_ops);
    END IF;
END
$$;
SQLDATA
  }
  ## REDHAT EXTENSION END 1426216

  $dbh->bz_add_column('profiles', 'prefs_admin_group', {TYPE => 'INT3'});
  $dbh->bz_add_fk('profiles', 'prefs_admin_group',
    {TABLE => 'groups', COLUMN => 'id', DELETE => 'SET NULL'});

  ## REDHAT EXTENSION START 1937776
  if (!$dbh->bz_column_info('components', 'default_to_private')) {
    $dbh->bz_start_transaction();

    $dbh->bz_add_column('components', 'default_to_private',
      {TYPE => 'BOOLEAN', NOTNULL => 1, DEFAULT => 0},
    );
    $dbh->bz_add_column('rh_sub_components', 'default_to_private',
      {TYPE => 'BOOLEAN', NOTNULL => 1, DEFAULT => 0},
    );

    if (lc(Bugzilla->localconfig->{db_driver}) eq 'Pg') {

      # Set components
      my $SQL = <<'EOS';
      UPDATE components
      SET default_to_private = 1
      WHERE name IN (
        'kernel',
        'distribution',
        'kernel-pegas',
        'kernel-alt'
      )
      AND product_id IN (
        SELECT id
        FROM PRODUCTS
        WHERE name ~ '^Red Hat Enterprise Linux [5-9]$'
      )
EOS

      $dbh->do($SQL);

      # Set sub-components
      $SQL = <<'EOS';
      UPDATE rh_sub_components
      SET default_to_private = 1
      WHERE component_id IN (
        SELECT id FROM components
        WHERE name IN (
          'kernel',
          'distribution',
          'kernel-pegas',
          'kernel-alt'
        )
        AND product_id IN (
          SELECT id
          FROM PRODUCTS
          WHERE name ~ '^Red Hat Enterprise Linux [5-9]$'
        )
      )
EOS

      $dbh->do($SQL);

      # set entitlements
      $SQL = <<'EOS';
      UPDATE components SET default_to_private = 1
      WHERE
      name = 'entitlements'
      AND product_id IN (
        SELECT id
        FROM PRODUCTS
        WHERE name IN (
          'Red Hat Enterprise Linux 6',
          'Red Hat Enterprise Linux 7',
          'Red Hat Enterprise Linux 8',
          'Red Hat Enterprise Linux 9',
          'Red Hat Enterprise Virtualization Manager',
          'Red Hat OpenStack',
          'OpenShift Container Platform',
          'Red Hat Satellite 5',
          'Red Hat Satellite 6',
          'Red Hat Certification Program',
          'Red Hat Certificate System',
          'Red Hat Directory Server',
          'Red Hat Gluster Storage',
          'Red Hat Ceph Storage',
          'Red Hat Quickstart Cloud Installer'
        )
      )
EOS

      $dbh->do($SQL);
    }

    $dbh->bz_commit_transaction();

  }
  ## REDHAT EXTENSION END 1937776

  ## REDHAT EXTENSION START 1142506
  if (Bugzilla->is_pg) {

    # When a job is created, add a starttime note
    $dbh->do(<<'SQLDATA');
      DROP TRIGGER IF EXISTS insert_starttimes ON ts_job;

      CREATE OR REPLACE FUNCTION trgfunc_insert_starttime()
        RETURNS trigger
        LANGUAGE 'plpgsql'

      AS $BODY$
        BEGIN
          INSERT INTO ts_note(jobid, notekey, value)
          VALUES (NEW.jobid, 'starttime', NEW.insert_time::text::bytea)
          ;

          RETURN NULL;
        END;
      $BODY$;

      CREATE TRIGGER insert_starttimes
        AFTER INSERT ON ts_job
        FOR EACH ROW
        EXECUTE FUNCTION trgfunc_insert_starttime()
      ;
SQLDATA

    # When the exit status expires, remove the starttime note
    $dbh->do(<<'SQLDATA');
      DROP TRIGGER IF EXISTS clean_starttimes ON ts_exitstatus;

      CREATE OR REPLACE FUNCTION trgfunc_delete_starttimes()
        RETURNS trigger
        LANGUAGE 'plpgsql'

      AS $BODY$
        BEGIN
          DELETE FROM ts_note
          WHERE notekey = 'starttime'
          AND jobid = OLD.jobid
          ;

          RETURN NULL;
        END;
      $BODY$;

      CREATE TRIGGER clean_starttimes
        AFTER DELETE ON ts_exitstatus
        FOR EACH ROW
        EXECUTE FUNCTION trgfunc_delete_starttimes()
      ;
SQLDATA

  }
  ## REDHAT EXTENSION END 1142506

  ## REDHAT EXTENSION START 2139001
  $dbh->bz_start_transaction();
  $dbh->bz_add_column('bugs', 'locked',
    {TYPE => 'BOOLEAN', NOTNULL => 1, DEFAULT => 'FALSE'});
  $dbh->bz_commit_transaction();
  ## REDHAT EXTENSION END 2139001

  return;
}

sub redhat_install_before_final_checks {
  my ($self, $args) = @_;

  my $dbh = Bugzilla->dbh;

  ## REDHAT EXTENSION START 465941
  add_setting('private_comment_default', ['on', 'off'], 'off');
  ## REDHAT EXTENSION END 465941

  ## REDHAT EXTENSION START 212492
  add_setting('warn_duplicate', ['on', 'off'], 'on');
  ## REDHAT EXTENSION END 212492

  ## REDHAT EXTENSION START 554957
  add_setting('merge_bug_data', ['Always', 'Never', 'Prompt'], 'Never');
  ## REDHAT EXTENSION END 554957

  ## REDHAT EXTENSION START 904164
  add_setting('product_in_subject', ['on', 'off'], 'off');
  ## REDHAT EXTENSION END 904164

  add_setting('view_customer_cases_on_sfdc', ['as_sfdc', 'as_portal'],
    'as_portal');

  my $count = $dbh->selectrow_array(
    q{SELECT count(*)
          FROM setting_value
        WHERE name = 'view_customer_cases_on_sfdc'
          AND value = 'as_ascension'}
  );

  if ($count) {
    add_setting('view_customer_cases_on_sfdc', ['as_sfdc', 'as_portal'],
      'as_portal', undef, 1);
  }

  if (!$dbh->bz_column_info('bugs', 'cf_deadline')) {
    my $field = Bugzilla::Field->new({name => 'cf_deadline'});
    if ($field) {
      $dbh->do("DELETE FROM fielddefs WHERE name = 'cf_deadline'");
    }
  }

  ## REDHAT EXTENSION START 1267048 1434222
  add_setting('headers_in_body', ['on', 'off'],                    'off');
  add_setting('status_colours',  ['off', 'pastel', 'colorbrewer'], 'off');
  add_setting('def_table_size',  [25, 50, 100, 500, -1],           25);
  ## REDHAT EXTENSION END 1267048 1434222
  add_setting('def_buglist_size', [20, 50, 100, 200], 20);

  ## REDHAT EXTENSION START 2149539
  my $max = Bugzilla->params->{'max_search_results'};
  $count = $dbh->selectrow_array(
    q{
        SELECT count(*)
          FROM setting_value
        WHERE name = 'def_buglist_size'
          AND value = ?
      }, undef, $max
  );

  if ($count == 0) {
    add_setting('def_buglist_size', Bugzilla->buglist_limits, 20, undef, 1);
  }
  ## REDHAT EXTENSION END 2149539

  # Has to happen after extensions modify schema.
  _update_custom_fields();

  ## REDHAT EXTENSION START 1241769
  $dbh->bz_add_index('bugs', 'bugs_cf_clone_of_idx', ['cf_clone_of']);
  ## REDHAT EXTENSION END 1241769

  ## REDHAT EXTENSION START 1371333
  _add_doc_types();
  ## REDHAT EXTENSION END 1371333

  $count
    = $dbh->selectrow_array(
    q{SELECT count(*) FROM bugs_activity WHERE id = 28889734 AND removed = '2018-09-21 20:42:33'}
    );

  if ($count) {
    say "Adjusting bugs_activity entries";
    $count = 0;
    my $SQL
      = q{select bug_id, added, removed, id from bugs_activity where fieldid = 6965 and removed <> '' order by bug_id, bug_when};

    my $sth = $dbh->prepare($SQL);

    $SQL = q{UPDATE bugs_activity SET removed = ? WHERE id = ?};
    my $update_sth = $dbh->prepare($SQL);

    $sth->execute();

    my $prev_bug_id = 0;
    my $prev_added  = '';
    while (my ($bug_id, $added, $removed, $id) = $sth->fetchrow_array()) {
      if ($prev_bug_id && ($bug_id == $prev_bug_id) && ($removed ne $added)) {
        say "Bug: $bug_id, record: $id changing '$removed' to '$prev_added'";
        $update_sth->execute($prev_added, $id);
      }
      $prev_bug_id = $bug_id;
      $prev_added  = $added;
      $count++;
    }

    say "Adjusted $count entries";
  }

## REDHAT EXTENSION START 1708003
  my $eb_group = Bugzilla::Group->new({name => 'editbugs'});

  if ($eb_group->category && $eb_group->category ne 'Admin') {

    my $com_group    = Bugzilla::Group->new({name => 'community'});
    my $bz_group     = Bugzilla::Group->new({name => 'team Bugzilla'});
    my $nobody_group = Bugzilla::Group->new({name => 'nobody'});

    unless ($com_group) {
      $com_group = Bugzilla::Group->create({
        name        => "community",
        description => "Anyone associated with any community in Red Hat Bugzilla",
        isbuggroup  => 1,
        isactive    => 1,
        secure_mail => 0,
        category    => 'Public',
      });
    }

    unless ($nobody_group) {
      $nobody_group = Bugzilla::Group->create({
        name        => "nobody",
        description =>
          "A group to stop users accessing things. This group should have no direct or indirect members.",
        isbuggroup  => 1,
        isactive    => 1,
        secure_mail => 0,
        category    => 'Public',
      });
    }

    # move current groups on edit bugs to community
    eval {
      $dbh->do(q{UPDATE group_group_map SET grantor_id = ? WHERE grantor_id = 7},
        undef, $com_group->id);
      $dbh->do(q{UPDATE group_group_map SET member_id = ? WHERE member_id = 7},
        undef, $com_group->id);
    };

    # add admin/bot groups to editbugs
    $eb_group->set_category('Admin');
    $eb_group->update();

    my $sth_insert = $dbh->prepare(
      'INSERT INTO group_group_map
      (member_id, grantor_id, grant_type) VALUES (?, 7, 0)'
    );

    # admin, expunge, edit_all_bugs
    my @grp_ids = qw/6 502 319/;
    eval {
      foreach my $id (@grp_ids) {
        $sth_insert->execute($id);
      }
    };

    # Setup editbugs groups on Red Hat products
    my $rh_group = Bugzilla::Group->new({name => 'redhat'});

    foreach my $prod (Bugzilla::Product->get_all) {

      # Red Hat
      if ($prod->classification->name =~ m/(?:Red Hat|JBoss|Internal)/) {
        $prod->set_group_controls($rh_group, {editbugs => 1,});
      }
      elsif ($prod->classification->name =~ m/(?:Retired)/) {
        $prod->set_group_controls($nobody_group, {editbugs => 1,});
      }
      elsif ($prod->name eq 'Bugzilla') {
        $prod->set_group_controls($bz_group, {editbugs => 1,});
      }
      elsif ($prod->classification->name =~ m/(?:Workflows)/) {
        ## nop
      }
      else {
        $prod->set_group_controls($com_group, {editbugs => 1,});
      }

      $prod->update;
    }
  }
## REDHAT EXTENSION END 1708003

## REDHAT EXTENSION START 1729338
  my $ap_group  = Bugzilla::Group->new({name => 'all_partners'});
  my $class_obj = Bugzilla::Classification->new({name => 'Red Hat'});

  if ($ap_group && $class_obj) {
    $count = $dbh->selectrow_array(
      q{select * from group_control_map where product_id = 370 and editbugs = 1 and group_id = ?},
      undef, $ap_group->id
    );

    if ($ap_group && $class_obj && !$count) {
      $ap_group->set_is_active(1);
      $ap_group->update();

      foreach my $prod (@{$class_obj->products}) {

        # Red Hat
        if ($prod->is_active) {
          $prod->set_group_controls($ap_group, {editbugs => 1,});
          $prod->update;
        }
      }
    }
  }
## REDHAT EXTENSION END 1729338

## REDHAT EXTENSION START 1708041
  $dbh->do(
    q{DELETE FROM team_edit_groups WHERE group_id = 161 AND team_id IN (select id from agile_team where name like 'sst_%')}
  );

  $dbh->do(
    q{UPDATE agile_role SET can_edit_team = 1 WHERE name in ('Product Manager', 'Team Lead')}
  );
## REDHAT EXTENSION END 1708041

## REDHAT EXTENSION START 1257476
  my $sql = <<EOS;
  UPDATE fielddefs
  SET type = ?
  WHERE name IN (
    'bug_agile_pool.pool_id',
    'bug_agile_pool.pool_order',
    'docs_contact',
    'percentage_complete',
    'reporter_accessible',
    'rh_rule',
    'votes'
  )
EOS

  $dbh->do($sql, undef, FIELD_TYPE_INTEGER);

  $sql = <<EOS;
  UPDATE fielddefs
  SET type = ?
  WHERE name IN (
    'bug_id'
  )
EOS

  $dbh->do($sql, undef, FIELD_TYPE_BUG_ID);

  $sql = <<EOS;
  UPDATE fielddefs
  SET type = ?
  WHERE name IN (
    'creation_ts',
    'delta_ts'
  )
EOS

  $dbh->do($sql, undef, FIELD_TYPE_DATETIME);
## REDHAT EXTENSION END 1257476

  $sql = <<EOS;
  UPDATE fielddefs
  SET type = ?
  WHERE name IN (
    'ext_bz_bug_map.ext_bz_bug_id'
  )
EOS

  $dbh->do($sql, undef, FIELD_TYPE_FREETEXT);

  my $iu_group = Bugzilla::Group->new({name => 'invalid_account'});
  my $at_group = Bugzilla::Group->new({name => 'automaton'});
  unless ($iu_group && $at_group) {
    unless ($iu_group) {
      $iu_group = Bugzilla::Group->create({
        name        => "invalid_account",
        description => "Red Hat accounts that fail validation",
        isbuggroup  => 1,
        isactive    => 1,
        secure_mail => 0,
        category    => 'Red Hat',
      });
    }

    unless ($at_group) {
      $at_group = Bugzilla::Group->create({
        name        => "automaton",
        description => "Red Hat List and Bot accounts",
        isbuggroup  => 1,
        isactive    => 1,
        secure_mail => 0,
        category    => 'Red Hat',
      });
    }
  }

## REDHAT EXTENSION START 1821023
  my $spam_prod_name = Bugzilla->params->{'spam_product'};
  my $spam_prod      = Bugzilla::Product->new({name => $spam_prod_name});

  my $nobody_group_name = Bugzilla->params->{'nobody_group'};

  my $nobody_group = Bugzilla::Group->new({name => $nobody_group_name});

  unless ($spam_prod) {
    $dbh->bz_start_transaction();

    my $spam_comp_name  = Bugzilla->params->{'spam_component'};
    my $spam_class_name = Bugzilla->params->{'spam_classification'};

    my $nobody_user_name = Bugzilla->params->{'nobody_user'};

    my $nob_usr = Bugzilla::User->new({name => $nobody_user_name});
    unless ($nob_usr) {
      Bugzilla::User->create({login_name => $nobody_user_name, cryptpassword => '*'});
    }

    my $classif = Bugzilla::Classification->new({name => $spam_class_name});

    unless ($classif) {
      $classif = Bugzilla::Classification->create({
        name        => "Retired",
        description =>
          "Products that have reached end of life, or were never released.",
      });
    }

    unless ($nobody_group) {
      $nobody_group = Bugzilla::Group->create({
        name        => $nobody_group_name,
        description =>
          "A group to stop users accessing things. This group should have no direct or indirect members.",
        isbuggroup  => 1,
        isactive    => 1,
        secure_mail => 1,
        category    => 'Public',
      });
    }

    $spam_prod = Bugzilla::Product->create({
      classification     => $classif->name,
      name               => $spam_prod_name,
      description        => "Sinkhole for spam",
      isactive           => 1,
      allows_unconfirmed => 0,
      version            => '1.0',
    });

    # setting entry limits who can create bugs
    $spam_prod->set_group_controls(
      $nobody_group,
      {
        entry         => 1,
        editbugs      => 1,
        canedit       => 1,
        membercontrol => CONTROLMAPMANDATORY,
        othercontrol  => CONTROLMAPMANDATORY
      }
    );

    $spam_prod->set_always_private(1);
    $spam_prod->update();

    Bugzilla::Component->create({
      product      => $spam_prod,
      name         => $spam_comp_name,
      description  => "A can to keep spam in.",
      initialowner => $nobody_user_name,
    });
    $dbh->bz_commit_transaction();
  }
## REDHAT EXTENSION END 1821023

## REDHAT EXTENSION START 2094134
  unless ($spam_prod->group_controls->{$nobody_group->id}->{canedit} == 1) {
    $dbh->bz_start_transaction();
    $spam_prod->set_is_active(1);
    $spam_prod->set_group_controls(
      $nobody_group,
      {
        entry         => 1,
        editbugs      => 1,
        canedit       => 1,
        membercontrol => CONTROLMAPMANDATORY,
        othercontrol  => CONTROLMAPMANDATORY
      }
    );
    $spam_prod->update();
    $dbh->bz_commit_transaction();
  }
## REDHAT EXTENSION END 2094134

  if (!defined Bugzilla::Field->new({name => "longdescs.extra_groups"})) {
    Bugzilla::Field->create({
      name        => "longdescs.extra_groups",
      description => "Extra Private Groups",
      is_numeric  => 0,
      buglist     => 0,
      mailhead    => 0,
    });
  }

  my $bot_group = Bugzilla::Group->new({name => 'internal_bot'});
  unless ($bot_group) {
    $dbh->bz_start_transaction();
    $bot_group = Bugzilla::Group->create({
      name        => 'internal_bot',
      description => "Bot accounts for automation for internal users",
      isbuggroup  => 1,
      isactive    => 0,
      secure_mail => 0,
      category    => 'Red Hat',
      userregexp  => Bugzilla->params->{bot_name_regex},
      icon_url    => "fas fa-robot",
    });

    $dbh->do(
      'INSERT INTO group_group_map
        (member_id, grantor_id, grant_type)
        VALUES (?, ?, ?)', undef, $bot_group->id, $at_group->id, GROUP_MEMBERSHIP
    );
    $dbh->bz_commit_transaction();
  }

  $sql = <<'SQL';
SELECT count (*)
FROM workflow_groupreq
WHERE regex = '@(bot\.bugzilla\.)redhat\.com$'
SQL

  my $res = $dbh->selectall_arrayref($sql);
  $count = $res->[0]->[0];
  if ($count ne '0') {
    $dbh->do(<<'SQL');
UPDATE workflow_groupreq SET regex = '@(?:bot\.bugzilla\.)*redhat\.com$'
WHERE regex = '@(bot\.bugzilla\.)redhat\.com$'
SQL
  }

  my $ext_bot_group = Bugzilla::Group->new({name => 'external_bot'});
  unless ($ext_bot_group) {
    $dbh->bz_start_transaction();
    $at_group->set_category('Public');
    $at_group->set_description('Bot and List accounts');
    $at_group->set_is_active(0);
    $at_group->update();

    $ext_bot_group = Bugzilla::Group->create({
      name        => 'external_bot',
      description => "Bot accounts for automation for external users",
      isbuggroup  => 1,
      isactive    => 0,
      secure_mail => 0,
      category    => 'Public',
      icon_url    => "fas fa-robot",
    });

    $dbh->do(
      'INSERT INTO group_group_map
        (member_id, grantor_id, grant_type)
        VALUES (?, ?, ?)', undef, $ext_bot_group->id, $at_group->id, GROUP_MEMBERSHIP
    );
    $dbh->bz_commit_transaction();
  }


  ## REDHAT EXTENSION START 2130031
  $count = $dbh->selectrow_array(
    q{SELECT count(*)
          FROM fielddefs
         WHERE name = 'docs_contact'
           AND type = 10}
  );

  if ($count) {
    $dbh->do(q{update fielddefs set type = 0 where name = 'docs_contact'});
  }
  ## REDHAT EXTENSION END 2130031

  ## REDHAT EXTENSION START 2141441
  my $field = Bugzilla::Field->new({name => 'cf_last_closed'});
  if ($field && !$field->buglist) {
    $field->set_buglist(1);
    $field->update();
  }
  ## REDHAT EXTENSION END 2141441

  if (!defined Bugzilla::Field->new({name => "locked"})) {
    Bugzilla::Field->create({
      name        => "locked",
      description => "Locked",
      enter_bug   => 0,
      is_numeric  => 1,
      buglist     => 1,
      mailhead    => 1,
      long_desc   =>
        q{A bug that has been locked can only be updated by users with 'editbugs' rights on the bug's product.},
      sortkey => 64,
    });
  }

  ## REDHAT EXTENSION START 2186036
  if (lc(Bugzilla->localconfig->{db_driver}) ne 'sqlite') {
    my $bz_product = Bugzilla::Product->new({name => 'Bugzilla', cache => 1});

    my $ops_comp
      = Bugzilla::Component->new({product => $bz_product, name => 'Operations'});

    unless ($ops_comp) {
      $dbh->bz_start_transaction();

      $ops_comp = Bugzilla::Component->create({
        product     => $bz_product,
        name        => 'Operations',
        description =>
          "This component is for operational issues. If you need your account modified or there is an issue you think a Site-admin Sys-admin would need to action, this is the place to open a bug.",
        initialowner => 'bugbot@bot.bugzilla.redhat.com',
      });

      $dbh->bz_commit_transaction();
    }

    my %ops_subcomponents = (
      'Accounts' =>
        'If your account requires modification or you have issues logging in.',
      'Bug Templates' => 'Requests to edit and create bug description templates.',
      'General'       =>
        q{If there isn't a more suitable sub-component then use this sub-component.},
      'SPAM or Abuse' => 'Issues realted to SPAM or Abusive behaviour',
    );

    foreach my $sub_comp_name (keys %ops_subcomponents) {
      my $sub_comps
        = Bugzilla::Extension::SubComponents::RHSubComponent->match({
        component_id => $ops_comp->id, name => $sub_comp_name,
        });

      unless (scalar(@$sub_comps)) {
        $dbh->bz_start_transaction();

        Bugzilla::Extension::SubComponents::RHSubComponent->create({
          name               => $sub_comp_name,
          component          => $ops_comp,
          description        => $ops_subcomponents{$sub_comp_name},
          initialowner       => 'bugbot@bot.bugzilla.redhat.com',
          default_to_private => 1,
        });

        $dbh->bz_commit_transaction();
      }
    }
  }
  ## REDHAT EXTENSION END 2186036

  return;
}

sub _update_status_and_resolution {
  my $dbh = Bugzilla->dbh;

  # Create the redhat specific statuses and resolutions and remove unused
  my $field = Bugzilla::Field->new({name => 'bug_status'});

  # ATM this detects if we have the statuses set-up as we want
  my $test = Bugzilla::Field::Choice->type($field)->new({name => 'ASSIGNED'});
  return () if ($test);

  foreach my $new_value (
    qw/NEW ASSIGNED POST MODIFIED ON_DEV ON_QA VERIFIED RELEASE_PENDING CLOSED/)
  {
    my $value = Bugzilla::Field::Choice->type($field)->new({name => $new_value});
    next if $value;
    print "Creating value $new_value for field bug_status\n";
    Bugzilla::Field::Choice->type($field)->create({value => $new_value});
  }
  $dbh->do("DELETE FROM bug_status WHERE value = 'UNCONFIRMED'");
  foreach my $old_value (qw/REOPENED/) {
    my $value = Bugzilla::Field::Choice->type($field)->new({name => $old_value});
    next if !$value;
    print "Removing value $old_value for field bug_status\n";
    $value->remove_from_db();
  }

  $field = Bugzilla::Field->new({name => 'resolution'});
  foreach my $new_value (
    qw/NOTABUG DEFERRED CURRENTRELEASE NEXTRELEASE RAWHIDE
    ERRATA UPSTREAM CANTFIX INSUFFICIENT_DATA/
    )
  {
    my $value = Bugzilla::Field::Choice->type($field)->new({name => $new_value});
    next if $value;
    print "Creating value $new_value for field resolution\n";
    Bugzilla::Field::Choice->type($field)->create({value => $new_value});
  }
  foreach my $old_value (qw/INVALID/) {
    my $value = Bugzilla::Field::Choice->type($field)->new({name => $old_value});
    next if !$value;
    print "Removing value $old_value for field resolution\n";
    $value->remove_from_db();
  }

  # Update sort key for statuses
  my $count = 10;
  $field = Bugzilla::Field->new({name => 'bug_status'});
  foreach my $value (
    qw/NEW ASSIGNED POST MODIFIED ON_DEV ON_QA VERIFIED RELEASE_PENDING CLOSED/)
  {
    my $value = Bugzilla::Field::Choice->type($field)->new({name => $value});
    die("Nooooooo") if !$value;
    $value->set_sortkey($count);
    $value->update;
    $count += 10;
  }

  $dbh->do(q{UPDATE bug_status SET is_open = 0 WHERE value IN ('CLOSED')});

  # Initialize standard status workflow
  my @workflow = (
    [undef,             'NEW'],
    [undef,             'ASSIGNED'],
    ['ASSIGNED',        'NEW'],
    ['ASSIGNED',        'VERIFIED'],
    ['ASSIGNED',        'CLOSED'],
    ['ASSIGNED',        'MODIFIED'],
    ['ASSIGNED',        'ON_DEV'],
    ['ASSIGNED',        'ON_QA'],
    ['ASSIGNED',        'RELEASE_PENDING'],
    ['ASSIGNED',        'POST'],
    ['CLOSED',          'NEW'],
    ['CLOSED',          'ASSIGNED'],
    ['CLOSED',          'POST'],
    ['CLOSED',          'MODIFIED'],
    ['CLOSED',          'ON_QA'],
    ['MODIFIED',        'NEW'],
    ['MODIFIED',        'VERIFIED'],
    ['MODIFIED',        'ASSIGNED'],
    ['MODIFIED',        'CLOSED'],
    ['MODIFIED',        'ON_DEV'],
    ['MODIFIED',        'ON_QA'],
    ['MODIFIED',        'RELEASE_PENDING'],
    ['MODIFIED',        'POST'],
    ['NEW',             'VERIFIED'],
    ['NEW',             'ASSIGNED'],
    ['NEW',             'CLOSED'],
    ['NEW',             'MODIFIED'],
    ['NEW',             'ON_DEV'],
    ['NEW',             'ON_QA'],
    ['NEW',             'RELEASE_PENDING'],
    ['NEW',             'POST'],
    ['ON_DEV',          'NEW'],
    ['ON_DEV',          'VERIFIED'],
    ['ON_DEV',          'ASSIGNED'],
    ['ON_DEV',          'CLOSED'],
    ['ON_DEV',          'MODIFIED'],
    ['ON_DEV',          'ON_QA'],
    ['ON_DEV',          'RELEASE_PENDING'],
    ['ON_DEV',          'POST'],
    ['ON_QA',           'NEW'],
    ['ON_QA',           'VERIFIED'],
    ['ON_QA',           'ASSIGNED'],
    ['ON_QA',           'CLOSED'],
    ['ON_QA',           'MODIFIED'],
    ['ON_QA',           'ON_DEV'],
    ['ON_QA',           'RELEASE_PENDING'],
    ['ON_QA',           'POST'],
    ['POST',            'NEW'],
    ['POST',            'VERIFIED'],
    ['POST',            'ASSIGNED'],
    ['POST',            'CLOSED'],
    ['POST',            'MODIFIED'],
    ['POST',            'ON_DEV'],
    ['POST',            'ON_QA'],
    ['POST',            'RELEASE_PENDING'],
    ['RELEASE_PENDING', 'NEW'],
    ['RELEASE_PENDING', 'VERIFIED'],
    ['RELEASE_PENDING', 'ASSIGNED'],
    ['RELEASE_PENDING', 'CLOSED'],
    ['RELEASE_PENDING', 'MODIFIED'],
    ['RELEASE_PENDING', 'ON_DEV'],
    ['RELEASE_PENDING', 'ON_QA'],
    ['RELEASE_PENDING', 'POST'],
    ['VERIFIED',        'NEW'],
    ['VERIFIED',        'ASSIGNED'],
    ['VERIFIED',        'CLOSED'],
    ['VERIFIED',        'MODIFIED'],
    ['VERIFIED',        'ON_DEV'],
    ['VERIFIED',        'ON_QA'],
    ['VERIFIED',        'RELEASE_PENDING'],
    ['VERIFIED',        'POST'],
  );

  print "Removing old bug status transitions...\n";
  $dbh->do("DELETE FROM status_workflow");

  print
    "Now filling the 'status_workflow' table with valid bug status transitions for Red Hat...\n";
  my $sth_select = $dbh->prepare('SELECT id FROM bug_status WHERE value = ?');
  my $sth
    = $dbh->prepare(
    'INSERT INTO status_workflow (old_status, new_status, require_comment) VALUES (?, ?, 0)'
    );

  foreach my $transition (@workflow) {
    my ($from, $to);

    # If it's an initial state, there is no "old" value.
    $from = $dbh->selectrow_array($sth_select, undef, $transition->[0])
      if $transition->[0];
    $to = $dbh->selectrow_array($sth_select, undef, $transition->[1]);

    # If one of the bug statuses doesn't exist, the transition is invalid.
    next if (($transition->[0] && !$from) || !$to);
    $sth->execute($from, $to);
  }

  return;
}

sub _update_custom_fields {
  my $permitted = shift;

  my $dbh = Bugzilla->dbh;

  ## REDHAT EXTENSION START 917935
  # Add Red Hat specific custom fields
  my $file = bz_locations()->{extensionsdir} . '/RedHat/custom_fields.yaml';
  if (!-e $file) {
    die "Cannot open custom field file: $file";
  }

  my $fields = LoadFile($file);

  foreach my $field (@$fields) {
    next
      if ($permitted && @$permitted && none { $field->{name} eq $_ } @$permitted);

    # We need to use SQL to see if the field exists. Bugzilla::Fields
    # hides fields that are not visible to a non privileged user
    ## BUGBUG why is the upgrade not priveleged?
    my $old_field = Bugzilla::Field->new({name => $field->{name}});
    next if ($old_field);

    my $vgrps = delete $field->{view_groups}  // [];
    my $egrps = delete $field->{edit_groups}  // [];
    my $agrps = delete $field->{admin_groups} // [];

    ## REDHAT EXTENSION START 1672226
    ## BUGBUG this breaks the tests because it's based on live data
    ## and the Ids change for the fields and products.
    if (lc(Bugzilla->localconfig->{db_driver}) eq 'sqlite') {
      delete $field->{visibility_field_id};
      delete $field->{value_field_id};
      delete $field->{visibility_values};
    }
    ## REDHAT EXTENSION END 1672226

    ## REDHAT EXTENSION START 2215165
    ## Allow providing values in the YAML instead of hard coding them
    my $values          = delete $field->{values};
    ## These allow us to use names instead of IDs for relationships.
    ## Useful as it means we don't need a dump to map relationships.
    my $vis_field_name  = delete $field->{visibility_field_name};
    my $vis_value_names = delete $field->{visibility_value_names};
    ## REDHAT EXTENSION END 2215165

    my $new_field = Bugzilla::Field->create($field);

    if (lc(Bugzilla->localconfig->{db_driver}) ne 'sqlite') {
      $new_field->set_view_groups($vgrps);
      $new_field->set_edit_groups($egrps);
      $new_field->set_admin_groups($agrps);
    }

    ## REDHAT EXTENSION START 1672226
    if ($field->{name} eq 'cf_zstream_target_release') {
      my @vals = qw{
        8.0.0 8.1.0 8.2.0 8.3.0 8.4.0 8.5.0
        8.0.1 8.0.2 8.0.3 8.0.4 8.0.5
      };
      foreach my $val (@vals) {
        Bugzilla::Field::Choice->type($new_field)->create({value => $val});
      }
    }
    ## REDHAT EXTENSION END 1672226

    ## REDHAT EXTENSION START 1672734
    if ($field->{name} eq 'cf_epm_pri') {
      my @vals = (
        'Low', 'Medium', 'High', 'URGENT', 'VIP', 'BD', 'MVP', 'RUD (Only Highs)',
        'FastTrack'
      );
      my $count = 1;
      foreach my $val (@vals) {
        Bugzilla::Field::Choice->type($new_field)
          ->create({value => $val, sortkey => $count++});
      }
    }

    if ($field->{name} eq 'cf_epm_prf_state') {
      my @vals = (
        'PFR Submitted',
        'PFR Reviewed',
        'PFR Validated',
        'PFR RPL Ready',
        'PFR Release Assigned'
      );
      my $count = 1;
      foreach my $val (@vals) {
        Bugzilla::Field::Choice->type($new_field)
          ->create({value => $val, sortkey => $count++});
      }
    }
    ## REDHAT EXTENSION END 1672734

    ## REDHAT EXTENSION START 2215165
    if (defined $values && @{$values}) {
      my $count = 1;
      foreach my $val (@{$values}) {
        Bugzilla::Field::Choice->type($new_field)
          ->create({value => $val, sortkey => $count++});
      }
    }

    if (defined $vis_field_name && $vis_field_name ne '') {
      my $vis_names = my $vis_field
        = Bugzilla::Field->check({name => $vis_field_name});

      $new_field->set_visibility_field($vis_field->id);

      if (defined $vis_value_names && @{$vis_value_names}) {

        my @vis_ids;

        foreach my $val (@{$vis_value_names}) {

          my $choices = Bugzilla::Field::Choice->type($vis_field)->match({value => $val});

          if ($choices->[0]) {
            push(@vis_ids, $choices->[0]->id);
          }
        }

        $new_field->set_visibility_values(\@vis_ids) if (@vis_ids);
      }

      $new_field->update();
    }
    ## REDHAT EXTENSION END 2215165
  }
  ## REDHAT EXTENSION END 917935

  return;
}

sub _populate_releases_table {
  my $dbh = Bugzilla->dbh;

  # Adding a table for target releases to the database
  # If the releases table is empty, and we're still back in a Bugzilla
  # that has a bugs.product field, that means that we just created
  # the releases table and it needs to be populated.
  my $releases_exist = $dbh->selectrow_array("SELECT DISTINCT 1 FROM releases");
  if (!$releases_exist && $dbh->bz_column_info('bugs', 'product')) {
    print "Replacing blank releases...\n";

    # If we are upgrading from 2.8 or earlier, we will have *created*
    # the releases table with a product_id field, but Bugzilla expects
    # it to have a "product" field. So we change the field backward so
    # other code can run. The change will be reversed later in checksetup.
    if ($dbh->bz_column_info('releases', 'product_id')) {

      # Dropping the column leaves us with a releases_product_id_idx
      # index that is only on the "value" column. We need to drop the
      # whole index so that it can be correctly re-created later.
      $dbh->bz_drop_index('releases', 'releases_product_id_idx');
      $dbh->bz_drop_column('releases', 'product_id');
      $dbh->bz_add_column('releases', 'product',
        {TYPE => 'varchar(64)', NOTNULL => 1}, '');
    }

    # Populate the release table with all existing values in the database
    my $sth = $dbh->prepare(
      "SELECT DISTINCT r.value, b.product_id
                                   FROM bugs b, bugs_release r
                                   WHERE b.bug_id = r.bug_id"
    );
    $sth->execute();

    print "Populating releases table...\n";

    while (my ($value, $product) = $sth->fetchrow_array()) {

      # check if the value already exists
      my $sortkey = substr($value, 1);
      if ($sortkey !~ /^\d+$/) {
        $sortkey = 0;
      }
      else {
        $sortkey *= 10;
      }
      my $ms_exists = $dbh->selectrow_array(
        "SELECT value FROM releases
                  WHERE value = ? AND product = ?", undef, $value, $product
      );

      if (!$ms_exists) {
        $dbh->do(
          "INSERT INTO releases(value, product, sortkey)
                          VALUES (?,?,?)", undef, $value, $product, $sortkey
        );
      }
    }
  }

  return;
}

sub _add_products_defaultrelease {
  my $dbh = Bugzilla->dbh;

  # Added a defaultrelease field to the products table, so that
  # we know which release to initially assign bugs to.
  if (!$dbh->bz_column_info('products', 'defaultrelease')) {
    $dbh->bz_add_column('products', 'defaultrelease',
      {TYPE => 'varchar(64)', NOTNULL => 1, DEFAULT => "'---'"});
    my $sth = $dbh->prepare("SELECT id, defaultrelease FROM products");
    $sth->execute();
    while (my ($product_id, $default_ms) = $sth->fetchrow_array()) {
      my $exists = $dbh->selectrow_array(
        "SELECT value FROM releases
                  WHERE value = ? AND product_id = ?", undef, $default_ms, $product_id
      );
      if (!$exists) {
        $dbh->do("INSERT INTO releases (value, product_id) " . "VALUES (?, ?)",
          undef, $default_ms, $product_id);
      }
    }
  }

  return;
}

sub _add_doc_types {
  my $dbh = Bugzilla->dbh;

  my %types = (
    'If docs needed, set a value'        => 5,
    'Bug Fix'                            => 10,
    'Enhancement'                        => 20,
    'Rebase: Bug Fixes and Enhancements' => 30,
    'Rebase: Bug Fixes Only'             => 40,
    'Rebase: Enhancements Only'          => 50,
    'Known Issue'                        => 60,
    'Release Note'                       => 70,
    'Technology Preview'                 => 80,
    'Deprecated Functionality'           => 90,
    'Removed functionality '             => 95,
    'No Doc Update'                      => 100,
  );

  my %templates = (
    "Known Issue" =>
      "Cause: \n\nConsequence: \n\nWorkaround (if any): \n\nResult: ",
    "Rebase: Bug Fixes and Enhancements" =>
      "Important: if this rebase instead contains *only bug fixes,* or *only enhancements*, select the correct option from the Doc Type drop-down list.\n\nRebase package(s) to version: \n\nHighlights, important fixes, or notable enhancements: ",
    "Enhancement"            => "Feature: \n\nReason: \n\nResult: ",
    "Bug Fix"                => "Cause: \n\nConsequence: \n\nFix: \n\nResult: ",
    "Rebase: Bug Fixes Only" =>
      "Important: if this rebase also contains *enhancements* (or contains only enhancements), select the correct option from the Doc Type drop-down list.\n\nRebase package(s) to version: \n\nHighlights and important bug fixes: ",
    "Deprecated Functionality" =>
      "Deprecated functionality should describe removed or no longer supported features.",
    "Release Note" =>
      "A release note should describe in detail a new feature being introduced into the product.\n\nA release note is NOT a bug, errata enhancement, rebase, known issue, Technology Preview or deprecated functionality description; for these, select the corresponding Doc Type option.",
    "Technology Preview" =>
      "Package(s) providing the Technology Preview: \n\nDescription of the Technology Preview: ",
    "Rebase: Enhancements Only" =>
      "Important: if this rebase also contains *bug fixes* (or contains only bug fixes), select the correct option from the Doc Type drop-down list.\n\nRebase package(s) to version: \n\nHighlights and notable enhancements: ",
    "If docs needed, set a value" =>
      "If this bug requires documentation, please select an appropriate Doc Type value.",
    "No Doc Update"         => "",
    "Removed functionality" => "",

  );

  my $sth = $dbh->prepare(
    q{INSERT INTO cf_doc_type (value, sortkey,isactive) values (?, ?, 1)});

  my $sth1
    = $dbh->prepare(
    q{INSERT INTO doc_type_text (cf_doc_type_id, value) select cf_doc_type.id, ? FROM cf_doc_type where cf_doc_type.value = ?}
    );

  foreach my $key (keys(%templates)) {
    eval { $sth->execute($key, $types{$key}); };
    eval { $sth1->execute($templates{$key}, $key); };
  }

  return;
}

1;
