# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::RuleEngine::FlagGroup;

use strict;
use warnings;
use 5.10.1;

use Scalar::Util qw(blessed);

use Bugzilla::Constants;
use Bugzilla::Field;
use Bugzilla::Util;
use Bugzilla::Error;

use Bugzilla::Extension::RuleEngine::FlagGroupDetail;

use base qw(Bugzilla::Object);

###############################
####    Initialization     ####
###############################

use constant DB_TABLE   => 'rh_flag_group';
use constant LIST_ORDER => 'sortkey, name';

use constant DB_COLUMNS => qw(
  id
  name
  description
  current_detail_id
  sortkey
);

use constant UPDATE_COLUMNS => qw(
  name
  description
  current_detail_id
  sortkey
);

use constant VALIDATORS => {
  name              => \&_check_name,
  description       => \&_check_description,
  current_detail_id => \&_check_current_detail_id,
  sortkey           => \&_check_sortkey,
};

###############################
####      Validators       ####
###############################

sub _check_creation_ts {
  return Bugzilla->dbh->selectrow_array('SELECT LOCALTIMESTAMP(0)');
}

sub _check_current_detail_id {
  my ($invocant, $id) = @_;

  $id = trim($id);

  # Return blank if specified, or we are creating this flag group
  return '' if !$invocant || $id eq '';

  # Check that the detail belongs to this flag group
  my $object
    = Bugzilla::Extension::RuleEngine::FlagGroupDetail->check({id => $id});
  if ($object->flag_group_id != $invocant->id) {
    ThrowCodeError(
      'flag_group_id_mismatch',
      {
        other_id            => $object->id,
        other_flag_group_id => $object->rule_id,
        this_flag_group_id  => $invocant->id
      }
    );
  }
  return $object->id;
}

sub _check_description {
  my ($invocant, $description) = @_;

  $description = trim($description || '');
  return $description;
}

sub _check_name {
  my ($invocant, $name) = @_;

  $name = trim($name);
  $name || ThrowUserError('flag_group_not_specified');

  if (length($name) > 64) {
    ThrowUserError('flag_group_name_too_long', {'name' => $name});
  }

  my $rule = Bugzilla::Extension::RuleEngine::FlagGroup->new({name => $name});
  if ($rule && (!ref $invocant || $rule->id != $invocant->id)) {
    ThrowUserError("flag_group_already_exists", {name => $rule->name});
  }
  return $name;
}

sub _check_sortkey {
  my ($invocant, $sortkey) = @_;

  $sortkey ||= 0;
  my $stored_sortkey = $sortkey;
  if (!detaint_natural($sortkey) || $sortkey > MAX_SMALLINT) {
    ThrowUserError('flag_group_invalid_sortkey', {'sortkey' => $stored_sortkey});
  }
  return $sortkey;
}

###############################
####       Methods         ####
###############################

sub set_name              { $_[0]->set('name',              $_[1]); return; }
sub set_description       { $_[0]->set('description',       $_[1]); return; }
sub set_rule_group_id     { $_[0]->set('rule_group_id',     $_[1]); return; }
sub set_current_detail_id { $_[0]->set('current_detail_id', $_[1]); return; }
sub set_sortkey           { $_[0]->set('sortkey',           $_[1]); return; }

sub current {
  my $self = shift;
  my $dbh  = Bugzilla->dbh;

  if (!defined $self->{current_detail}) {
    $self->{current_detail} = Bugzilla::Extension::RuleEngine::FlagGroupDetail->new(
      $self->current_detail_id);
  }
  return $self->{current_detail};
}

sub all_details {
  my $self = shift;
  my $dbh  = Bugzilla->dbh;

  if (!$self->{'all_details'}) {
    my $ids = $dbh->selectcol_arrayref(
      q{
            SELECT id FROM rh_flag_group_detail
            WHERE flag_group_id = ?}, undef, $self->id
    );

    $self->{'all_details'}
      = Bugzilla::Extension::RuleEngine::FlagGroupDetail->new_from_list($ids);
  }
  return $self->{'all_details'};
}

###############################
####      Accessors        ####
###############################

sub description       { return $_[0]->{'description'}; }
sub creation_ts       { return $_[0]->{'creation_ts'}; }
sub rule_group_id     { return $_[0]->{'rule_group_id'}; }
sub current_detail_id { return $_[0]->{'current_detail_id'}; }
sub sortkey           { return $_[0]->{'sortkey'}; }

###############################
####       Helpers         ####
###############################

1;

__END__

=head1 NAME

Bugzilla::Extension::RuleEngine::FlagGroup - Bugzilla Rules Engine Flag Group class.

=head1 SYNOPSIS

    use Bugzilla::Extension::RuleEngine::FlagGroup;

    my $flag_group = new Bugzilla::Extension::RuleEngine::FlagGroup(1);
    my $flag_group = new Bugzilla::Extension::RuleEngine::FlagGroup({name => 'Acme'});

    my $id                = $flag_group->id;
    my $name              = $flag_group->name;
    my $description       = $flag_group->description;
    my $current_detail_id = $flag_group->current_detail_id;
    my $sortkey           = $flag_group->sortkey;

=head1 DESCRIPTION

Bugzilla::Extension::RuleEngine::FlagGroup represents a flag_group object. It
is an implementation of L<Bugzilla::Object>, and thus provides all methods
that L<Bugzilla::Object> provides.

The methods that are specific to L<Bugzilla::Extension::RuleEngine::FlagGroup> are listed
below.

A Flag group is either a regular expression or a list of flag types that is
part of the Bugzilla Rules Engine. A flag group is versioned with a specific
version called a flag group detail.

=head1 METHODS

=over

=item C<current>

=over

=item B<Description>

Returns the current version of the flag group.

=item B<Params>

none.

=item B<Returns>

A L<Bugzilla::Extension::RuleEngine::FlagGroupDetail> object.

=back

=item C<all_details>

=over

=item B<Description>

Returns all versions of the flag group.

=item B<Params>

none.

=item B<Returns>

An arrayref of L<Bugzilla::Extension::RuleEngine::FlagGroupDetail> objects.

=back

=back

=cut
