# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::RuleEngine::WebService::Util;

use 5.10.1;
use strict;
use warnings;

=head1 NAME

Bugzilla::Extension::RuleEngine::WebService::Util - AgileTools WS utility functions

=head1 SYNOPSIS

    use Bugzilla::Extension::RuleEngine::WebService::Util;

    my $rule_group = Bugzilla::Bugzilla::Extension::RuleEngine::RuleGroup->new($id);

    my $rule_group_hash = rule_group_to_hash($webservice, $rule_group);


=head1 DESCRIPTION

Utility functions for webservices

=cut

use constant PUBLIC_METHODS => qw(
  rule_group_to_hash
  group_to_hash
  user_to_hash
);

use base qw(Exporter);
our @EXPORT = qw(
  rule_group_to_hash
  group_to_hash
  user_to_hash
);


=head1 FUNCTIONS

=over

=item C<rule_group_to_hash($webservice, $rule_group)>

    Description: Converts a RuleGroup to a hash
    Params:      $webservice - WebService object
                 $rule_group - RuleGroup object
    Returns:     Hash ref

=cut

sub rule_group_to_hash {
  my ($ws, $rule_group) = @_;

  my $item = {
    name           => $ws->type('string',  $rule_group->name),
    is_active      => $ws->type('boolean', $rule_group->is_active),
    id             => $ws->type('int',     $rule_group->id),
    description    => $ws->type('string',  $rule_group->description),
    disabled_count => $ws->type('int',     $rule_group->disabled_count),
    user_name      => $ws->type('string',  $rule_group->user_name),
    rule_count     => $ws->type('int',     $rule_group->rule_count),
    view_groups    => [],
    edit_groups    => [],
    maintainers    => [],
  };

  $item->{classifications}
    = [map { $_->name } @{$rule_group->classifications}];

  $item->{products} = [map { $_->name } @{$rule_group->products}];

  $item->{view_groups}
    = [map { group_to_hash($ws, $_) } @{$rule_group->view_groups}]
    if ($rule_group->view_groups);
  $item->{edit_groups}
    = [map { group_to_hash($ws, $_) } @{$rule_group->edit_groups}]
    if ($rule_group->edit_groups);
  $item->{maintainers}
    = [map { user_to_hash($ws, $_) } @{$rule_group->maintainers}]
    if ($rule_group->maintainers);

  return $item;
}

=item C<group_to_hash($webservice, $group)>

    Description: Converts a Group to a hash
    Params:      $webservice - WebService object
                 $group - Group object
    Returns:     Hash ref

=cut

sub group_to_hash {
  my ($ws, $group) = @_;

  my $item = {
    id          => $ws->type('int',     $group->id),
    name        => $ws->type('string',  $group->name),
    description => $ws->type('string',  $group->description),
    direct      => $ws->type('boolean', $group->{direct}),
  };

  return $item;
}


=item C<user_to_hash($webservice, $rule_group_to_hash)>

    Description: Converts a User to a hash
    Params:      $webservice - WebService object
                 $user - User object
    Returns:     Hash ref

=cut

sub user_to_hash {
  my ($ws, $user) = @_;

  my $item = {
    id        => $ws->type('int',    $user->id),
    real_name => $ws->type('string', $user->name),
    name      => $ws->type('email',  $user->login),
    email     => $ws->type('email',  $user->email),
  };

  return $item;
}

1;

__END__

=back

