# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

# Unless otherwise noted, this is REDHAT EXTENSION 653315

package Bugzilla::Extension::SubComponents;

use strict;
use warnings;
use 5.10.1;
use base qw(Bugzilla::Extension);

use Bugzilla::Bug;
use Bugzilla::Config::Common qw(check_numeric);
use Bugzilla::Constants;
use Bugzilla::Error;
use Bugzilla::Extension::SubComponents::RHSubComponent;
use Bugzilla::Field;
use Bugzilla::Token;
use Bugzilla::Util qw(diff_arrays trick_taint trim);

use List::MoreUtils qw(any none);
use Scalar::Util qw(looks_like_number);

# We create a lot of functions in other modules
BEGIN {
  *Bugzilla::Bug::rh_sub_component_objs        = \&_rh_sub_component_objs;
  *Bugzilla::Bug::rh_sub_components            = \&_rh_sub_components;
  *Bugzilla::Bug::rh_sub_component_ids         = \&_rh_sub_component_ids;
  *Bugzilla::Component::get_sub_components     = \&_get_sub_components;
  *Bugzilla::Component::get_sub_component_list = \&_get_sub_component_list;
  *Bugzilla::Component::get_sub_component_list_tree
    = \&_get_sub_component_list_tree;
  *Bugzilla::Component::new_from_list       = \&_new_from_list;
  *Bugzilla::Component::sub_component_count = \&_sub_component_count;
  *Bugzilla::Product::has_sub_components    = \&_has_sub_components;
  *Bugzilla::User::sub_component_responsibilities
    = \&_get_sub_component_responsibilities;
  *Bugzilla::Component::full_name        = \&_full_name;
  *Bugzilla::Component::full_name_pretty = \&_full_name_pretty;
  *Bugzilla::Component::cc_list_string   = \&_cc_list_string;
  *Bugzilla::Component::type             = \&_type;
}

=head1 NAME

Bugzilla::SubComponents::Extension - Extension for supporting Sub Components

=head1 DESCRIPTION

This extension includes the code and hooks for the Red Hat customisation
that handles Sub Component. A single sub component is a
Bugzilla::Extension::SubComponents::RHSubComponent object.

=head1 METHODS

This only lists methods that map to other modules. Specifically, it excludes
hooks and internal functions

=cut

######################################
## Bugzilla::Component sub routines ##
######################################

sub _full_name {
  my $self = shift;
  return $self->name;
}

sub _full_name_pretty {
  my $self = shift;
  return $self->name;
}

sub _type {
  return 'component';
}

sub _cc_list_string {
  my $self = shift;

  my $str = '';
  $str = join(', ', map { $_->login } @{$self->initial_cc})
    if ($self->initial_cc);
  return ($str);
}

=over

=item _get_sub_components

Called as Bugzilla::Component->get_sub_components

Returns an arrayref of the top level sub components that the component has

=cut

sub _get_sub_components {
  my $self = shift;
  my $dbh  = Bugzilla->dbh;

  if (not defined $self->{sub_components}) {
    my $query = qq{
            SELECT id
            FROM rh_sub_components
            WHERE component_id = ? AND parent_id IS NULL
        };

    my $ids = $dbh->selectcol_arrayref($query, undef, $self->id);
    $self->{sub_components}
      = Bugzilla::Extension::SubComponents::RHSubComponent->new_from_list($ids);
  }

  return $self->{sub_components};
}

=item _get_sub_component_list

Called as Bugzilla::Component->get_sub_component_list

Returns an arrayref of the sub components that the component is ordered
by sort order and level. This is ideal for showing sub components in
a drop down list.

=cut

sub _get_sub_component_list {
  my $self = shift;
  my $dbh  = Bugzilla->dbh;

  if (not defined $self->{sub_components_list}) {
    my $list = $self->get_sub_components();
    if (scalar @$list) {
      $list = $self->get_sub_component_list_tree($list, 0);
    }
    $self->{sub_components_list} = $list;
  }

  return $self->{sub_components_list};
}

=item _get_sub_component_list_tree

Called as Bugzilla::Component->get_sub_component_list_tree

This is a helper function for _get_sub_component_list, and is called
recursively

=cut

sub _get_sub_component_list_tree {
  my $self  = shift;
  my $list  = shift;
  my $level = shift;

  if ($level > 50 || scalar(@$list) > 1000) {
    ThrowCodeError('sub_component_possible_loop', {sub_component => $list->[0]});
  }
  my @new_list = ();
  foreach my $sub_comp (@$list) {
    my $children
      = $self->get_sub_component_list_tree($sub_comp->children, $level + 1);
    push @new_list, $sub_comp, @$children;
  }

  return \@new_list;
}

=item _new_from_list

Called as Bugzilla::Component->new_from_list

Returns the same as the upstream call, but precomputes the sub_components
value for efficency gains

=cut

sub _new_from_list {
  my $self = shift;
  my $ids  = shift;
  my $dbh  = Bugzilla->dbh;

  my $list = Bugzilla::Object::new_from_list($self, $ids);

  return $list if scalar(@$ids) == 0;

  # Now we need to get all the sub components for the components
  # Doing it this way means we aren't going to make many
  # $self->get_sub_components calls
  my $query = q{
        SELECT component_id, id
        FROM rh_sub_components
        WHERE component_id IN (? } . (',?' x $#$ids) . q{)
           AND parent_id IS NULL
    };

  # Turn this into $sub_comps{<id} = [<sub_comp_id>, ...];
  my $rows      = $dbh->selectall_arrayref($query, undef, @$ids);
  my %sub_comps = ();
  foreach (@$rows) {
    push @{$sub_comps{$_->[0]}}, $_->[1];
  }

  # Go through each component
  foreach my $comp (@$list) {
    if (not defined $sub_comps{$comp->id}) {

      # We want to cache negative results
      $comp->{sub_components} = [];
    }
    else {
      my $sub_comps = $sub_comps{$comp->id};
      $comp->{sub_components}
        = Bugzilla::Extension::SubComponents::RHSubComponent->new_from_list(
        $sub_comps);
    }
  }

  return $list;
}

=item _sub_component_count

Called as Bugzilla::Component->sub_component_count

Returns the number of top level sub components that the component has

=cut

sub _sub_component_count {
  my $self = shift;
  my $dbh  = Bugzilla->dbh;

  if (not defined $self->{sub_component_count}) {
    my $query = qq{
            SELECT COUNT(*)
            FROM rh_sub_components
            WHERE component_id = ?
        };

    ($self->{sub_component_count})
      = $dbh->selectrow_array($query, undef, $self->id);
  }

  return $self->{sub_component_count};
}

####################################
## Bugzilla::Product sub routines ##
####################################

=item _has_sub_components

Called as Bugzilla::Product->has_sub_components

Returns the number of components that have sub components.

=cut

sub _has_sub_components {
  my $self = shift;
  my $dbh  = Bugzilla->dbh;

  if (not defined $self->{has_sub_components}) {
    my $query = qq{
            SELECT COUNT(DISTINCT component_id)
            FROM components c
                JOIN rh_sub_components s ON (c.id = s.component_id)
            WHERE c.product_id = ?
        };

    ($self->{has_sub_components}) = $dbh->selectrow_array($query, undef, $self->id);
    $self->{has_sub_components} ||= 0;
  }

  return $self->{has_sub_components};
}

####################################
##  Bugzilla::User sub routines   ##
####################################

=item _get_sub_component_responsibilities

Called as Bugzilla::User->sub_component_responsibilities

Returns a list of hashes of products and their components.
The components hash contains a parent component and a list of its
child sub components (not in tree order)

=cut

sub _get_sub_component_responsibilities {
  my $self = shift;
  my $dbh  = Bugzilla->dbh;

  return $self->{'sub_comp_resp'} if defined $self->{'sub_comp_resp'};
  return [] unless $self->id;

  my $sql = '
        SELECT
            c.product_id AS product_id,
            sc.component_id AS component_id,
            sc.id AS sub_component_id
        FROM rh_sub_components sc
        JOIN components c ON (sc.component_id = c.id)
        LEFT JOIN rh_sub_component_cc sccc
            ON (sc.id = sccc.rh_sub_component_id)
        WHERE sc.initialowner = ?
        OR sc.initialqacontact = ?
        OR sc.initialdocscontact = ?
        OR sccc.user_id = ?
    ';

  my $id   = $self->id;
  my $list = $dbh->selectall_arrayref($sql, {Slice => {}}, $id, $id, $id, $id);

  unless ($list) {
    return $self->{'sub_comp_resp'} = [];
  }

  my @prod_ids = map { $_->{'product_id'} } @$list;
  my $products = Bugzilla::Product->new_from_list(\@prod_ids);
  require Bugzilla::Component;

  my @comp_ids   = map { $_->{'component_id'} } @$list;
  my $components = Bugzilla::Component->new_from_list(\@comp_ids);

  my @sub_comp_ids = map { $_->{'sub_component_id'} } @$list;
  my $sub_components
    = Bugzilla::Extension::SubComponents::RHSubComponent->new_from_list(
    \@sub_comp_ids);

  my @component_list;
  foreach my $comp (@$components) {
    my @sub_components = grep { $_->component_id == $comp->id } @$sub_components;
    push @component_list, {component => $comp, sub_components => \@sub_components,};
  }

  my @product_list;
  foreach my $prod (@$products) {
    my @comp_list
      = grep { $_->{component}->product_id == $prod->id } @component_list;
    push @product_list, {product => $prod, components => \@comp_list,};
  }

  return \@product_list;
}


##################
## INSTALLATION ##
##################

sub db_schema_abstract_schema {
  my ($self, $args) = @_;

  my $schema = $args->{schema};

  $schema->{rh_sub_components} = {
    FIELDS => [
      id   => {TYPE => 'MEDIUMSERIAL', NOTNULL => 1, PRIMARYKEY => 1},
      name => {TYPE => 'VARCHAR(64)',  NOTNULL => 1},
      component_id => {
        TYPE       => 'INT3',
        NOTNULL    => 1,
        REFERENCES => {TABLE => 'components', COLUMN => 'id', DELETE => 'CASCADE'}
      },
      parent_id => {
        TYPE       => 'INT3',
        REFERENCES => {TABLE => 'rh_sub_components', COLUMN => 'id'}
      },
      initialowner => {
        TYPE       => 'INT3',
        NOTNULL    => 1,
        REFERENCES => {TABLE => 'profiles', COLUMN => 'userid'}
      },
      initialqacontact => {
        TYPE       => 'INT3',
        REFERENCES => {TABLE => 'profiles', COLUMN => 'userid', DELETE => 'SET NULL'}
      },
      initialdocscontact => {
        TYPE       => 'INT3',
        REFERENCES => {TABLE => 'profiles', COLUMN => 'userid', DELETE => 'SET NULL'}
      },
      description => {TYPE => 'MEDIUMTEXT', NOTNULL => 1},
      isactive  => {TYPE => 'BOOLEAN',      NOTNULL => 1, DEFAULT => 'TRUE'},
      sortkey   => {TYPE => 'INT2',         NOTNULL => 1, DEFAULT => 0},
      level     => {TYPE => 'INT2',         NOTNULL => 1, DEFAULT => 0},
      full_name => {TYPE => 'VARCHAR(255)', NOTNULL => 1, DEFAULT => "''"},
    ],
    INDEXES => [
      rh_sub_components_component_id_idx =>
        {FIELDS => [qw(component_id name parent_id)], TYPE => 'UNIQUE'},
      rh_sub_components_name_idx      => ['name'],
      rh_sub_components_full_name_idx => ['full_name'],
      rh_sub_component_comp_id_idx    => ['component_id'],
    ],
  };

  $schema->{rh_sub_component_cc} = {
    FIELDS => [
      user_id => {
        TYPE       => 'INT3',
        NOTNULL    => 1,
        REFERENCES => {TABLE => 'profiles', COLUMN => 'userid', DELETE => 'CASCADE'}
      },
      rh_sub_component_id => {
        TYPE    => 'INT3',
        NOTNULL => 1,
        REFERENCES =>
          {TABLE => 'rh_sub_components', COLUMN => 'id', DELETE => 'CASCADE'}
      },
    ],
    INDEXES => [
      rh_sub_component_cc_user_id_idx =>
        {FIELDS => [qw(rh_sub_component_id user_id)], TYPE => 'UNIQUE'},
      rh_sub_component_cc_comp_id_idx => ['rh_sub_component_id'],
    ],
  };

  $schema->{bug_rh_sub_components} = {
    FIELDS => [
      bug_id => {
        TYPE       => 'INT3',
        NOTNULL    => 1,
        REFERENCES => {TABLE => 'bugs', COLUMN => 'bug_id', DELETE => 'CASCADE'}
      },
      rh_sub_component_id => {
        TYPE    => 'INT3',
        NOTNULL => 1,
        REFERENCES =>
          {TABLE => 'rh_sub_components', COLUMN => 'id', DELETE => 'CASCADE'}
      },

    ],
    INDEXES => [
      rh_sub_components_bug_id_idx =>
        {FIELDS => [qw(bug_id rh_sub_component_id)], TYPE => 'UNIQUE'},
      rh_sub_components_rh_sub_component_id_idx => ['rh_sub_component_id'],
    ],
  };

  return;
}

sub install_update_db {
  my ($self, $args) = @_;
  my $dbh = Bugzilla->dbh;

  my $sub_comp_field = Bugzilla::Field->new({name => 'rh_sub_components'});
  if (!$sub_comp_field) {
    $sub_comp_field = Bugzilla::Field->create({
      name         => 'rh_sub_components',
      description  => 'Sub Component',
      mailhead     => 1,
      is_mandatory => 0,
      buglist      => 1,
    });
  }

  # Give the sub component field the same sortkey as the component.
  # Name is used as the second order field so Component will appear first
  my $comp_field = Bugzilla::Field->new({name => 'component'});
  if ($comp_field and $sub_comp_field) {
    $sub_comp_field->set_sortkey($comp_field->sortkey);
    $sub_comp_field->update;
  }

  ## REDHAT EXTENSION START 1067210
  $dbh->bz_add_index('rh_sub_components', 'rh_sub_component_comp_id_idx',
    ['component_id']);

  $dbh->bz_add_index('rh_sub_component_cc', 'rh_sub_component_cc_comp_id_idx',
    ['rh_sub_component_id']);
  ## REDHAT EXTENSION END 1067210

  return;
}

###########
## PAGES ##
###########

sub template_before_process {
  my ($self, $args) = @_;
  my $file = $args->{'file'};

  if (Bugzilla->params->{usesubcomponents} and $file eq 'search/form.html.tmpl') {
    _query_sub_components($args);
  }

  return;
}

sub page_before_template {
  my ($self, $args) = @_;
  return unless Bugzilla->params->{usesubcomponents};

  my $page = $args->{page_id};

  # If we aren't viewing the sub component page, we don't need to do anything
  return unless $page =~ m{^subcomponents/};

  # Make sure we have permissions to edit the component.
  my $user = Bugzilla->login(LOGIN_REQUIRED);

  $user->in_group('editcomponents')
    || scalar(@{$user->get_products_by_permission('editcomponents')})
    || ThrowUserError("auth_failure",
    {group => "editcomponents", action => "edit", object => "components"});

  my $vars = $args->{vars};
  my $cgi  = Bugzilla->cgi;

  # Check that the product and component have been specified
  my $product_name = $cgi->param('product')
    || ThrowUserError('required_field', {field => {name => 'Product'}});
  my $component_name = $cgi->param('component')
    || ThrowUserError('required_field', {field => {name => 'Component'}});

  # Check they are valid
  my $product = Bugzilla::Product->check({name => $product_name});
  my $component
    = Bugzilla::Component->check({product => $product, name => $component_name});

  # Stash the values way (component is a TT reserved word)
  $vars->{showbugcounts} = (defined $cgi->param('showbugcounts'));
  $vars->{product}       = $product;
  $vars->{comp}          = $component;

  if ($page eq 'subcomponents/list.html') {

    # List sub components for a specified component
    _page_list($vars);
  }
  elsif ($page eq 'subcomponents/edit.html') {
    _page_edit($vars);
  }
  elsif ($page eq 'subcomponents/del.html') {
    _page_confirm_delete($vars);
  }

  return;
}

sub _page_list {
  my ($vars) = @_;
  my $dbh    = Bugzilla->dbh;
  my $cgi    = Bugzilla->cgi;
  my $input  = Bugzilla->input_params;

  # often used variables
  my $token  = $input->{'token'};
  my $action = $input->{'action'};
  my $id     = $input->{'sub_component_id'};

  if ($action eq 'add') {
    check_token_data($token, 'add_sub_component');

    # Do the user matching
    Bugzilla::User::match_field({
      'initialowner'     => {'type' => 'single'},
      'initialqacontact' => {'type' => 'single'},
      ## REDHAT EXTENSION 876015: Add initialdocscontact
      'initialdocscontact' => {'type' => 'single'},
      'initialcc'          => {'type' => 'multi'},
    });

    my $name               = trim($cgi->param('name')             || '');
    my $description        = trim($cgi->param('description')      || '');
    my $default_assignee   = trim($cgi->param('initialowner')     || '');
    my $default_qa_contact = trim($cgi->param('initialqacontact') || '');
    ## REDHAT EXTENSION 876015: Add default_docs_contact
    my $default_docs_contact = trim($cgi->param('initialdocscontact') || '');
    my @initial_cc           = $cgi->param('initialcc');
    my $sortkey              = trim($cgi->param('sortkey') || 0);
    my $parent_id            = trim($cgi->param('parent_id') || '');
    my $default_pool         = $cgi->param('default_pool');                     ## RED HAT EXTENSION 1613297
    my $default_to_private   = $cgi->param('default_to_private');

    my $sub_component
      = Bugzilla::Extension::SubComponents::RHSubComponent->create({
      name             => $name,
      component        => $vars->{comp},
      description      => $description,
      initialowner     => $default_assignee,
      initialqacontact => $default_qa_contact,
      ## REDHAT EXTENSION 876015: Add initialdocscontact
      initialdocscontact => $default_docs_contact,
      initial_cc         => \@initial_cc,
      sortkey            => $sortkey,
      parent_id          => $parent_id,
      default_pool       => $default_pool,                       ## RED HAT EXTENSION 1613297
      agile_team         => scalar $cgi->param('agile_team'),    ## RED HAT EXTENSION 1622371
      default_to_private => $default_to_private,
      });

    delete_token($token);

    $vars->{message}       = 'sub_component_created';
    $vars->{sub_component} = $sub_component;
  }

  elsif ($action eq 'delete') {
    $id || ThrowCodeError('param_required', {param => 'sub_component_id'});

    my $sub_component
      = Bugzilla::Extension::SubComponents::RHSubComponent->new($id);

    check_token_data($token, 'delete_sub_component');
    $sub_component->remove_from_db();
    delete_token($token);

    $vars->{sub_component} = $sub_component;
    $vars->{message}       = 'sub_component_deleted';
  }

  elsif ($action eq 'update') {
    $id || ThrowCodeError('param_required', {param => 'sub_component_id'});

    my $sub_component
      = Bugzilla::Extension::SubComponents::RHSubComponent->new($id);

    check_token_data($token, 'edit_sub_component');

    # Do the user matching
    Bugzilla::User::match_field({
      'initialowner'     => {'type' => 'single'},
      'initialqacontact' => {'type' => 'single'},
      ## REDHAT EXTENSION 876015: Add initialdocscontact
      'initialdocscontact' => {'type' => 'single'},
      'initialcc'          => {'type' => 'multi'},
    });

    my $name               = trim($cgi->param('name')             || '');
    my $default_assignee   = trim($cgi->param('initialowner')     || '');
    my $default_qa_contact = trim($cgi->param('initialqacontact') || '');
    ## REDHAT EXTENSION 876015: Add default_docs_contact
    my $default_docs_contact = trim($cgi->param('initialdocscontact') || '');
    my $description          = trim($cgi->param('description')        || '');
    my @initial_cc           = $cgi->param('initialcc');
    my $isactive             = $cgi->param('isactive');
    my $default_to_private   = $cgi->param('default_to_private');
    my $sortkey              = trim($cgi->param('sortkey') || 0);
    my $parent_id            = $cgi->param('parent_id');
    my $default_pool = $cgi->param('default_pool');    ## RED HAT EXTENSION 1613297

    $sub_component->set_name($name);
    $sub_component->set_description($description);
    $sub_component->set_default_assignee($default_assignee);
    $sub_component->set_default_qa_contact($default_qa_contact);
    ## REDHAT EXTENSION 876015: Add default_docs_contact
    $sub_component->set_default_docs_contact($default_docs_contact);
    $sub_component->set_cc_list(\@initial_cc);
    $sub_component->set_is_active($isactive);
    $sub_component->set_default_to_private($default_to_private);
    $sub_component->set_sortkey($sortkey);
    $sub_component->set_parent_id($parent_id);
    $sub_component->set_default_pool($default_pool);                     ## RED HAT EXTENSION 1613297
    $sub_component->set_agile_team(scalar $cgi->param('agile_team'));    ## RED HAT EXTENSION 1622371

    my $changes = $sub_component->update();

    # If the parent id or name has changed, we may need to update other
    # sub components for the component. The easiest thing to do is to just
    # update them all.
    my $sub_components
      = $sub_component->match({component_id => $sub_component->component_id});
    foreach my $sub_comp (@$sub_components) {
      $sub_comp->update();
    }

    delete_token($token);

    $vars->{sub_component} = $sub_component;
    $vars->{message}       = 'sub_component_updated';
  }

  return;
}

sub _page_edit {
  my ($vars) = @_;
  my $dbh    = Bugzilla->dbh;
  my $input  = Bugzilla->input_params;

  my $action = $input->{'action'} || '';

  if ($action eq 'new') {
    $vars->{action} = 'add';
    $vars->{token}  = issue_session_token('add_sub_component');
  }
  else {
    my $id = $input->{'sub_component_id'};
    $id || ThrowCodeError('param_required', {param => 'sub_component_id'});

    $vars->{action} = 'update';
    $vars->{sub_component}
      = Bugzilla::Extension::SubComponents::RHSubComponent->check({id => $id});
    $vars->{initial_cc_names}
      = join(', ', map($_->login, @{$vars->{sub_component}->initial_cc}));
    $vars->{token} = issue_session_token('edit_sub_component');

    if ($vars->{comp}->id != $vars->{sub_component}->component_id) {
      ThrowUserError(
        'sub_component_incorrect_product',
        {
          sub_component => $vars->{sub_component},
          comp          => $vars->{comp},
          product       => $vars->{product}
        }
      );
    }
  }

  return;
}

sub _page_confirm_delete {
  my ($vars) = @_;
  my $dbh    = Bugzilla->dbh;
  my $input  = Bugzilla->input_params;

  my $id = $input->{'sub_component_id'};
  $id || ThrowCodeError('param_required', {param => 'sub_component_id'});

  $vars->{sub_component}
    = Bugzilla::Extension::SubComponents::RHSubComponent->new($id);
  $vars->{token} = issue_session_token('delete_sub_component');

  return;
}

############
## PANELS ##
############

sub config_modify_panels {
  my ($self, $args) = @_;

  my $panels = $args->{panels};

  push(
    @{$panels->{bugfields}{params}},
    {name => 'usesubcomponents', type => 'b', default => '1',},
    {
      name    => 'subcomponentspacing',
      type    => 't',
      default => '4',
      checker => \&check_numeric,
    }
  );

  return;
}

##############
## BUG SUBS ##
##############

=item _rh_sub_component_objs

Called as Bugzilla::Bug->rh_sub_component_objs

Returns an arrayref of sub component objects the bug has

=cut

sub _rh_sub_component_objs {
  my $self = shift;

  if (not defined $self->{rh_sub_component_objs}) {
    $self->{rh_sub_component_objs}
      = Bugzilla::Extension::SubComponents::RHSubComponent->new_from_list(
      $self->rh_sub_component_ids);
  }

  return $self->{rh_sub_component_objs};
}

=item _rh_sub_components

Called as Bugzilla::Bug->rh_sub_components

Returns an arrayref of sub component names (flattened, seperate by spaces)
the bug has

=cut

sub _rh_sub_components {
  my $self = shift;

  if (not defined $self->{rh_sub_components}) {
    $self->{rh_sub_components} = [];
    foreach my $sub_comp (@{$self->rh_sub_component_objs}) {
      push @{$self->{rh_sub_components}}, join(' ', @{$sub_comp->name_tree});
    }
  }

  return $self->{rh_sub_components};
}

=item _rh_sub_component_ids

Called as Bugzilla::Bug->rh_sub_component_ids

Returns an arrayref of sub component ids the bug has

=cut

sub _rh_sub_component_ids {
  my $self = shift;
  my $dbh  = Bugzilla->dbh;

  if (not defined $self->{rh_sub_component_ids}) {
    my $query = qq{
            SELECT rh_sub_component_id
            FROM bug_rh_sub_components
            WHERE bug_id = ?
        };

    $self->{rh_sub_component_ids}
      = $dbh->selectcol_arrayref($query, undef, $self->id);
  }

  return $self->{rh_sub_component_ids};
}

###############
## BUG HOOKS ##
###############

sub bug_fields {
  my ($self, $args) = @_;
  my $fields = $args->{fields};
  push(@$fields, 'rh_sub_components');

  return;
}

sub post_bug_fields {
  my ($self, $args) = @_;

  my $bug_fields = $args->{bug_fields};
  my $input      = Bugzilla->input_params;

  push @$bug_fields, grep {/^rh_sub_component$/} keys %$input;

  return;
}

sub bug_before_create {
  my ($self, $args) = @_;

  # No sub components? No problems :)
  return if not Bugzilla->params->{usesubcomponents};

  my $params = $args->{params};

  # params->{rh_sub_components} is what we return, so we need to capture
  # the value passed in from RPC (if used) first
  my $xml_sub_comps = delete $params->{rh_sub_components};

  # Set the value we will return
  $params->{rh_sub_components} = [];

  # params->{product} from the UI is an object, from the RPC it is a string.
  my $product
    = ref $params->{product}
    ? $params->{product}
    : Bugzilla::Product->check({name => $params->{product}});

  # This is called after the ExtraValues extensions, so we need to get
  # all components
  my @component_names = ($params->{component});
  push @component_names, @{$params->{'cf_extra_component'}}
    if defined $params->{'cf_extra_component'};

  if (Bugzilla->usage_mode != USAGE_MODE_BROWSER) {

    # This section deals with RPC calls
    if (defined $xml_sub_comps) {

      if (ref $xml_sub_comps ne 'HASH') {
        ThrowUserError('subcomponent_not_a_hash', {ref_type => ref $xml_sub_comps});
      }

      foreach my $component_name (keys %$xml_sub_comps) {
        my $value = $xml_sub_comps->{$component_name};

        # Silent ignore if this component is not selected
        next unless any { lc $_ eq lc $component_name } @component_names;

        # If we are given an array, then join the names together
        $value = join ' ', @$value if ref $value;

        # If we are given a number, it is an id, not a name
        my $field = $value =~ /^\d+$/ ? 'id' : 'full_name';

        # Get the component object, since we need the id
        my $component
          = Bugzilla::Component->new({product => $product, name => $component_name});

        # Get the sub component object
        my $rh_sub_components
          = Bugzilla::Extension::SubComponents::RHSubComponent->match(
          {component_id => $component->id, $field => $value});
        if (scalar(@$rh_sub_components) == 0) {
          ThrowUserError('subcomponent_does_not_exist',
            {id => $value, comp => $component});
        }

        # ... and matches only one sub component
        elsif (scalar(@$rh_sub_components) > 1) {
          ThrowUserError('subcomponent_not_unique', {id => $value, comp => $component});
        }

        # ... and is active
        elsif (!$rh_sub_components->[0]->is_active) {
          ThrowUserError('subcomponent_not_active', {id => $value, comp => $component});
        }

        # Seems all is good, add this sub component to the list
        # of sub components for this bug
        push @{$params->{rh_sub_components}}, $rh_sub_components->[0];
      }

      if (!@{$params->{rh_sub_components}}) {
        ThrowUserError('subcomponent_not_valid');
      }
    }
  }
  else {
    # And this section deals with the cgi
    foreach my $key (grep {/^rh_sub_component$/} keys %$params) {
      my $value = delete $params->{$key};

      next unless defined $value and length($value) > 0;

      # Check that the component is actually selected
      my $component
        = Bugzilla::Component->new({product => $product, name => $component_names[0]
        });
      next unless any { lc $_ eq lc $component->name } @component_names;
      my ($component_id) = $component->id();

      # Check that the sub component is valid ...
      my $rh_sub_component;

      if (looks_like_number($value)) {
        $rh_sub_component
          = Bugzilla::Extension::SubComponents::RHSubComponent->new($value);
      }
      else {
        my $sub_comps
          = Bugzilla::Extension::SubComponents::RHSubComponent->match({
          component_id => $component->id, name => $value,
          });

        if (scalar(@$sub_comps)) {
          $rh_sub_component = $sub_comps->[0];
        }
      }

      if (not $rh_sub_component) {
        ThrowUserError('subcomponent_does_not_exist',
          {id => $value, comp => $component});
      }

      # ... and belongs to this component
      elsif ($rh_sub_component->component_id != $component_id) {
        ThrowUserError('subcomponent_not_correct',
          {rh_sub_component => $rh_sub_component, comp => $component});
      }
      elsif (!$rh_sub_component->is_active) {
        ThrowUserError('subcomponent_not_active', {id => $value, comp => $component});
      }

      push @{$params->{rh_sub_components}}, $rh_sub_component;
    }
  }

  if ( (not defined $params->{rh_sub_components})
    || ($#{$params->{rh_sub_components}} == -1))
  {
    my $component = Bugzilla::Component->new(
      {product => $product, name => $component_names[0]});
    if ($component && $component->sub_component_count) {
      ThrowUserError('subcomponent_mandatory',
        {comp => $component, product => $product});
    }
  }

  return;
}

sub bug_before_set_all {
  my ($self, $args) = @_;

  # No sub components? No problems :)
  return if not Bugzilla->params->{usesubcomponents};

  my $bug    = $args->{bug};
  my $params = $args->{params};
  my $input  = Bugzilla->input_params;
  my $dbh    = Bugzilla->dbh;

  # This is the arrayref we use to record the sub components of this bug
  # By the end, it will be defined one way or another
  my $sub_comps = undef;

  # Get a list of existing sub components. This is to check that inactive
  # components can remain selected
  my $existing_sub_comps = $bug->rh_sub_component_ids;

  # This is called after the ExtraValues extensions, but before we have
  # actually made any changes
  my $product = $bug->product_obj;
  if ($params->{product}) {
    $product = Bugzilla::Product->new({name => $params->{product}});
  }

  my @component_names = @{$bug->all_components};
  if ($params->{component}) {
    @component_names = ($params->{component});
    if (defined $params->{cf_extra_component}) {
      push @component_names, @{$params->{cf_extra_component}};
    }
  }

  my $privs;
  if ($bug->check_can_change_field('rh_sub_components', 0, 1, \$privs)) {
    if (
      Bugzilla->usage_mode == USAGE_MODE_BROWSER
      && (defined $input->{"defined_rh_sub_component"}
        && $input->{"defined_rh_sub_component"})
      )
    {
      # If the user is using a web browser, then we know they will
      # always send the values (providing they can edit them)
      $sub_comps = [];
## The old way had multiple select boxes, one for each component with a sub-component
## The new way has one select and updates it via JSONRPC
## Not entirely sure if this loop is required, maybe for a product with multiple components and subcomponents?
## We don't have anything using that ATM so it's not testable, but I'll leave this here in case that eventuates
## and we need this loop. Don't have time to test due to performance upgrade focus.
##            foreach my $key (@sub_comp_keys) {
      my $component = Bugzilla::Component->new(
        {product => $product, name => $component_names[0]});
      my ($component_id) = $component->id();
      my $value = $input->{"rh_sub_component"};

      # If they didn't select a value, we don't need to do any checks.
      next if !$value;

      # Check that the component is actually selected
      next unless any { lc $_ eq lc $component->name } @component_names;

      # If they have requested we don't change it
      if ($input->{dontchange} and $value eq $input->{dontchange}) {

        # Get the value (if any) from the the existing list
        push @$sub_comps,
          grep { $_->component_id == $component_id } @$existing_sub_comps;
        next;
      }

      # Check that the sub component is valid ...
      my $rh_sub_component
        = Bugzilla::Extension::SubComponents::RHSubComponent->new($value);
      if (not $rh_sub_component) {
        ThrowUserError('subcomponent_does_not_exist',
          {id => $value, comp => $component});
      }

      # ... and belongs to this component
      elsif ($rh_sub_component->component_id != $component_id) {
        ThrowUserError('subcomponent_not_correct',
          {rh_sub_component => $rh_sub_component, comp => $component});
      }
      elsif (
        !$rh_sub_component->is_active && none { $rh_sub_component->id == $_ }
        @$existing_sub_comps
        )
      {
        ThrowUserError('subcomponent_not_active', {id => $value, comp => $component});
      }

      push @$sub_comps, $rh_sub_component;
##            }
    }
    elsif (my $xml_sub_comps = $params->{rh_sub_components}) {

      # Any RPC call is made
      $sub_comps = [];

      # They are making an RPC call, and have supplied a value.
      # This overrides ALL sub components on a bug.
      if (ref $xml_sub_comps ne 'HASH') {
        ThrowUserError('subcomponent_not_a_hash', {ref_type => ref $xml_sub_comps});
      }

      foreach my $component_name (keys %$xml_sub_comps) {
        my $value = $xml_sub_comps->{$component_name};

        # Silent ignore if this component is not selected
        next unless any { lc $_ eq lc $component_name } @component_names;

        # If we are given an array, then join the names together
        $value = join ' ', @$value if ref $value;

        # If we are given a number, it is an id, not a name
        my $field = $value =~ /^\d+$/ ? 'id' : 'full_name';

        # Get the component object, since we need the id
        my $component
          = Bugzilla::Component->new({product => $product, name => $component_name});

        # Get the sub component object
        my $rh_sub_components
          = Bugzilla::Extension::SubComponents::RHSubComponent->match(
          {component_id => $component->id, $field => $value});

        # Check that the sub component is valid
        if (scalar(@$rh_sub_components) == 0) {
          ThrowUserError('subcomponent_does_not_exist',
            {id => $value, comp => $component});
        }

        # ... and matches only one sub component
        elsif (scalar(@$rh_sub_components) > 1) {
          ThrowUserError('subcomponent_not_unique', {id => $value, comp => $component});
        }

        # ... and is active or is currently selected
        elsif (!$rh_sub_components->[0]->is_active
          && none { $rh_sub_components->[0]->id == $_ } @$existing_sub_comps)
        {
          ThrowUserError('subcomponent_not_active', {id => $value, comp => $component});
        }

        # Seems all is good, add this sub component to the list
        # of sub components for this bug
        push @$sub_comps, $rh_sub_components->[0];
      }

      if (!@$sub_comps) {
        ThrowUserError('subcomponent_not_valid');
      }
    }

    # else they are using RPC, and have not supplied a value, use current ones
  }
  else {
    # They don't have permission to change the values.
    # Lets throw an error if they try
    if ($input->{rh_sub_components} or $input->{rh_sub_component}) {
      ThrowUserError('illegal_change',
        {field => 'rh_sub_components', privs => $privs});
    }
  }

  if (not defined $sub_comps) {

    # Either the user doesn't have permissions to change sub components
    # or an RPC call has been made without passing the sub_components
    # value. We need to check that the current sub components are valid
    $sub_comps = [];
    foreach my $sub_comp (@{$bug->rh_sub_component_objs}) {
      my $component = $sub_comp->component;
      if (($product->name eq $component->product->name)
        && any { $_ eq $component->name } @component_names)
      {
        push @$sub_comps, $sub_comp;
      }
    }
  }

  $bug->{rh_sub_component_objs} = $sub_comps;
  $bug->{rh_sub_component_ids}  = [map { $_->id } @$sub_comps];

  # Now some tricky magic if the reset user flags are given
  if (scalar(@$sub_comps)) {
    if (delete $params->{reset_assigned_to}) {
      $params->{assigned_to} = $sub_comps->[0]->default_assignee;
    }
    if (delete $params->{reset_qa_contact}) {
      $params->{qa_contact} = $sub_comps->[0]->default_qa_contact;
    }
    if (delete $params->{reset_docs_contact}) {
      $params->{docs_contact} = $sub_comps->[0]->default_docs_contact;
    }
  }

  return;
}

sub bug_start_of_update {
  my ($self, $args) = @_;

  my $bug       = $args->{bug};
  my $old_bug   = $args->{old_bug};
  my $timestamp = $args->{timestamp};
  my $changes   = $args->{changes};
  my $dbh       = Bugzilla->dbh;

  # Now we get the difference between the old and new
  my ($removed_sub_comps, $added_sub_comps)
    = diff_arrays($old_bug->rh_sub_component_objs,
    $bug->rh_sub_component_objs, 'id');

  # Leave if there is no change
  return unless scalar(@$removed_sub_comps) or scalar(@$added_sub_comps);

  # Make the change
  foreach my $sub_comp (@$removed_sub_comps) {
    my $query
      = "DELETE FROM bug_rh_sub_components WHERE bug_id = ? AND rh_sub_component_id = ?";
    $dbh->do($query, undef, $bug->id, $sub_comp->id);
  }

  foreach my $sub_comp (@$added_sub_comps) {
    my $query
      = "INSERT INTO bug_rh_sub_components (bug_id, rh_sub_component_id) VALUES(?, ?)";
    $dbh->do($query, undef, $bug->id, $sub_comp->id);
  }

  # Log the change
  $changes->{rh_sub_components} = [
    join(' ', map { join(' ', @{$_->name_tree}) } @$removed_sub_comps),
    join(' ', map { join(' ', @{$_->name_tree}) } @$added_sub_comps),
  ];

  ## REDHAT EXTENSION START 1117645
  # If we are changing sub components, and they have default CC: addresses,
  # then add these users to the bug
  foreach my $sub_comp (@$added_sub_comps) {
    foreach my $cc (@{$sub_comp->initial_cc}) {
      $bug->add_cc($cc, 1);
    }
  }
  ## REDHAT EXTENSION END 1117645

  return;
}

sub bug_end_of_update {
  my ($self, $args) = @_;
  my $dbh  = Bugzilla->dbh;
  my $user = Bugzilla->user;
  my $bug  = $args->{bug};

  # if a component has sub-components, then one must be set
  # and it must be valid for the component
  if ( $user->login ne Bugzilla->params->{system_user}
    && $bug->isopened
    && $bug->component_obj->sub_component_count())
  {
    @{$bug->rh_sub_component_objs()}
      || ThrowUserError('subcomponent_mandatory',
      {comp => $bug->component_obj, product => $bug->product_obj});

    foreach my $sub_comp (@{$bug->rh_sub_component_objs()}) {
      if ($bug->component_obj->id != $sub_comp->component_id) {
        ThrowUserError(
          'sub_component_incorrect_product',
          {
            sub_component => $sub_comp,
            comp          => $bug->component_obj,
            product       => $bug->product_obj,
          }
        );
      }
    }
  }

  return;
}

sub bug_remove_from_db {
  my ($self, $args) = @_;

  my $bug_id = $args->{bug_id};
  my $dbh    = Bugzilla->dbh;

  my $query = "DELETE FROM bug_rh_sub_components WHERE bug_id = ?";
  $dbh->do($query, undef, $bug_id);

  return;
}

sub bug_prefetch {
  my ($self, $args) = @_;
  my $fields    = $args->{'fields'};
  my $in_clause = $args->{'in_clause'};

  ## REDHAT EXTENSION START 965089
  $fields->{sub_components} = {
    value => 'rh_sub_component_ids',
    sql   => qq{
            SELECT bug_id, rh_sub_component_id
            FROM bug_rh_sub_components
            WHERE bug_id $in_clause
            ORDER BY rh_sub_component_id
        },
  };

  return;
}

sub webservice {
  my ($self, $args) = @_;

  my $dispatch = $args->{dispatch};

  # Historically, the way to call this was SubComponent.method
  # The docs suggest that the way to do this is SubComponents.method (plural)
  # We'll accept both, to preserve backwards compatability for existing scripts
  # and to not confused people who read the documentation
  $dispatch->{'SubComponents'} = 'Bugzilla::Extension::SubComponents::WebService';
  $dispatch->{'SubComponent'}  = 'Bugzilla::Extension::SubComponents::WebService';

  return;
}

#################
## QUERY HOOKS ##
#################

sub buglist_columns {
  my ($self, $args) = @_;

  my $columns = $args->{'columns'};
  my $dbh     = Bugzilla->dbh;

  $columns->{rh_sub_components}{name} = Bugzilla->dbh->sql_group_concat(
    'DISTINCT map_rh_sub_component_names.full_name');

  return;
}

sub buglist_column_joins {
  my ($self, $args) = @_;
  my $joins = $args->{'column_joins'};

  $joins->{rh_sub_components} = {
    table   => 'bug_rh_sub_components',
    then_to => {
      as    => 'map_rh_sub_component_names',
      table => 'rh_sub_components',
      from  => 'map_rh_sub_components.rh_sub_component_id',
      to    => 'id',
    },
  };

  return;
}

sub _query_sub_components {

  # This is quite complex, since we need to merge all trees into a single tree
  # Hopefully the comments are going to help
  my $args = shift;
  my $vars = $args->{'vars'};
  my $dbh  = Bugzilla->dbh;

  # We start by getting all sub components
  my $query = q{
        SELECT s.id
        FROM rh_sub_components s
            JOIN components c      ON (s.component_id      = c.id)
            JOIN products   p      ON (c.product_id        = p.id)
            JOIN classifications l ON (p.classification_id = l.id)
        WHERE 1=1
    };

  ## REDHAT EXTENSION 1410695 only retrun this list if component is set.
  return
    unless (exists($vars->{default}{component})
    && $vars->{default}{component} ne '');

  # But restricted to the selected classifications, products or components
  my %pairs = (qw/classification l.name product p.name component c.name/);
  my @vars  = ();
  while (my ($key, $column) = each %pairs) {
    if (defined($vars->{default}{$key})) {
      my $values = $vars->{default}{$key};
      $query
        .= " AND $column IN (" . join(',', split(//, ('?' x scalar(@$values)))) . ')';
      push @vars, @$values;
    }
  }

  # Get all the matching sub components objects.
  trick_taint($_) foreach (@vars);
  my $ids = $dbh->selectcol_arrayref($query, undef, @vars);
  my $sub_components
    = Bugzilla::Extension::SubComponents::RHSubComponent->new_from_list($ids);

  sub _walk_tree {
    my $tree = shift;
    my $name = shift;
    my $this = shift @$name;
    if (scalar(@$name)) {
      if (not exists $tree->{uc $this}) {
        $tree->{uc $this} = {};
      }
      _walk_tree($tree->{uc $this}, $name);
    }
    elsif (not exists $tree->{uc $this}{_name}) {
      $tree->{uc $this}{_name} = $this;
    }

    return;
  }

  # We now turn this into a hash of hash of hash.... to represent the levels
  # eg, {STORAGE}{_name} = 'Storage' {STORAGE}{EXT2}{_name} = 'Ext2'
  my $tree = {};
  foreach my $sub_component (@$sub_components) {
    _walk_tree($tree, $sub_component->name_tree);
  }

  sub _walk_hash {
    my $array      = shift;
    my $inner_tree = shift;
    my $parents    = shift;

    foreach my $row (sort keys %$inner_tree) {
      my $name = delete $inner_tree->{$row}{_name};
      push @$array, [@$parents, $name];
      _walk_hash($array, $inner_tree->{$row}, [@$parents, $name]);
    }

    return;
  }

  # Finally, we flatten it to an array of arrays. Each array is a row to
  # display. Each inner array is the tree for the sub components.
  # e.g. [ [ Storage ], [ Storage, Ext2 ] ]
  my $array = [];
  _walk_hash($array, $tree, []);

  # Add this to the stash to present the user
  $vars->{sub_component_list} = $array;

  return;
}

sub search_operator_field_override {
  my ($self, $args) = @_;

  my $operators = $args->{'operators'};

  $operators->{rh_sub_components} = {
    _non_changed => sub { _rh_sub_components_nonchanged(@_) }
  };

  return;
}

sub _rh_sub_components_nonchanged {
  my ($invocant, $args) = @_;

  $args->{full_field} = "rh_sub_components.full_name";
  $invocant->_do_operator_function($args);
  $args->{term} = qq{
        bugs.bug_id IN (
            SELECT bug_rh_sub_components.bug_id
            FROM bug_rh_sub_components
                JOIN rh_sub_components ON (bug_rh_sub_components.rh_sub_component_id = rh_sub_components.id)
            WHERE $args->{term}
        )
    };

  return;
}

sub webservice_error_codes {
  my ($self, $args) = @_;

  my $error_map = $args->{error_map};

  $error_map->{sub_component_has_bugs} = 105;

  return;
}

sub sanitycheck_check {
  my ($self, $args) = @_;
  my $status = $args->{status};

  $status->('sub_component_check');

  my ($count) = Bugzilla->dbh->selectrow_array("
    SELECT count(b.bug_id)
      FROM bug_rh_sub_components AS bsc
        INNER JOIN bugs AS b ON b.bug_id = bsc.bug_id
        INNER JOIN rh_sub_components AS sc ON bsc.rh_sub_component_id = sc.id
      WHERE sc.component_id != b.component_id;
    ");
  if ($count) {
    $status->('sub_component_alert', undef, 'alert');
    $status->('sub_component_repair');
  }

  return;
}

sub sanitycheck_repair {
  my ($self, $args) = @_;
  return unless Bugzilla->cgi->param('sub_component_repair');

  my $status = $args->{'status'};
  my $dbh    = Bugzilla->dbh;
  $status->('sub_component_repairing');

  my $rows = $dbh->selectall_arrayref("
    SELECT b.bug_id, bsc.rh_sub_component_id
      FROM bug_rh_sub_components AS bsc
        INNER JOIN bugs AS b ON b.bug_id = bsc.bug_id
        INNER JOIN rh_sub_components AS sc ON bsc.rh_sub_component_id = sc.id
     WHERE sc.component_id != b.component_id
    ", {Slice => {}});
  foreach my $row (@$rows) {
    $dbh->do("
      DELETE
        FROM bug_rh_sub_components
       WHERE bug_id=? AND rh_sub_component_id=?
      ", undef, $row->{bug_id}, $row->{rh_sub_component_id});
  }

  return;
}

=back

=cut

__PACKAGE__->NAME;
