# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::SubComponents::WebService;

use strict;
use warnings;
use 5.10.1;
use base qw(Bugzilla::WebService);
use Bugzilla::Product;
use Bugzilla::Component;
use Bugzilla::User;
use Bugzilla::Constants;
use Bugzilla::Error;
use Bugzilla::Util qw(trim);
use Bugzilla::WebService::Util qw(validate);
use Bugzilla;

use Bugzilla::Extension::SubComponents::RHSubComponent;

use List::MoreUtils qw(any);

use constant PUBLIC_METHODS => qw(
  get
  update
  create
  delete
);

# function to return subcomponent information when passed either
# an array of subcomponent ids or an array of hashes containing
# subcomponent, component and product names
# $call = $rpc->call( 'SubComponent.get',
#                          {
#                               ids => [1,2],
#                               names => [{
#                                          subcomponent => 'testSubComponent1',
#                                          component => 'testingComponent1',
#                                          product => 'TestProduct1'
#                                       },
#                                       {
#                                           subcomponent => 'testSubComponent2',
#                                           component => 'testComponent1',
#                                           product => 'TestProduct'
#                                       }]
#                          }
#                   );

sub get {
  my ($self, $params) = validate(@_, 'ids', 'names');

  my $accessible_subcomponents = _get_accessible_subcomponents($params);

  my @subcomponents = map { {
    id                 => $self->type('int',     $_->id),
    name               => $self->type('string',  $_->name),
    component_id       => $self->type('int',     $_->component_id),
    component_name     => $self->type('string',  $_->component->name),
    product_id         => $self->type('int',     $_->component->product_id),
    product_name       => $self->type('string',  $_->component->product->name),
    parent_id          => $self->type('int',     $_->parent_id),
    description        => $self->type('string',  $_->description),
    is_active          => $self->type('boolean', $_->is_active),
    name_tree          => [map { $self->type('string', $_) } @{$_->name_tree}],
    default_assignee   => $self->type('string', $_->default_assignee->login),
    default_qa_contact => $self->type(
      'string', (defined $_->default_qa_contact ? $_->default_qa_contact->login : '')
    ),
    ## REDHAT EXTENSION START 876015
    default_docs_contact => $self->type(
      'string',
      (defined $_->default_docs_contact ? $_->default_docs_contact->login : '')
    ),

    #  REDHAT EXTENSION END 876015
    default_cc => [map $self->type('string', $_->login), @{$_->initial_cc}],
    level      => $self->type('int', $_->level),
    sort_key   => $self->type('int', $_->sortkey),
    ## REDHAT EXTENSION START 1650366
    agile_team => $self->type(
      'string', defined $_->agile_team_obj ? $_->agile_team_obj->name : ''
    ),
    ## REDHAT EXTENSION END 1650366

    ## REDHAT EXTENSION BEGIN 1937776
    default_to_private => $self->type('boolean', $_->default_to_private),
    ## REDHAT EXTENSION END 1937776

  } } @$accessible_subcomponents;

  return {subcomponents => \@subcomponents};
}

# function to update subcomponent information when passed either
# an array of subcomponent ids or an array of hashes containing
# subcomponent, component and product names
# $call = $rpc->call( 'SubComponent.update',
#                          {ids => [1,2],
#                               names => [{
#                                          subcomponent => 'testSubComponent1',
#                                          component => 'testingComponent',
#                                          product => 'TestProduct'
#                                       },
#                                       {
#                                           subcomponent => 'testSubComponent2',
#                                           component => 'testComponent',
#                                           product => 'TestProduct'
#                                       }]
#                               updates => {
#                                           default_assignee => 'testuser@redhat.com',
#                                          }
#                          }
#                   );

sub update {
  my ($self, $params) = @_;

  my $user = Bugzilla->login(LOGIN_REQUIRED);

  $user->in_group('editcomponents')
    || scalar(@{$user->get_products_by_permission('editcomponents')})
    || ThrowUserError("auth_failure",
    {group => "editcomponents", action => "edit", object => "components"});

  # Check to see if we can admin these subcomponents;
  my @accessible_subcomponents
    = grep { $user->check_can_admin_product($_->component->product->name) }
    @{_get_accessible_subcomponents($params)};

  # if the user tries to change the subcomponent name for several
  # subcomponents of the same component then throw an error
  my %unique_product_comps;
  foreach my $sub_component (@accessible_subcomponents) {
    if ($unique_product_comps{$sub_component->id}
      && (any { $_ =~ /name/ } keys %{$params->{updates}}))
    {
      ThrowUserError("multiple_components_update_not_allowed");
    }
    else {
      $unique_product_comps{$sub_component->id} = 1;
    }
  }

  my %mapped_returns = (
    initialowner     => 'default_assignee',
    initialqacontact => 'default_qa_contact',
    ## REDHAT EXTENSION 876015: Add initialdocscontact
    initialdocscontact => 'default_docs_contact',
    cc_list            => 'default_cc',
  );

  my %changes;

  foreach my $sub_component (@accessible_subcomponents) {
    foreach my $field (keys %{$params->{updates}}) {
      my $method = "set_" . $field;
      if ($field eq "default_cc") {
        $method = "set_cc_list";
      }
      elsif ($field eq 'sort_key') {
        $method = 'set_sortkey';
      }

      $sub_component->$method($params->{updates}{$field});
    }
  }

  # map the field names in the returned changes hash to the
  # right field name for API consistency and convert
  # the user ids in the changes hash to login names
  foreach my $subcomponent_obj (@accessible_subcomponents) {
    my $returned_changes = $subcomponent_obj->update();
    foreach my $changed_field (keys %{$returned_changes}) {
      my $mapped_field = $mapped_returns{$changed_field} || $changed_field;
      ## REDHAT EXTENSION 876015: Add default_docs_contact
      if ( $mapped_field eq 'default_assignee'
        || $mapped_field eq 'default_qa_contact'
        || $mapped_field eq 'default_docs_contact'
        || $mapped_field eq 'default_cc')
      {
        my $user_list    = $returned_changes->{$changed_field};
        my @old_userlist = split(/[,\s]+/, $user_list->[0]);
        my @new_userlist = split(/[,\s]+/, $user_list->[1]);

        my $old_userlist = join(', ',
          (map $_->{login_name}, @{Bugzilla::User->new_from_list(\@old_userlist)}));
        my $new_userlist = join(', ',
          (map $_->{login_name}, @{Bugzilla::User->new_from_list(\@new_userlist)}));

        $changes{$subcomponent_obj->id}{$mapped_field} = [$old_userlist, $new_userlist];
      }
      else {
        $changes{$subcomponent_obj->id}{$mapped_field}
          = $returned_changes->{$changed_field};
      }
    }
  }

  # If the parent id or name has changed, we may need to update other
  # sub components for the component. The easiest thing to do is to just
  # update them all.
  foreach my $sub_component (@accessible_subcomponents) {
    my $sub_components
      = $sub_component->match({component_id => $sub_component->component_id});
    foreach my $sub_comp (@$sub_components) {
      $sub_comp->update();
    }
  }

  return {subcomponents_updates => \%changes};
}

# function to create a subcomponent
# the function can be called like this with the
# subcomponent, product, initialowner and description as required arguments
# $call = $rpc->call('SubComponent.create',
#                       {
#                           subcomponent=> 'TestingSubComponent',
#                           component => 'TestComponent'
#                           product => 'TestProduct',
#                           description => 'testing component',
#                           default_assignee => 'testuser@redhat.com',
#                           default_qa_contact => 'admin@redhat.com',
#                           default_cc => ['testuserb@redhat.com, 'testuserc@redhat.com']
#                       });
#
sub create {
  my ($self, $params) = @_;

  my $user = Bugzilla->login(LOGIN_REQUIRED);

  $user->in_group('editcomponents')
    || scalar(@{$user->get_products_by_permission('editcomponents')})
    || ThrowUserError("auth_failure",
    {group => "editcomponents", action => "edit", object => "components"});

  my $product_name       = trim($params->{product});
  my $component_name     = trim($params->{component});
  my $subcomponent_name  = trim($params->{subcomponent});
  my $default_assignee   = trim($params->{default_assignee});
  my $default_qa_contact = trim($params->{default_qa_contact});
  ## REDHAT EXTENSION 876015: Add default_docs_contact
  my $default_docs_contact = trim($params->{default_docs_contact});
  my $description          = trim($params->{description});
  my $default_cc           = $params->{default_cc};
  my $parent_id            = $params->{parent_id};

  my $product = $user->check_can_admin_product($product_name);

  my $component
    = Bugzilla::Component->check({product => $product, name => $component_name,});

  my $args = {
    name             => $subcomponent_name,
    component        => $component,            # Bugzilla::Component
    description      => $description,
    initialowner     => $default_assignee,
    initialqacontact => $default_qa_contact,
    ## REDHAT EXTENSION 876015: Add default_docs_contact
    initialdocscontact => $default_docs_contact,
    initial_cc         => $default_cc,
    parent_id          => $parent_id,
  };

  # If the user has supplied args for is_active, set that.
  if (defined($params->{is_active})) {
    $args->{isactive} = $params->{is_active} == 0 ? undef : 1;
  }

  ## REDHAT EXTENSION START 1937776
  if (defined($params->{default_to_private})) {
    $args->{default_to_private} = $params->{default_to_private} == 0 ? undef : 1;
  }
  ## REDHAT EXTENSION END 1937776

  # Set the sort key if the user has supplied it.
  $args->{sortkey} = $params->{sort_key} if defined $params->{sort_key};

  my $subcomponent
    = Bugzilla::Extension::SubComponents::RHSubComponent->create($args);

  return {id => $self->type('int', $subcomponent->id)};
}


# subroutine to hold repeated code between Component.get and
# Component.update, this function returns the unique accessible
# Component objects to the user from list of subcomponent ids and
# Component/product names.
sub _get_accessible_subcomponents {
  my $params        = shift;
  my @subcomponents = ();

  if (defined $params->{ids}) {
    push @subcomponents,
      @{Bugzilla::Extension::SubComponents::RHSubComponent->new_from_list(
        $params->{ids}
      )
      };
  }


  if (defined $params->{names}) {

    # to get the subcomponent objects for product/component combination
    # first obtain the product object from the passed product name
    foreach my $name_hash (@{$params->{names}}) {
      push @subcomponents,
        @{Bugzilla::Extension::SubComponents::RHSubComponent->match({
          component_id => Bugzilla::Component->check({
            name    => $name_hash->{component},
            product => Bugzilla::Product->check($name_hash->{product})
          })->id,
          name => $name_hash->{subcomponent}
        })
        };
    }
  }

  my %seen_component_ids = ();
  my %accessible_product_ids
    = map { $_->id => 1 } @{Bugzilla->user->get_accessible_products};

  my @accessible_subcomponents;
  foreach my $subcomponent (@subcomponents) {

    # Skip if we already included this subcomponent
    next if $seen_component_ids{$subcomponent->id}++;

    # Can the user see this product?
    next unless $accessible_product_ids{$subcomponent->component->product_id};

    push @accessible_subcomponents, $subcomponent;
  }

  return \@accessible_subcomponents;
}

sub delete {
  my ($self, $params) = @_;

  my $user = Bugzilla->login(LOGIN_REQUIRED);

  $user->in_group('editcomponents')
    || scalar(@{$user->get_products_by_permission('editcomponents')})
    || ThrowUserError("auth_failure",
    {group => "editcomponents", action => "edit", object => "components"});

  # Check to see if we can admin these subcomponents;
  my @accessible_subcomponents
    = grep { $user->check_can_admin_product($_->component->product->name) }
    @{_get_accessible_subcomponents($params)};

  foreach my $subcomponent (@accessible_subcomponents) {

    # Run through the list of subcomponents and make sure each can be deleted.
    if (!Bugzilla->params->{'allowbugdeletion'} && $subcomponent->bug_count) {
      ThrowUserError('sub_component_has_bugs',
        {nb => $subcomponent->bug_count, name => $subcomponent->name,});
    }

    if (scalar(@{$subcomponent->children})) {
      ThrowUserError('sub_component_has_children', {sub_component => $subcomponent,});
    }
  }

  my @deleted_ids;

  foreach my $subcomponent (@accessible_subcomponents) {
    push @deleted_ids, $subcomponent->id;
    $subcomponent->remove_from_db;
  }

  return {deleted => \@deleted_ids};
}

1;

__END__

=head1 NAME

Bugzilla::Extension::SubComponents::WebService - The SubComponent API

=head1 DESCRIPTION

This part of the Bugzilla API allows you to get info about accessible subcomponents,
create new subcomponents and also edit the attributes of existing subcomponents.

=head1 METHODS

See L<Bugzilla::WebService> for a description of what B<STABLE>, B<UNSTABLE>,
and B<EXPERIMENTAL> mean, and for more information about error codes.

=head2 Get SubComponents

=over

=item C<get> B<EXPERIMENTAL>

=over

=item B<Description>

Returns information about L<Bugzilla::SubComponent|SubComponents> in accessible
L<Bugzilla::Component|Components>.

You may retreive a list of all sub componets for a component using the Product.get
method. Documentation is available in L<Bugzilla::WebService::Product>

=item B<Params>

At least one of the following two parameters must be specified:

=over

=item C<ids> (array) - An array of integers, representing subcomponents ids.

=item C<names> (array) - An array of hashes, each hash has product, component and subcomponent name.

=back

=item B<Returns>

Information about all accessible product subcomponents passed. SubComponents
that do not exist or are not accessible are silently ignored.

The return value will be a hash containing one item, C<subcomponents>,
that is an array of hashes. Each hash describes a subcomponent,
and has the following items:

=over

=item id

C<int> The unique integer ID that Bugzilla uses to identify this subcomponent.
Even if the name of the subcomponent changes, this ID will stay the same.

=item name

C<string> The name of the subcomponent.

=item name_tree

C<array> An array of strings containing the names of all the sub components above this sub component.

=item description

C<string> The description of the subcomponent.

=item product_id

C<int> The id of the product that the subcomponent belongs to.

=item product_name

C<string> The name of the product that this component belongs to

=item component_id

C<int> The id of the component that this subcomponent belongs to

=item component_name

C<string> The name of the component that this subcomponent belongs to

=item default_assignee

C<string> The login name of the default owner of the subcomponent if
the user is logged in, if the user is not logged in then only the
realname of the default owner will be returned.

=item default_qa_contact

C<string> The login name of the default qa contact of the subcomponent if
the user is logged in, if the user is not logged in then only the
realname of the default qa contact will be returned.

=item default_docs_contact

C<string> The login name of the default docs contact of the subcomponent if
the user is logged in, if the user is not logged in then only the
realname of the default docs contact will be returned.

=item default_cc

C<array> The default CC list for this subcomponent. It is an array
of login names if the user is logged in. If the user is not logged in
then only the realnames of the default cc list members  will be returned.

=item level

C<int> How many levels deep is this subcomponent. 0 is a top level subcomponent.

=item sort_key

C<int> The sort key of this subcomponent.

=item is_active

C<boolean> True if the component is active.  False if it is inactive.

=item parent_id

C<int> B<optional> The ID of the parent subcomponent.  Will be absent if there is no parent.

=item agile_team

C<string> The name of the Agile Team bugs in this subcomponent will default to.

=back

=item B<Errors>

=over

=item 51 (Invalid Object)

A non existing product or subcomponent name was passed to the function, as a result no
product or subcomponent object existed for that invalid name.

=item 106 (Invalid Product)

You were required to specify a product, and either you didn't, or you
specified an invalid product (or a product that you can't access).

=back

=back

=back

=head2 Create SubComponent

=over

=item C<create> B<EXPERIMENTAL>

=over

=item B<Description>

Creates new subcomponent for an existing product.

=item B<Params>

A hash containing the following subcomponent fields:

=over

=item product

C<string> B<Required> The name of the product that the subcomponent
will be added to.

=item component

C<string> B<Required> The name of the component this subcomponent will be a child of

=item subcomponent

C<string> B<Required> The name of the new sub component.

=item description

C<string> B<Required> The description of the new subcomponent.

=item default_assignee

C<string> B<Required> The initial owner of the new subcomponent.

=item default_qa_contact

C<string> B<Optional> The initial qa contact for the new subcomponent.

=item default_docs_contact

C<string> B<Optional> The initial docs contact for the new subcomponent.

=item default_cc

C<array> B<Optional> Array of the initial cclist for the new subcomponent.

=item parent_id

C<int> B<Optional> If this subcomponent has a parent subcomponent, this will contain its id.

=item sort_key

C<int> B<Optional> The sort key for this subcomponent.

=item is_active

C<boolean> B<Optional> True to create the subcomponent active.  False to create the subcomponent inactive.

=item default_to_private

C<boolean> True if bugs for this subcomponent default to private.

=back

=item B<Returns>

A hash containing one item, C<id>, which is the id of the newly
created subcomponent

=item B<Errors>

=over

=item 304 (Authorization Required)
Logged-in user not authorized to create subcomponents.

=item 106 (Product Access Denied)
Logged-in user doesn't have access to the specified product to create
subcomponents for it.

=item 50 (Product Not Specified)
Logged-in user didn't specify any product names for the subcomponent they want
to create.

=item 106 (Product Doesn't Exist)
The specified product doesn't exist.

=item 105 (Component Name Is Too Long)
The name of the subcomponent is longer than 64 characters.

=item 105 (Component Requires Default Assignee)
User didn't specify a default assignee for the new subcomponent.

=item 105 (Blank Component Description Not Allowed)
User didn't specify a description for the new subcomponent.

=item 105 (Blank Component Name Not Allowed)
User didn't specifiy name for the new subcomponent.

=item 105 (Component Already Exists)
Trying to create a subcomponent which has similar name to an existing subcomponent.

=item 51 (Invalid Username)
Invalid login name was provided for the subcomponent's default assignee,
default qa contact, default docs contact or one of the default cc.

=back

=back

=back

=head2 Update SubComponents

=over

=item C<update> B<EXPERIMENTAL>

=over

=item B<Description>

Updates single or multiple subcomponents information.

=item B<Params>

A hash containing either C<ids>, which is an array of integers of the sub component ids.
Or C<names>, which is an array of hashes. Each hash has a product name, component name and subcomponent name.
And, C<updates> which is a hash containing the subcomponent fields that will be updated.
the update hash may contain any of the following items:

=over

=item name

C<string> The name of the subcomponent.

=item description

C<string> The description of the subcomponent.

=item default_assignee

C<string> The login name of the default owner of the subcomponent.

=item default_qa_contact

C<string> The login name of the default qa contact of the subcomponent.

=item default_docs_contact

C<string> The login name of the default docs contact of the subcomponent.

=item default_cc

C<array> A list of the login names in the default cc list of the subcomponent.

=item parent_id

C<int> B<optional> The parent id of this subcomponent

=item sort_key

C<int> B<Optional> The sort key for this sub component.

=item is_active

C<boolean> B<Optional> True to enable the subcomponent, False to disable the subcomponent.

=item default_to_private

C<boolean> True if bugs for this subcomponent default to private.

=back

=item B<Returns>

A hash containing one item, C<subcomponents_updates>, which is a hash of hashes,
with subcomponent ids as the hash keys pointing to hashes that contains
the changes that were made to the subcomponent information. The hash may contain
any of the following:

=over

=item name

C<array> An array of (strings) that holds the old and the new
names of the subcomponent.

=item description

C<array> An array of (strings) that holds the old and the new
descriptions of the subcomponent.

=item default_assignee

C<array> An Array of (strings) that holds the old and the new
login names of the default owner of the subcomponent.

=item default_qa_contact

C<array> An array of (strings) that holds the old and the new
login names of the default qa contact of the subcomponent.

=item default_docs_contact

C<array> An array of (strings) that holds the old and the new
login names of the default docs contact of the subcomponent.

=item default_cc

C<array> An array of (strings) that holds the old and the new
lists of the login names in the default cc list of the subcomponent.

=back

=item B<Errors>

=over

=item 50 (Invalid Product)

You were required to specify a product, and either you didn't, or you
specified an invalid product (or a product that you can't access).

=item 304 (Authorization Required)
Logged-in user not authorized to edit subcomponents.

=item 105 (Component Already Exists)
Trying to change subcomponent name to similar name of an existing subcomponent
of the same product.

=item 51 (Invalid Username)
Invalid login name was provided for the subcomponent's default assignee,
default qa contact, default docs contact or one of the default cc.

=item 105 (Component Name Is Too Long)
The name of the subcomponent is longer than 64 characters.

=item 105 (Multiple Components Update Not Allowed)
Can not do multiple subcomponents update when changing the name
of multiple subcomponents of the same product.

=back

=back

=back

=head2 Delete SubComponents

=over

=item C<delete> B<EXPERIMENTAL>

=over

=item B<Description>

Deletes one or more subcomponents.

=item B<Params>

A hash containing either C<ids>, which is an array of integers of the sub component ids you wish to delete.
Or C<names>, which is an array of hashes. Each hash has a product name, component name and subcomponent name.

If any of the subcomponents has bugs belonging to that sub component, the entire delete operation will fail.
That is to say, all subcomponents you wish to delete must have zero bugs.

=item B<Returns>

A hash containing one item, C<deleted> which contains an array of subcomponent ids that were deleted.

=back

=item B<Errors>

=over

=item 50 (Invalid Product)

You were required to specify a product, and either you didn't, or you
specified an invalid product (or a product that you can't access).

=item 304 (Authorization Required)
Logged-in user not authorized to edit subcomponents.

=item 105 (Component Already Exists)
Trying to change subcomponent name to similar name of an existing subcomponent
of the same product.

=item 105 (Sub Component Has Bugs)
Attempting to delete a sub component that has bugs in it.

=back

=back
