# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::Workflows;

use 5.10.1;
use strict;
use warnings;

use parent qw(Bugzilla::Extension);

use Bugzilla::Extension::Workflows::Util;
use Bugzilla::Extension::Workflows::GroupApproval;
use Bugzilla::Extension::Workflows::GroupRequest;
use Bugzilla::Extension::Workflows::ManageComponents;
use Bugzilla::Extension::Workflows::GroupReview;
use Bugzilla::Extension::Workflows::CreateBotAccount;
use Bugzilla::Extension::Workflows::APIKeyExpiry;
use Bugzilla::Extension::Workflows::GroupInvite;

use Bugzilla::Constants;
use Bugzilla::Error;
use Bugzilla::Token;
use Bugzilla::Extension::RuleEngine::Util qw(message_2_array);

our $VERSION = '0.01';

BEGIN {
  *Bugzilla::User::can_admin_group_workflow = \&can_admin_group_workflow;
  *Bugzilla::User::apply_for_groups         = \&apply_for_groups;
  *Bugzilla::User::days_since_login         = \&_user_days_since_login;
  *Bugzilla::User::get_review_bug           = \&_user_get_review_bug;
  *Bugzilla::User::create_review_bug        = \&_user_create_review_bug;
  *Bugzilla::User::get_all_review_bugs      = \&_user_get_all_review_bugs;

  *Bugzilla::Bug::age = \&_bug_age;

  *Bugzilla::Group::group_workflow = \&group_workflow;
  *Bugzilla::Group::can_apply_for  = \&can_apply_for;
  *Bugzilla::Group::can_approve    = \&can_approve;
  *Bugzilla::Group::can_invite     = \&can_invite;
  ## REDHAT EXTENSION BEGIN 1882298
  *Bugzilla::Group::admin_can_manage = \&admin_can_manage;
  ## REDHAT EXTENSION END 1882298
  *Bugzilla::Flag::user_can_edit     = \&_flag_user_can_edit;
  *Bugzilla::FlagType::user_can_edit = \&_flagtype_user_can_edit;

  *Bugzilla::Product::editors = \&_product_editors;

  ## REDHAT EXTENSION START 2148313
  *Bugzilla::User::contacts     = \&_user_contacts;
  *Bugzilla::User::admins       = \&_user_admins;
  *Bugzilla::User::admin_logins = \&_user_admin_logins;
  ## REDHAT EXTENSION END 2148313

}

# See the documentation of Bugzilla::Hook ("perldoc Bugzilla::Hook"
# in the bugzilla directory) for a list of all available hooks.
sub install_before_final_checks {
  my ($self, $args) = @_;

  my $dbh = Bugzilla->dbh;

  my $group = Bugzilla::Group->new({name => 'approvals_admin'});

  unless ($group) {
    $group = Bugzilla::Group->create({
      name        => "approvals_admin",
      description => "Admin group for approvals",
      isbuggroup  => 1,
      isactive    => 1,
      secure_mail => 1,
      category    => 'Admin',
    });
  }

  my $trusted_group = Bugzilla::Group->new({name => 'trusted_approvers'});

  unless ($trusted_group) {
    $trusted_group = Bugzilla::Group->create({
      name        => "trusted_approvers",
      description => "Users who are trusted and can bypass manager sign off",
      isbuggroup  => 1,
      isactive    => 0,
      secure_mail => 0,
      category    => 'Admin',
    });
  }

  my $classif = Bugzilla::Classification->new({name => 'Workflows'});

  unless ($classif) {
    $classif
      = Bugzilla::Classification->create({
      name => "Workflows", description => "User driven workflows",
      });
  }

  my $bvp
    = Bugzilla::Extension::BugViewPlus::Template->new({name => 'Group Request'});

  unless ($bvp) {
    $bvp = Bugzilla::Extension::BugViewPlus::Template->create({
      name        => 'Group Request',
      is_active   => 1,
      description => 'Group membership request text',
      content     =>
        "Please explain why you are requesting membership in these groups:\n\n",
    });
  }

  my $prod = Bugzilla::Product->new({name => 'Group Membership'});
  unless ($prod) {

    $dbh->bz_start_transaction();
    $prod = Bugzilla::Product->create({
      classification     => $classif->name,
      name               => "Group Membership",
      description        => "Workflow to manage group membership requests.",
      isactive           => 1,
      allows_unconfirmed => 0,
      version            => '1.0',
    });

    $prod->set_group_controls(
      $group,
      {
        entry         => 1,
        editbugs      => 1,
        membercontrol => CONTROLMAPMANDATORY,
        othercontrol  => CONTROLMAPMANDATORY
      }
    );
    $prod->set_always_private(1);
    $prod->update();

    Bugzilla::Component->create({
      product      => $prod,
      name         => "Self service",
      description  => "Allows a user to start a request for their own memberships.",
      initialowner => Bugzilla->params->{system_user},
      bvp_template_id => $bvp->id,
    });
    $dbh->bz_commit_transaction();
  }

  my $comp
    = Bugzilla::Component->new({product => $prod, name => "Membership Review"});

  unless ($comp) {
    Bugzilla::Component->create({
      product      => $prod,
      name         => "Membership Review",
      description  => "Allows Blessed users to review group membership.",
      initialowner => Bugzilla->params->{system_user},
    });
  }

  ## REDHAT EXTENSION START 2148313
  my $sys_user = Bugzilla::User->new({name => Bugzilla->params->{system_user}});

  my $rule_group = Bugzilla::Extension::RuleEngine::RuleGroup->new(
    {name => 'Group Membership Workflow'});

  if ($rule_group) {
    $rule_group->set_name('Workflows');
    $rule_group->set_description('Rules for Workflows');
    $rule_group->update();
  }
  else {
    $rule_group
      = Bugzilla::Extension::RuleEngine::RuleGroup->new({name => 'Workflows'});
    unless ($rule_group) {
      $rule_group = Bugzilla::Extension::RuleEngine::RuleGroup->create({
        name        => 'Workflows',
        description => 'Rules for Workflows',
        user_id     => $sys_user->id
      });
    }
  }

  $prod = Bugzilla::Product->new({name => 'Automated Maintenance'});
  unless ($prod) {

    $dbh->bz_start_transaction();

    $prod = Bugzilla::Product->create({
      classification     => $classif->name,
      name               => "Automated Maintenance",
      description        => "Automated maintenance processes.",
      isactive           => 1,
      allows_unconfirmed => 0,
      version            => '1.0',
      rule_group         => $rule_group->id,
    });

    $prod->set_group_controls(
      $group,
      {
        entry         => 1,
        editbugs      => 1,
        membercontrol => CONTROLMAPMANDATORY,
        othercontrol  => CONTROLMAPMANDATORY
      }
    );

    $prod->set_always_private(1);
    $prod->update();

    my $comp2 = Bugzilla::Component->create({
      product         => $prod,
      name            => "api_key_expiry",
      description     => "Automated expiration tasks for APIKeys",
      initialowner    => Bugzilla->params->{system_user},
      bvp_template_id => $bvp->id,
    });

    $dbh->bz_commit_transaction();
  }

  my $flag_type = Bugzilla::FlagType->new({name => '1_week_warning'});

  unless ($flag_type) {
    $dbh->bz_start_transaction();

    my %args = (
      sortkey          => 1,
      name             => '1_week_warning',
      inclusions       => [$prod->id . ':0'],
      description      => '1 Week Warning',
      is_requestable   => 0,
      exclusions       => [],
      is_multiplicable => 0,
      request_group    => 'approvals_admin',
      is_active        => 1,
      target_type      => 'bug',
      view_group       => '',
      category         => '',
    );

    $flag_type = Bugzilla::FlagType->create(\%args);

    $dbh->bz_commit_transaction();
  }

  my $rule_obj = Bugzilla::Extension::RuleEngine::Rule->new(
    {name => 'APIKey 1 Week Warning'});

  unless ($rule_obj) {
    $dbh->bz_start_transaction();

    $rule_obj = Bugzilla::Extension::RuleEngine::Rule->create({
      name          => 'APIKey 1 Week Warning',
      description   => 'A rule to send 1 week warning mail for APIKey expiry',
      rule_group_id => $rule_group->id,
      is_periodic   => 1,
    });

    my $match = {
      product    => {values => ['Automated Maintenance']},
      component  => {values => ['api_key_expiry']},
      bug_status => {values => ["All open"]},
      flag_types => [{status => ["X"], name => "1_week_warning"}],
      "custom"   => [{
        "value"  => "7d",
        "negate" => 0,
        "op"     => "lessthaneq",
        "field"  => "deadline"
      }],
    };

    my $action = {
      flag_types => {"1_week_warning" => {"status" => "+", "condition" => ""}},
      "comment"  =>
        {"text" => "The APIKey this bug is related to will expire within 7 days!"},
    };

    my $rule_detail = Bugzilla::Extension::RuleEngine::RuleDetail->create({
      match      => $match,
      action     => $action,
      creator_id => $sys_user->id,
      rule_id    => $rule_obj->id,
      isactive   => 1,
    });

    $rule_obj->set_current_detail_id($rule_detail->id);
    $rule_obj->update;

    $dbh->bz_commit_transaction();
  }

  my $rule_obj2
    = Bugzilla::Extension::RuleEngine::Rule->new({name => 'APIKey expired'});

  unless ($rule_obj2) {
    $dbh->bz_start_transaction();
    $rule_obj2 = Bugzilla::Extension::RuleEngine::Rule->create({
      name          => 'APIKey expired',
      description   => 'A rule to closed banned APIKey bugs',
      rule_group_id => $rule_group->id,
      is_periodic   => 1,
    });

    my $match2 = {
      product    => {values => ['Automated Maintenance']},
      component  => {values => ['api_key_expiry']},
      bug_status => {values => ["All open"]},
      "custom"   => [{
        "value"  => "now",
        "negate" => 0,
        "op"     => "lessthaneq",
        "field"  => "deadline"
      }],
    };

    my $action2 = {bug_status => "CLOSED", resolution => "COMPLETED"};

    my $rule_detail2 = Bugzilla::Extension::RuleEngine::RuleDetail->create({
      match      => $match2,
      action     => $action2,
      creator_id => $sys_user->id,
      rule_id    => $rule_obj2->id,
      isactive   => 1,
    });

    $rule_obj2->set_current_detail_id($rule_detail2->id);
    $rule_obj2->update;

    $dbh->bz_commit_transaction();
  }
  ## REDHAT EXTENSION END 2148313

  return;
}

sub install_update_db {

  ## REDHAT EXTENSION START 2148313
  my $field  = Bugzilla::Field->new({name => 'resolution'});
  my $choice = Bugzilla::Field::Choice->type($field)->new({name => 'COMPLETED'});
  if (!defined $choice) {
    print "Creating value COMPLETED for field resolution\n";
    Bugzilla::Field::Choice->type($field)
      ->create({value => 'COMPLETED', sortkey => 20, isactive => 1});
  }
  ## REDHAT EXTENSION END 2148313

  return;
}

sub db_schema_abstract_schema {
  my ($self, $args) = @_;

  my $schema = $args->{schema};

  ## Group Request
  $schema->{workflow_groupreq} = {
    FIELDS => [
      id          => {TYPE => 'MEDIUMSERIAL', NOTNULL => 1, PRIMARYKEY => 1},
      managed     => {TYPE => 'BOOLEAN',      NOTNULL => 1, DEFAULT    => 'FALSE'},
      invite_only => {TYPE => 'BOOLEAN',      NOTNULL => 1, DEFAULT    => 'FALSE'},
      regex       => {TYPE => 'TINYTEXT',     NOTNULL => 1, DEFAULT    => "''"},
      group_id    => {
        TYPE       => 'INT3',
        NOTNULL    => 1,
        REFERENCES =>
          {TABLE => 'groups', COLUMN => 'id', DELETE => 'CASCADE', NOTNULL => 1}
      },
      flag_id =>
        {TYPE => 'INT3', REFERENCES => {TABLE => 'flagtypes', COLUMN => 'id'}},
    ],
    INDEXES => [
      workflow_groupreq_group_unique_idx =>
        {FIELDS => [qw(group_id)], TYPE => 'UNIQUE'},
    ],
  };

  $schema->{workflow_group_membership_reviews} = {
    FIELDS => [
      id      => {TYPE => 'MEDIUMSERIAL', NOTNULL => 1, PRIMARYKEY => 1},
      user_id => {
        TYPE       => 'INT3',
        NOTNULL    => 1,
        REFERENCES =>
          {TABLE => 'profiles', COLUMN => 'userid', DELETE => 'CASCADE', NOTNULL => 1}
      },
      group_id => {
        TYPE       => 'INT3',
        NOTNULL    => 1,
        REFERENCES =>
          {TABLE => 'groups', COLUMN => 'id', DELETE => 'CASCADE', NOTNULL => 1}
      },
      bug_id => {
        TYPE       => 'INT3',
        NOTNULL    => 1,
        REFERENCES =>
          {TABLE => 'bugs', COLUMN => 'bug_id', DELETE => 'CASCADE', NOTNULL => 1}
      },
    ],
    INDEXES => [
      workflow_bug_group_reviews_unique_idx =>
        {FIELDS => [qw(user_id group_id)], TYPE => 'UNIQUE'},
    ],
  };

  return;
}

sub page_before_template {
  my ($self, $args) = @_;

  my $page = $args->{page_id};
  my $vars = $args->{vars};

  if ($page =~ m{^workflows/group_request\.}) {
    page_group_request($vars);
  }
  elsif ($page =~ m{^workflows/group_req_admin\.}) {
    page_group_admin($vars);
  }
  elsif ($page =~ m{^workflows/manage_components\.}) {
    page_manage_components($vars);
  }
  elsif ($page =~ m{^workflows/group_review\.}) {
    page_group_review($vars);
  }
  elsif ($page =~ m{^workflows/create_bot\.}) {
    page_create_bot($vars);
  }
  elsif ($page =~ m{^workflows/group_invite\.}) {
    page_group_invite($vars);
  }

  return;
}

sub template_before_process {
  my ($self, $args) = @_;
  my $file = $args->{'file'};
  my $vars = $args->{vars};

  my $user;

  if ($file eq 'global/header.html.tmpl') {
    $user = Bugzilla->user;
    if ($user->id) {
      my $dbh = Bugzilla->dbh;

      my $SQL = <<'EOS';
SELECT count(*)
FROM flags
WHERE
  requestee_id = ?
  AND status = '?'
  AND type_id IN
    ( SELECT id
      FROM flagtypes
      WHERE name LIKE 'approvals_request_%')
EOS

      my $count = $dbh->selectrow_array($SQL, undef, $user->id);
      if ($count) {
        message_2_array($vars);

        push(
          @{$vars->{message}},
          ['outstanding_personal_group_requests', {personal_requests => $count}]
        );
      }

      $SQL = <<'EOS';
SELECT count(*)
FROM flags
WHERE
bug_id IN (
  SELECT bug_id
  FROM bugs b
  WHERE product_id = (SELECT id FROM products WHERE name = 'Group Membership')
  AND component_id = (
    SELECT id
    FROM components
    WHERE name = 'Self service'
    AND product_id = b.product_id
  )
  AND bug_status <> 'CLOSED'
)
AND status = '?'
AND type_id IN
  ( SELECT id
    FROM flagtypes
    WHERE name LIKE 'approvals_request_%'
    AND grant_group_id IN
      ( SELECT group_id
        FROM user_group_map
        WHERE user_id = ?
        UNION
        SELECT grantor_id
        FROM group_group_map
        WHERE grant_type = 0
        AND member_id IN
          ( SELECT group_id
            FROM user_group_map
            WHERE user_id = ?
          )
      )
  )
EOS

      $count = $dbh->selectrow_array($SQL, undef, $user->id, $user->id);

      if ($count) {
        message_2_array($vars);

        push(
          @{$vars->{message}},
          ['outstanding_approval_group_requests', {approval_group_requests => $count}]
        );
      }

      $SQL = <<'EOS';
SELECT count(*)
FROM flags
WHERE
bug_id IN (
  SELECT bug_id
  FROM bugs b
  WHERE product_id = (SELECT id FROM products WHERE name = 'Group Membership')
  AND component_id = (
    SELECT id
    FROM components
    WHERE name = 'Membership Review'
    AND product_id = b.product_id
  )
  AND bug_status <> 'CLOSED'
  AND DATE_PART('day', NOW() - creation_ts) >= 30
)
AND status = '?'
AND type_id IN
  ( SELECT id
    FROM flagtypes
    WHERE name LIKE 'approvals_request_%'
    AND grant_group_id IN
      ( SELECT group_id
        FROM user_group_map
        WHERE user_id = ?
        UNION
        SELECT grantor_id
        FROM group_group_map
        WHERE grant_type = 0
        AND member_id IN
          ( SELECT group_id
            FROM user_group_map
            WHERE user_id = ?
          )
      )
  )
EOS
      $count = $dbh->selectrow_array($SQL, undef, $user->id, $user->id);

      if ($count) {
        message_2_array($vars);

        push(
          @{$vars->{message}},
          ['outstanding_approval_group_reviews', {approval_group_reviews => $count}]
        );
      }
    }
  }

  return;
}

sub webservice {
  my ($self, $args) = @_;

  my $dispatch = $args->{dispatch};
  $dispatch->{GroupRequest}
    = "Bugzilla::Extension::Workflows::WebService::GroupRequest";

  $dispatch->{ManageComponents}
    = "Bugzilla::Extension::Workflows::WebService::ManageComponents";

  $dispatch->{GroupReview}
    = "Bugzilla::Extension::Workflows::WebService::GroupReview";

  $dispatch->{GroupInvite}
    = "Bugzilla::Extension::Workflows::WebService::GroupInvite";

  return;
}

sub bug_end_of_update {
  my ($self, $args) = @_;
  my $dbh = Bugzilla->dbh;

  my ($bug, $old_bug, $timestamp, $changes)
    = @$args{qw(bug old_bug timestamp changes)};

  if ($bug->product eq 'Group Membership' && $bug->component eq 'Self service') {
    group_membership_bug_update($args);
  }
  elsif ($bug->product eq 'Group Membership'
    && $bug->component eq 'Membership Review')
  {
    group_membership_review_bug_update($args);
  }
  ## RED HAT EXTENSION START 2148313
  elsif ($bug->product eq 'Automated Maintenance'
    && $bug->component eq 'api_key_expiry')
  {
    api_key_expiry_bug_update($args);
  }
  ## RED HAT EXTENSION END 2148313

  return;
}

sub _flag_user_can_edit {
  my ($self, $user) = @_;

  my $can_set_flag = 1;
  my $flag_type    = $self->type;
  my $flag_name    = $flag_type->name;
  if ($flag_name =~ m/^approvals_request_(.+)$/) {
    my $group_name = $1;
    $group_name =~ s/\./ /g;
    my $group   = Bugzilla::Group->check({name => $group_name});
    my $subject = $self->bug->short_desc;
    my $uname;
    if ($self->bug->component eq 'Self service') {
      $subject =~ m/Group membership request for (.*)$/;
      $uname = $1;
    }
    elsif ($self->bug->component eq 'Membership Review') {
      $subject =~ m/Group membership review for (.*) in (.*)$/;
      $uname = $1;
    }
    if ($uname) {
      my $requestor = Bugzilla::User->new({name => $uname});
      if ($requestor) {
        $can_set_flag = $group->can_approve($user, $requestor);
      }
      else {
        $can_set_flag = $user->in_group('approvals_admin');
      }
    }
    else {
      $can_set_flag = 0;
    }
  }

  return $can_set_flag;
}

sub _flagtype_user_can_edit {
  my ($self, $user, $bug) = @_;

  my $can_edit = 0;

  my $flag_name = $self->name;
  if ($flag_name =~ m/^approvals_request_(.+)$/) {
    my $group_name = $1;
    $group_name =~ s/\./ /g;
    my $group   = Bugzilla::Group->check({name => $group_name});
    my $subject = $bug->short_desc;
    $subject =~ m/Group membership request for (.*)$/;
    my $uname     = $1;
    my $requestor = Bugzilla::User->check($uname);
    $can_edit = $group->can_approve($user, $requestor);
  }
  else {
    $can_edit = $user->can_request_flag($self);
  }

  return $can_edit;
}

sub _user_days_since_login {
  my ($self) = @_;

  unless (defined($self->{days_since_login})) {
    my $from_date = $self->last_seen_date;

    if (!$from_date || $from_date eq '') {
      $from_date = '1970-01-01';
    }

    my $dbh = Bugzilla->dbh;

    $self->{days_since_login}
      = $dbh->selectrow_array("SELECT DATE_PART('day', NOW() - ?)", undef,
      $from_date);
  }

  return $self->{days_since_login};
}

sub _user_get_review_bug {
  my ($self, $group) = @_;

  my $bug;

  my $dbh = Bugzilla->dbh;

  my $sql = <<EOS;
SELECT bug_id
FROM workflow_group_membership_reviews
WHERE group_id = ?
  AND user_id = ?
EOS

  my $bug_id = $dbh->selectrow_array($sql, undef, $group->id, $self->id);

  if ($bug_id) {
    $bug = Bugzilla::Bug->check($bug_id);
  }

  return $bug;
}

sub _user_get_all_review_bugs {
  my ($self) = @_;

  my $bugs = Bugzilla::Bug->new_from_where(
    'bug_id in (SELECT bug_id FROM workflow_group_membership_reviews WHERE user_id = ?)',
    [$self->id]
  );

  return $bugs;
}

sub _bug_age {
  my ($self) = @_;

  unless (defined($self->{age})) {
    my $dbh = Bugzilla->dbh;

    $self->{age} = $dbh->selectrow_array("SELECT DATE_PART('day', NOW() - ?)",
      undef, $self->creation_ts);
  }

  return $self->{age};
}

sub _user_create_review_bug {
  my ($self, $requestor, $group, $message) = @_;

  my $params = {
    product    => 'Group Membership',
    component  => 'Membership Review',
    short_desc => "Group membership review for "
      . $self->login . " in "
      . $group->name,
    version  => '1.0',
    reporter => $requestor,
    groups   => ['approvals_admin'],
  };

##  push(@{$params->{cc}}, $self->login);

  my $gwf = $group->group_workflow;

  ThrowUserError("approvals_not_permitted", {name => $group->name})
    unless ($gwf && $gwf->managed);

  if ($gwf->approval_group && $gwf->approval_group->name ne 'approvals_admin') {
    push(@{$params->{groups}}, $gwf->approval_group->name);
  }

  my $flag_vals = {type_id => $gwf->flag_id, status => '?'};
  push(@{$params->{flags}}, $flag_vals);

  my $descr = <<EOT;
The user has been notified they need to change their password to regain access to their account.
EOT

  $params->{comment} = $descr;

  Bugzilla->set_user(Bugzilla::User->check(Bugzilla->params->{system_user}));

  my $dbh = Bugzilla->dbh;

  $dbh->bz_start_transaction();

  my $bug = Bugzilla::Bug->create($params);

  # No bug_mail
  $dbh->do('UPDATE bugs SET lastdiffed = delta_ts WHERE bug_id = ?',
    undef, $bug->id);

  $dbh->do(
    'INSERT INTO workflow_group_membership_reviews (user_id, bug_id, group_id) VALUES (?, ?, ?)',
    undef, $self->id, $bug->id, $group->id
  );

  my %args;
  $args{names}   = [$self->login];
  $args{action}  = 'reconfirm';
  $args{message} = $message;

  Bugzilla->user->manage_user_access(\%args);

  $dbh->bz_commit_transaction();

  Bugzilla->set_user($requestor);

  return ($bug);
}

sub object_end_of_update {
  my ($self, $args) = @_;

  my ($object, $old_object, $changes) = @$args{qw(object old_object changes)};

  # Only trigger this is the forgot password process triggers it
  my $action = Bugzilla->cgi->param('a');
  my $script = Bugzilla->cgi->script_name;

  if ( $object->isa('Bugzilla::User')
    && defined $changes->{'cryptpassword'}
    && $action
    && $action eq 'chgpw'
    && $script
    && $script =~ m{/token.cgi$})
  {
    my $cur_user = Bugzilla->user;
    Bugzilla->set_user(Bugzilla::User->check(Bugzilla->params->{system_user}));

    my $bugs = $object->get_all_review_bugs;
    my $params;
    my $dbh = Bugzilla->dbh;

    my $msg = 'The user has updated their password.';
    $params->{bug_status} = 'CLOSED';
    $params->{resolution} = 'CURRENTRELEASE';
    $params->{comment}    = {body => $msg};

    foreach my $bug (@$bugs) {
      my @flags;

      # Remove all flags
      foreach my $flag (@{$bug->flags}) {
        push(@flags, {id => $flag->id, status => 'X'});
      }

      $bug->set_all($params);
      $bug->set_flags(\@flags, []);
      $bug->update();
      $dbh->do("DELETE FROM workflow_group_membership_reviews WHERE bug_id = ?",
        undef, $bug->id);
    }

    Bugzilla->set_user($cur_user);
  }

  ## REDHAT EXTENSION START 2148313
  ## close bugs if APIKey is revoked
  if ( $object->isa('Bugzilla::User::APIKey')
    && defined $changes->{'revoked'}
    && $changes->{'revoked'}->[1] == 1)
  {
    api_key_expiry_key_update($args);
  }
  ## REDHAT EXTENSION END 2148313

  return;
}

sub _product_editors {
  my ($self) = @_;

  my $group_controls  = $self->group_controls;
  my @edit_components = map { $group_controls->{$_}->{group} }
    grep { $group_controls->{$_}{editcomponents} == 1 } keys %$group_controls;

  my %results;
  foreach my $group (@edit_components) {
    my @members;
    foreach my $member (@{$group->members_non_inherited}) {
      push(@members, $member->login)
        if ($member->is_enabled
        && !$member->in_group('admin')
        && !$member->in_group('automaton'));
    }
    $results{$group->name} = \@members;
  }

  return \%results;
}

sub config_modify_panels {
  my ($self, $args) = @_;

  my $panels = $args->{panels};

  push(
    @{$panels->{core}{params}},
    ## REDHAT EXTENSION START 2039558
    {name => 'general_groups', type => 't', default => '',},
    ## REDHAT EXTENSION END 2039558
  );

  return;
}

## REDHAT EXTENSION START 2148313
sub _user_contacts {
  my $self = shift();
  my $to   = $self->email;

  if ($self->prefs_admin_group) {
    my $group = $self->prefs_edit_group;

    $to = join(", ",
      map { (!$_->prefs_admin_group && !$_->is_bot) ? ($_->email) : () }
        @{$group->members_direct()});
  }

  return $to;
}

sub _user_admin_logins {
  my $self = shift();
  my @logins;

  if ($self->prefs_admin_group) {
    my $group = $self->prefs_edit_group;

    @logins = map { (!$_->prefs_admin_group && !$_->is_bot) ? ($_->login) : () }
      @{$group->members_direct()};
  }
  else {
    push(@logins, $self->login);
  }

  return \@logins;
}

sub _user_admins {
  my $self = shift();
  my @admins;

  if ($self->user->prefs_admin_group) {
    my $group = $self->user->prefs_edit_group;

    @admins = grep { (!$_->prefs_admin_group && !$_->is_bot) }
      @{$group->members_direct()};
  }
  else {
    push(@admins, $self->user);
  }

  return \@admins;
}
## REDHAT EXTENSION END 2148313

__PACKAGE__->NAME;

__END__

=head1 NAME

Bugzilla::Extension::Workflows

=head1 DESCRIPTION

This extensions allows creation of custom workflows within Bugzilla.

=head1 Workflows

L<Group Approval|Bugzilla::Extension::Workflows::GroupApproval>

L<Manage Components|Bugzilla::Extension::Workflows::ManageComponents>

L<Group Membership Review|Bugzilla::Extension::Workflows::GroupReview>

L<Create Bot Account|Bugzilla::Extension::Workflows::CreateBotAccount>
