#!/usr/bin/perl -w
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

use strict;
use warnings;
use 5.10.1;

use lib qw(. /var/www/html/bugzilla/);

use Bugzilla;
use Bugzilla::User::APIKey;
use Bugzilla::Mailer;
use Bugzilla::Util;
use Bugzilla::Error;

use Getopt::Long;
use Pod::Usage;
use Data::Dumper;

=head1 NAME

check_apikeys.pl

=head1 SYNOPSIS

check_apikeys.pl <options>

  --help|h         Display help text
  --man            Display man page


=head1 DESCRIPTION

This command wil:

 * Automatically ban revoked expiring APIKeys
 * Start the ban workflow for unrevoked expiring APIKeys
 * Automatically revoke APIKeys that have not been used in the last 30 days

=head1 EXAMPLES

 $ check_apikeys.pl

=cut

local $| = 1;

my $help  = 0;
my $man   = 0;
my $count = 0;

my $get_option = GetOptions("help|h" => \$help, "man" => \$man,)
  or pod2usage(2);

pod2usage(1)             if $help;
pod2usage(-verbose => 2) if $man;

my $dbh      = Bugzilla->dbh;
my $template = Bugzilla->template;

Bugzilla->set_user(Bugzilla::User->check(Bugzilla->params->{system_user}));

say "Starting APIKEY maintenance";
ban();

## Bug 2174291 - Revoke revoking
## revoke();

say "Finished APIKEY maintenance, processed $count keys";

exit;

### subs

sub ban {

  my $params = {
    product   => 'Automated Maintenance',
    component => 'api_key_expiry',
    version   => '1.0',
    groups    => ['approvals_admin'],
  };

  my $sql = "
    SELECT id
    FROM user_api_keys
    WHERE
      banned = 0
    AND 
    "
    . $dbh->sql_to_days('NOW()') . " - "
    . $dbh->sql_to_days('creation_ts')
    . " > 335";

  ##say $sql;

  my $api_key_ids = $dbh->selectcol_arrayref($sql);

  my $apikeys = Bugzilla::User::APIKey->new_from_list($api_key_ids);

  foreach my $apikey (@$apikeys) {
    my ($message, $vars,);

    next if ($apikey->expiry_bug);

    my $id           = $apikey->id;
    my $dscr         = $apikey->description;
    my $apikey_login = $apikey->user->login;

    if ($apikey->revoked) {
      $apikey->set_banned(1);
      $apikey->update();
      say "Banned revoked key  $id '$dscr' $apikey_login";
    }
    else {
      my $deadline = $dbh->selectrow_array(
        'SELECT '
          . $dbh->sql_date_format(
          $dbh->sql_date_math('NOW()', '+', 60, 'DAY'), '%Y-%m-%d'
          )
      );

      $params->{deadline}   = $deadline;
      $params->{cc}         = $apikey->user->admin_logins;
      $params->{short_desc} = "APIKey expiry for key $id '$dscr'";
      my $url = correct_urlbase() . 'userprefs.cgi?tab=apikey';

      if ($apikey->user->is_bot) {
        $url .= "&other_user=$apikey_login";
      }

      $message = <<"SQL";
This is an automated process for expiring APIKeys.

APIKeys have a lifetime of 12 months and this APIKey is approaching EOL.

This APIKey will be automatically banned on $deadline. You need to take action
to replace use of this APIKey with a new one before this date.

You will be notified again when 7 days remain before this date.

You will be notified again when the key is banned and this bug is closed.

If the owner[s] of this APIKey revoke the key before the expiry date then this
APIKey will be banned immediately and this bug will be closed.

APIKey details:

ID: $id
Description: $dscr
Owner: $apikey_login

You can manage this APIKey at $url

SQL

      $params->{comment} = $message;

      $dbh->bz_start_transaction();
      my $bug = Bugzilla::Bug->create($params);
      $dbh->bz_commit_transaction();
      my $bug_id = $bug->id;
      say
        "Created expiry workflow bug $bug_id for APIKey $id '$dscr' owned by $apikey_login";
    }

    $count++;
  }

  return;
}

sub revoke {
  my $sql = "
    SELECT id
    FROM user_api_keys
    WHERE
        revoked = 0
    AND banned = 0
    AND
    (
        (
            last_used IS NOT NULL
            AND "
    . $dbh->sql_to_days('NOW()') . " - " . $dbh->sql_to_days('last_used') . " > 30
        )
        OR
        (
            last_used IS NULL
            AND "
    . $dbh->sql_to_days('NOW()') . " - "
    . $dbh->sql_to_days('creation_ts') . " > 30
      )
    )";

  ##say $sql;

  my $api_key_ids = $dbh->selectcol_arrayref($sql);

  my $apikeys = Bugzilla::User::APIKey->new_from_list($api_key_ids);

  foreach my $apikey (@$apikeys) {
    my ($message, $vars);

    next if ($apikey->expiry_bug);

    my $id           = $apikey->id;
    my $dscr         = $apikey->description;
    my $apikey_login = $apikey->user->login;

    $apikey->set_revoked(1);
    $apikey->update();

    $vars->{apikey} = $apikey;
    $template->process('email/apikey_revoked.txt.tmpl', $vars, \$message)
      || ThrowTemplateError($template->error());

    MessageToMTA($message);

    $count++;
    say "Revoked unused APIKey $id '$dscr' $apikey_login";
  }

  return;
}
