# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::Workflows::GroupApproval;

use strict;
use warnings;
use 5.10.1;

use parent qw(Exporter);

our @EXPORT = qw(
  page_group_admin
  page_group_request
  group_membership_bug_update
  can_admin_group_workflow
  group_workflow
  can_apply_for
  can_approve
  can_invite
  apply_for_groups
  admin_can_manage
);

use Bugzilla::Constants;
use Bugzilla::Error;
use Bugzilla::Util;
use Bugzilla::Extension::Workflows::Util;
use Bugzilla::Token;

sub page_group_admin {
  my ($vars) = @_;
  my $dbh    = Bugzilla->dbh;
  my $input  = Bugzilla->input_params;

  Bugzilla->login(LOGIN_REQUIRED);
  my $user = Bugzilla->user;
  $user->can_admin_group_workflow();

  $vars->{'groups'} = [Bugzilla::Group->get_all];

  return;
}

sub page_group_request {
  my ($vars) = @_;
  my $dbh    = Bugzilla->dbh;
  my $input  = Bugzilla->input_params;
  my $cgi    = Bugzilla->cgi;
  my $action = $cgi->param('action');

  Bugzilla->login(LOGIN_REQUIRED);
  my $user = Bugzilla->user;

  if ($action) {
    if ($action eq 'create') {
      my $token = $input->{'token'};
      check_token_data($token, 'create_group_request');
      my @apply_groups;

      foreach my $cat (split(',', Bugzilla->params->{'group_categories'})) {
        $cat = lc($cat);
        $cat =~ s/\s//g;

        my @grps = $cgi->param("groups_$cat");
        foreach my $grp (@grps) {
          my $group = Bugzilla::Group->check({name => $grp});
          $group->can_apply_for($user)
            || ThrowUserError("approvals_not_permitted", {name => $grp});

          push(@apply_groups, $group);
        }
      }

      if (@apply_groups) {
        my $comment = $cgi->param('rationale');
        my $bug     = apply_for_groups($user, $user, $comment, \@apply_groups);
        $vars->{'message'} = 'group_request_created';
        $vars->{'bug'}     = $bug;
      }

      delete_token($token);
    }
  }
  else {
    $vars->{'token'} = issue_session_token('create_group_request');
  }

  my %groups;
  foreach my $group (Bugzilla::Group->get_all) {
    if ($group->can_apply_for($user)) {
      push(@{$groups{$group->category}}, $group);
    }
  }

  $vars->{'groups'} = \%groups;

  return;
}

sub apply_for_groups {
  my ($requestee, $requestor, $rationale, $groups) = @_;
  my $dbh = Bugzilla->dbh;

  ThrowUserError("approvals_rationale_required")
    if (!$rationale || $rationale eq '');

  my $params = {
    product    => 'Group Membership',
    component  => 'Self service',
    short_desc => "Group membership request for " . $requestee->login,
    version    => '1.0',
    reporter   => $requestee,
    groups     => ['approvals_admin'],
  };

  push(@{$params->{cc}},     $requestee->login);
  push(@{$params->{groups}}, 'approvals_admin');

  my $manager;

  foreach my $group (@$groups) {
    my $gwf = $group->group_workflow;

    ThrowUserError("approvals_not_permitted", {name => $group->name})
      unless ($gwf && $gwf->managed);

    if ($gwf->invite_only) {
      my @reasons;
      unless ($group->can_invite($requestor, $requestee, \@reasons)) {
        ThrowUserError("approvals_cannot_invite",
          {group => $group, requestee => $requestee, reasons => \@reasons});
      }
    }

    my $flag_vals = {type_id => $gwf->flag_id, status => '?'};

    if ($gwf->cc_list) {
      push(@{$params->{cc}}, split(/[,\s]+/, $gwf->cc_list));
    }

    my $regex;
    $regex = $gwf->regex if ($gwf && $gwf->regex);
    my $bu_regex = Bugzilla->params->{bot_or_user_regex};

    if ($gwf->approval_group) {
      push(@{$params->{groups}}, $gwf->approval_group->name);
      my $users = $gwf->approval_group->members_non_inherited();
      foreach my $member (@$users) {
        if ($member->is_enabled && $member->email_enabled) {
          push(@{$params->{cc}}, $member->email);
        }
      }
    } else {
      # LDAP is slow, only do it if necessary
      if ($requestee->login =~ /\@redhat.com$/) {
        $manager = get_manager($requestee);
        if ($manager) {
          push(@{$params->{cc}}, $manager->login);
        }
      }

      if ($manager) {
        $flag_vals->{requestee} = $manager->login;
      }
      elsif (!$gwf->cc_list
        && (!$regex || ($regex !~ m{\@redhat\\\.com\$} && $regex ne $bu_regex)))
      {
        ThrowUserError(
          "approvals_missing_approval_group",
          {group => $group, requestee => $requestee}
        );
      }
    }

    push(@{$params->{flags}}, $flag_vals);
  }

  my $descr = '';

  if ($requestor->id == $requestee->id) {
    $descr = $requestee->identity
      . " has requested membership of the following groups:\n\n";
  }
  else {
    $descr
      = $requestor->identity
      . " has requested membership of the following groups for "
      . $requestee->identity . ":\n\n";
  }

  $descr .= join("\n", map { $_->name } @$groups);

  $descr .= "\n\nThe rationale provided was:\n\n$rationale";

  $params->{comment} = $descr;

  Bugzilla->set_user(Bugzilla::User->check(Bugzilla->params->{system_user}));

  $dbh->bz_start_transaction();

  my $bug = Bugzilla::Bug->create($params);

  $dbh->bz_commit_transaction();

  Bugzilla->set_user($requestor);

  Bugzilla::Hook::process('post_bug_end_of_create', {bug => $bug});

  # Avoid bug update header
  my $mode = Bugzilla->usage_mode;
  Bugzilla->usage_mode(USAGE_MODE_CMDLINE);
  $bug->send_changes();
  Bugzilla->usage_mode($mode);

  Bugzilla::Hook::process('bug_create_after_mail', {bug => $bug});

  return ($bug);
}

sub group_workflow {
  my $group = shift;

  if (!defined($group->{group_workflow})) {

    my $gwf = Bugzilla::Extension::Workflows::GroupRequest->match(
      {group_id => $group->id})->[0];

    $group->{group_workflow} = $gwf if ($gwf);
  }

  return ($group->{group_workflow});
}


sub can_admin_group_workflow {
  my $user = shift;

  ($user->in_group('admin') || $user->in_group('approvals_admin'))
    || ThrowUserError(
    "auth_failure",
    {
      group  => 'approvals_admin',
      action => "edit",
      object => "workflow_groupreq_admin"
    }
    );

  return;
}

sub can_apply_for {
  my $group = shift;
  my $user  = shift;

  # if this is being called later in the approvals process we skip this check
  my $skip_invite_check = shift // 0;

  my $can = 1;

  my $gwf = $group->group_workflow;

  my $regex;
  $regex = $gwf->regex if ($gwf && $gwf->regex);

  if (!$gwf || !$gwf->managed || ($regex && $user->login !~ /$regex/)) {
    $can = 0;
  }
  elsif ($user->in_group_id($group->id)) {
    $can = 0;
  }
  elsif (!$skip_invite_check && $gwf->managed && $gwf->invite_only) {
    $can = 0;
  }

  return ($can);
}

sub can_approve {
  my $group    = shift;
  my $approver = shift;
  my $user     = shift;

  my $can = 0;

  my $gwf = $group->group_workflow;

  # if we aren't managed, use bless
  if (!$gwf || !$gwf->managed) {
    if ($group->category =~ m/^(Public)$/) {
      $can = $approver->can_bless($group->id);
    }
  }

  # allow approval group members to approve
  elsif ($gwf->approval_group && $approver->in_group_id($gwf->approval_group->id))
  {
    $can = 1;
  }

  # allow managers to approve other requests
  elsif (!$gwf->approval_group) {
    if ($user->is_bot) {
      ## BUGBUG this leaves the requestee blank and therefore can only be
      ## done by an admin
      $can = 1;
    }
    else {
      my $manager = get_manager($user);
      if ($manager && ($approver->id == $manager->id)) {
        $can = 1;
      }
      elsif ($approver->in_group('trusted_approvers')) {
        $can = 1;
      }
      elsif ($manager) {
        my $man_man;
        eval { $man_man = get_manager($manager); };

        # my manager's manager
        if ($man_man && ($approver->id == $man_man->id)) {
          $can = 1;
        }
      }
    }
  }

  return ($can);
}

sub can_invite {
  my $group   = shift;
  my $inviter = shift;
  my $user    = shift;
  my $reasons = shift;

  my $can = 1;

  my $gwf = $group->group_workflow;

  if ($gwf && $gwf->managed) {
    unless ($gwf->invite_only) {
      $can = 0;
      push(@$reasons, 'This is not an invite only group');
    }

    if ($gwf->approval_group && !$inviter->in_group_id($gwf->approval_group->id)) {
      $can = 0;
      push(@$reasons, 'You are not in the approval group');
    }

    my $regex = $gwf->regex;
    if ($regex && $regex ne '' && $user->login !~ m/$regex/) {
      $can = 0;
      push(@$reasons, "The user does not match the regular expression '$regex'");
    }
  }

  return $can;
}

## REDHAT EXTENSION BEGIN 1882298
sub admin_can_manage {
  my $group   = shift;
  my $user    = shift;
  my $reasons = shift;

  my $in_group = $user->in_group_id_direct($group->id);

  my $can   = 1;
  my $gwf   = $group->group_workflow;
  my $regex = '';
  $regex = $gwf->regex if ($gwf && $gwf->regex);

  if ($group->category eq 'Agile') {
    $can = 0;
    push(@$reasons, "Use the Agile Team interface to manage team membership");
  }
  elsif (!$in_group && $group->category eq 'Deprecated') {
    $can = 0;
    push(@$reasons, "Do not add users to Deprecated groups.");
  }
  elsif (!$in_group && $regex && $regex ne '' && $user->login !~ /$regex/) {
    $can = 0;
    push(@$reasons, "The user does not match the regular expression '$regex'");
  }

  return $can;
}
## REDHAT EXTENSION END 1882298

sub group_membership_bug_update {
  my ($args) = @_;
  my $dbh = Bugzilla->dbh;

  my ($bug, $old_bug, $timestamp, $changes)
    = @$args{qw(bug old_bug timestamp changes)};

  return if ($bug->bug_status eq 'CLOSED');

  my $approver = Bugzilla->user;
  my $subject  = $bug->short_desc;
  $subject =~ m/Group membership request for (.*)$/;
  my $uname     = $1;
  my $requestor = Bugzilla::User->new({name => $uname});

  my (@added, @removed);
  my $msg     = '';
  my $skipped = 0;

  ThrowUserError("approvals_subject_change_blocked")
    if (exists $changes->{'short_desc'});

  if ($requestor && exists $changes->{'flagtypes.name'}) {

    my @flags_added = split(/,\s*/, $changes->{'flagtypes.name'}->[1]);

    foreach my $change (@flags_added) {
      my ($group_name, $value);

      if ($change =~ m/^approvals_request_(.+)([\?\+\-])$/) {
        $group_name = $1;
        $value      = $2;
      }

      next unless ($group_name);

      my $flag_name = "approvals_request_$group_name";
      $group_name =~ s/\./ /g;

      my $flagtype = Bugzilla::FlagType->check({name => $flag_name});
      my $group    = Bugzilla::Group->check({name => $group_name});

      ThrowUserError("approvals_flag_mismatch",
        {group => $group, flag_name => $flag_name})
        unless ($flagtype->id == $group->group_workflow->flag_id);

      if ($value eq '+') {
        if ($requestor->in_group($group_name)) {
          $skipped++;
          $msg
            .= $requestor->identity . "is already a member of $group_name, skipping.\n";
        }
        else {
          $group->can_apply_for($requestor, 1)
            || ThrowUserError("approvals_not_permitted", {name => $group_name});

          ThrowUserError("approvals_cannot_approve",
            {group => $group, requestee => $requestor})
            unless ($group->can_approve($approver, $requestor));

          push(@added, $group_name);
          Bugzilla->logger->info(
                "Approvals system granted membership to $group_name group to "
              . $requestor->identity
              . " authorized by "
              . $approver->identity);
        }
      }
      elsif ($value eq '?') {
        $group->can_apply_for($requestor, 1)
          || ThrowUserError("approvals_not_permitted", {name => $group_name});

        my @reasons;
        my $gwf = $group->group_workflow;

        if ( $gwf
          && $gwf->invite_only
          && !$group->can_invite($approver, $requestor, \@reasons))
        {
          ThrowUserError("approvals_cannot_invite",
            {group => $group, requestee => $requestor, reasons => \@reasons});
        }
      }
      elsif ($value eq '-') {
        push(@removed, $group_name) if ($requestor->in_group($group_name));
        Bugzilla->logger->info(
              "Approvals system removed membership of $group_name group for "
            . $requestor->identity
            . " authorized by "
            . $approver->identity);
      }
    }


    Bugzilla->set_user(Bugzilla::User->check(Bugzilla->params->{system_user}));
    $requestor->set_groups({add    => \@added})   if (@added);
    $requestor->set_groups({remove => \@removed}) if (@removed);
    $requestor->update() if (@added || @removed);
  }

  my $all_flags_done = 1;
  foreach my $flag (@{$bug->flags}) {
    my $flag_name = $flag->type->name;
    next unless ($flag_name =~ /^approvals_request_(.+)$/);

    if ($flag->status ne '-' && $flag->status ne '+') {
      $all_flags_done = 0;
      last;
    }
  }

  if (@added || @removed || $all_flags_done || $skipped) {
    my $fresh_bug = Bugzilla::Bug->check($bug->id);
    my $params;

    foreach my $grp (@added) {
      $msg .= $requestor->login . " added to $grp by " . $approver->login . "\n";
    }

    foreach my $grp (@removed) {
      $msg .= $requestor->login . " removed from $grp by " . $approver->login . "\n";
    }

    if ($all_flags_done) {
      $msg .= 'All requests have been completed';
      $params->{bug_status} = 'CLOSED';
      $params->{resolution} = 'CURRENTRELEASE';
    }

    $params->{comment} = {body => $msg};
    $fresh_bug->set_all($params);
    $fresh_bug->update();
  }

  Bugzilla->set_user($approver);

  return;
}

1;


__END__

=head1 NAME

Bugzilla::Extension::Workflows::GroupApproval

=head1 DESCRIPTION

This workflow uses bugs and flags to control group membership requests. Two forms
are provided, one for admins to configure the groups being managed, and one for
users to initiate the workflow. The process revolves around standard bug
process, with the requestees either acking or nacking flags, and the extension
processing updates and closing the bugs when all flags are acked or nacked.

Security is paramount for this process. A new product is created specifically for
this workflow. Only the system account is able to create bugs in this product;
users use the form to trigger bug creation. Users only have access to bugs that
they are directly involved in. There are minimal changes required for this, it is
mostly configuring the product and flags properly, then manipulating the CC &
groups when creating bugs.

This Workflow intercepts Group changes made in the Admin UI or RPC interfaces,
and redirects them through this workflow if the groups requested are being
managed.

B<All> groups in the B<Red Hat>, B<Partner>, and B<Admin> categories B<must> be managed and
B<no one> will be able to grant membership to such groups until they are configured
properly.

=head3 Request Creation Page

This page, accessible via the I<< "My Links"->"Request Group Membership" >> menu,
consists of a form the user fills out to start the workflow.

The form has these fields:

Groups:
  For each Group classification with managed groups there is a multi-select field for
  the user to select groups they wish to become a member of.

These fields show groups the user qualifies for based on the rules specified on
the admin page.

The user will never see groups they cannot apply for, which includes groups they
are already members of, including inherited groups.

If a user cannot see a group they wish to become a member of then they will need
to contact the site administrators to see why they do not qualify. See the
I<< "Help"->"Contacts" >> menu for contact details.

Rationale:
  A text field for the user to enter why they wish membership of these groups.

Submit:
  A button to submit the form.

This form can be pre-populated by adding I<&groups=foo> to the URL, this can be
set multiple times if multiple groups are required.

=head3 Request creation workflow

Once the user submits the form the request workflow it triggered.

=over 4

=item Request is validated

=item A bug is created

=item The user is added as a CC to the bug

=item For each group requested

=over 4

=item If the group has an approval group

=over 4

=item the approval group is added to the bug

=item the approvals_request flag for the group is set to "?"

=back

=item else

=over 4

=item the user's manager is looked up in LDAP

=item the user's manager is added as a CC on the bug

=item the approvals_request flag for the group is set to "?" targeted at the user's manager

=back

=back

=item The bug is submitted

=item Email is sent to all related parties

=back

=head3 Approver workflow

Once the approver submits a bug update the extension processes the bug changes.
This happens after the normal bug update process, so most invalid flag changes
will have prevented the update from getting this far.

The extension performs a secondary check to ensure all rules are met my both
requester and approver.

Approvers can track their outstanding requests on the I<< "My Links"->"My Requests" >> menu.

=over 4

=item For each flag modified

=over 4

=item If the flag was ACK'd

The Requester is added to the requested group

=item if the flag was NAK'd

The Requester is removed from the group

=item else

No changes are made.

=back

=item Adds a comment to the bug to record the changes it made.

=item If all flags have been completed, closes the bug

=back

=head3 Admin Configuration Page

This page consists of several tabs, one for each group category, group settings
are edited inline. There are several tabs to simplify the user experience.

Documentation for the API can be found at
L<Bugzilla::Extension::Workflows::WebService::GroupRequest>

The tables have the following columns.

name:
  The name of the group.

managed:
  Are we exposing this group in the approvals system? Users will not see this group in the self service form if this is not enabled.

invite only:
  Is membership in this group invite only? If so only users in the approval group
  can nominate users for membership. Users will not see this group in the self
  service form.

login regex:
  A regex to limit who can join this group. This can be a broader regex than the
  auto approve set on the group. If a group is not invite only then this field is
  also used to restrict who can see this group on the self service form.

Approval Group:
  The group that can ack requests for this flag. If not set then the user's
  manager will be looked up in LDAP and asked to approve. The manager must have
  a Bugzilla account for this to be possible. Groups for non Red Hat accounts
  must always have an approval group set to be actionable. This is a drop down and is synced to the grant_group on the flag.

CCs:
  A comma-separated list of email addresses to CC when this flag is requested.
  This sets the "CC List" on the flag.

Flag:
  The name of the flag tracking requests for this group.

Actions:
  Icons for:

=over 4

=item * Saving changes

=item * a link to the group admin page

=item * a link to the flag admin page

=item * a link to a search for pending requests for this group

=back

Each group that is managed will has it's own flag "approvals_request_<group_name>"
this flag will have it's grant_group set to the approval_group if it is set

For manager approved groups the flags is "specifically requestable". For group
approved groups the flags is not "specifically requestable" and is locked to
the grant_group.

If the flag for the group does not exist it is created when the changes are
saved.

=head3 Product Setup

To properly secure the process is set-up in the following way:

=over 4

=item A group "approval_admin" is created, used for bugs ON, secure_mail ON

=item A classification "Workflows" is created, sortkey 300

=item A new product "Group Membership" is created

=item A new component "Self service" is created

=item The product has its access control set to

approval_admin "Mandatory/Mandatory, ENTRY, editcomponents, editbugs"

=item The product has "Bugs must always be private" ON

=back

=head3 Admin Workflow

When an admin modifies a group on the admin page the request is handled as follows.

=over 4

=item If the manage check box is enabled

=over 4

=item If the group flag does not exist, create it with the following values

  Name: "approvals_request_<group_name>"
  Description: "Approval flag for the <group_name> group."
  Inclusion: "Group Membership":__Any__
  Sort Key: 9001
  Active: True
  Requestable: True
  CC List: As supplied
  specifically requestable: False if no regex, else True
  multiplicable: False
  View Group: None
  Request Group: approval_admin
  Grant Group: "Approval Group" if set
  Flag Type Category: unset, possibly a new type if it helps the UI experience.

=item If the group flag is not requestable, make it requestable

=back

=item If the manage check box is disabled

=over 4

=item  If the group flag is requestable, set requestable to False

=back

=item If the approval group is set, ensure the group flag is not "specifically requestable"

=item If the approval group  is not set, ensure the group flag is "specifically requestable"

=item Set the CC List field on the group flag

=item Ensure the group flag grant group is "Approval Group" if set

=item If the "Approval Group" is set, ensure the group is "Shown/Shown" on the product access controls

=back

=head3 Partner group restriction change

Previously if a Partner group had a regex set for auto group membership then
Red Hat Bugzilla would limit group membership to users matching that regex.

Now, if a group workflow has a regex set then this new regex will override the
auto membership regex when validating a user's membership in that Partner group.

=head3 TODO

=over 4

=item Make sure auto membership still works

=back

=cut

