# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::Workflows::ManageComponents;

use strict;
use warnings;
use 5.10.1;

use parent qw(Exporter);

our @EXPORT = qw(
  page_manage_components
  component_search
);

use Bugzilla;
use Bugzilla::Constants;
use Bugzilla::Util;
use Bugzilla::Error;
use Bugzilla::Extension::Workflows::Util;
use List::MoreUtils qw(any);

BEGIN {
  *Bugzilla::User::check_can_manage_comp = \&_check_can_manage_comp;
  *Bugzilla::User::can_manage_comp       = \&_can_manage_comp;
  *Bugzilla::User::can_manage_comp_field = \&_can_manage_comp_field;
}

sub page_manage_components {
  my ($vars) = @_;
  my $dbh    = Bugzilla->dbh;
  my $input  = Bugzilla->input_params;

  Bugzilla->login(LOGIN_REQUIRED);
  my $user = Bugzilla->user;

  $vars->{'pool_names'}    = $input->{'pools'};
  $vars->{'team_names'}    = $input->{'teams'};
  $vars->{'show_nested'}   = $input->{'show_nested'};
  $vars->{'show_inactive'} = $input->{'show_inactive'};

  if ($input->{'product'}) {
    my $prod = Bugzilla::Product->check($input->{'product'});

    $vars->{'product'} = $prod->name;
  }

  Bugzilla::active_pools_to_vars($vars);

  $vars->{'templates'} = [Bugzilla::Extension::BugViewPlus::Template->get_all()];

  return;
}

sub component_search {
  my ($vars, $components) = @_;

  my %seen = ();

  if ($vars->{'product'}) {
    my $prod = Bugzilla::Product->check($vars->{'product'});

    foreach my $comp (@{$prod->components}) {
      _proc_component($comp, $components, \%seen, $vars);
    }

    $vars->{'product'} = $prod->name;
  }
  elsif ($vars->{'teams'}) {
    foreach my $team_name (@{$vars->{'teams'}}) {
      my $team = Bugzilla::Extension::AgileTools::Team->check($team_name);
      _process_team($team, $components, \%seen, $vars);
    }
  }
  elsif ($vars->{'pools'}) {
    foreach my $pool_name (@{$vars->{'pools'}}) {
      my $pool = Bugzilla::Extension::AgileTools::Pool->check($pool_name);
      _process_pool($pool, $components, \%seen, $vars);
    }
  }

  return;
}

sub _process_team {
  my ($team, $components, $seen, $vars) = @_;


  foreach my $pool (@{$team->pools(1)}) {
    _process_pool($pool, $components, $seen, $vars);
  }

  foreach my $comp (@{Bugzilla::Component->match({agile_team => $team->id})}) {
    _proc_component($comp, $components, $seen, $vars);
  }

  # also search at the sub_component level
  foreach my $sc (@{Bugzilla::Extension::SubComponents::RHSubComponent->match(
      {agile_team => $team->id}
    )
  })
  {
    _proc_sub_component($sc, $components, $seen, 0, $vars);
  }

  foreach my $prod (@{Bugzilla::Product->match({agile_team => $team->id})}) {
    foreach my $comp (@{$prod->components}) {
      _proc_component($comp, $components, $seen, $vars);
    }
  }

  return;
}

sub _process_pool {
  my ($pool, $components, $seen, $vars) = @_;

  foreach my $comp (@{Bugzilla::Component->match({default_pool => $pool->id})}) {
    _proc_component($comp, $components, $seen, $vars);
  }

  # also search at the sub_component level
  foreach my $sc (@{Bugzilla::Extension::SubComponents::RHSubComponent->match(
      {default_pool => $pool->id}
    )
  })
  {
    _proc_sub_component($sc, $components, $seen, 0, $vars);
  }

  return;
}

sub _proc_component {
  my ($comp, $components, $seen, $vars) = @_;

  my $c_id = "component_" . $comp->id;

  if (!defined($seen->{$c_id})) {
    my $show = _should_show($comp, 0, $vars);
    $seen->{$c_id} = $show;
    if ($show) {
      push(@$components, $comp);

      foreach my $sc (@{$comp->get_sub_components}) {
        _proc_sub_component($sc, $components, $seen, $show, $vars);
      }
    }
  }
  return;
}

sub _proc_sub_component {
  my ($sub_comp, $components, $seen, $show_par, $vars) = @_;

  my $c_id = "sub_component_" . $sub_comp->id;
  if (!defined($seen->{$c_id})) {
    my $show = _should_show($sub_comp, $show_par, $vars);
    $seen->{$c_id} = $show;
    if ($show) {
      push(@$components, $sub_comp);

      foreach my $sc (@{$sub_comp->children}) {
        _proc_sub_component($sc, $components, $seen, ($show_par || $show), $vars);
      }
    }
  }

  return;
}

sub _should_show {
  my ($comp, $show_par, $vars) = @_;

  my $should = 0;

  if (defined $vars->{show_nested} && $vars->{show_nested} && $show_par) {
    $should = 1;
  }
  elsif ((!defined $vars->{show_inactive} || !$vars->{show_inactive})
    && !$comp->is_active)
  {
    $should = 0;
  }
  elsif ($vars->{teams}) {
    if (
      $comp->default_pool_obj && (
        any { $_ eq $comp->default_pool_obj->team->name }
        @{$vars->{teams}}
      )
      )
    {
      $should = 1;
    }
    elsif ($comp->my_team && (any { $_ eq $comp->my_team->name } @{$vars->{teams}}))
    {
      $should = 1;
    }
  }
  elsif ($vars->{pools}) {
    if (
      $comp->default_pool && (
        any { $_ eq $comp->default_pool_obj->name }
        @{$vars->{pools}}
      )
      )
    {
      $should = 1;
    }
  }
  else {
    $should = 1;
  }

  return $should;
}

# $user->check_can_manage_comp
sub _check_can_manage_comp {
  my $self      = shift;
  my $component = shift;

  $self->can_manage_comp($component)
    || ThrowUserError("manage_components_cant_manage", {component => $component});

  return;
}

# $user->can_manage_comp($component)
sub _can_manage_comp {
  my $self      = shift;
  my $component = shift;

  $self->{can_manage_comp} = {} unless (defined $self->{can_manage_comp});

  my $c_id = $component->type . "_" . $component->id;

  if (not defined($self->{can_manage_comp}->{$c_id})) {
    $self->{can_manage_comp}->{$c_id} = 0;

    if ($self->in_group('editcomponents')) {
      $self->{can_manage_comp}->{$c_id} = 1;
    }
    elsif ($self->in_group("editcomponents", $component->product_id)) {
      $self->{can_manage_comp}->{$c_id} = 1;
    }

    if ($component->default_pool) {
##      if ($component->default_pool_obj->team->user_can_edit($self)) {
      if ($self->in_group_id($component->default_pool_obj->team->group_id)) {
        $self->{can_manage_comp}->{$c_id} = 1;
      }
    }

    if (my $team = $component->my_team) {
      if ($self->in_group_id($team->group_id)) {
        $self->{can_manage_comp}->{$c_id} = 1;
      }
    }
  }

  return $self->{can_manage_comp}->{$c_id};
}

# TODO not used yet, may delete it.
sub _can_manage_comp_field {
  my $self      = shift;
  my $component = shift;
  my $field     = shift;

  my $can = 0;

  if ($self->can_manage_comp($component)) {
    $can = 1;
  }

  return $can;
}


1;


__END__

=head1 NAME

Bugzilla::Extension::Workflows::ManageComponents

=head1 DESCRIPTION

This workflow allows teams and product owners to edit component roles without
granting editcomponents to large numbers of users.

A user is allowed to manage components this way when:

=over 4

=item They have editcomponents right on the product

=item They are in the component's default pool's team

=item They are in the component's sub_team

=back

For the API see L<Bugzilla::Extension::Workflows::WebService::ManageComponents>.

The columns in the the table are:

Name:
  The name of the component or sub_component.
Product:
  The name of the product the component belongs to.
Default assignee:
  The login of the default assignee.
Default qa:
  The login of the default QA contact.
Default docs:
  The login of the default docs contact.
Initial CCs:
  The logins of all users to be added to the CC list of new bugs.
Default pool:
  The Default pool to assign bugs to.
Sub-team:
  A Team that is responsible for the component that is not the team that owns the default pool.
Template:
  The BVP bug template to use for ne bugs.
Description:
  The public description got the component.
Actions:
  Icon that allow actions related to the current row.
